// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include "Connect.h"

using namespace mcPolymer_ns;

class ConnectTest : public ::testing::Test {
public:
	ConnectTest() {
	}
	~ConnectTest() {
	}

protected:
	Fragment *f1;
	Fragment *f2;
	Fragment *f3;
	Fragment *f4;

private:
	void SetUp() override {
		f1 = new Fragment();
		f2 = new Fragment();
		f3 = new Fragment();
		f4 = new Fragment();
	}
	void TearDown() override {
		delete f1;
		delete f2;
		delete f3;
		delete f4;
	}
};

TEST_F(ConnectTest, Construction)
{
	Connect *con = new Connect();
	EXPECT_NE(con, nullptr);
	delete con;
	con = new Connect(f1, f2, ConnectType::Termination);
	EXPECT_NE(con, nullptr);
	delete con;

	Connect con1(f1, f2, ConnectType::Crosslink);
	Connect con2(&con1);
	Connect con4(f2, f1, ConnectType::Crosslink);
	EXPECT_EQ(con1.getConnect1(), con2.getConnect1());
	EXPECT_EQ(con1.getConnect2(), con2.getConnect2());
	EXPECT_EQ(con1.getType(), con2.getType());

	EXPECT_NE(con1.getConnect1(), con4.getConnect1());
	EXPECT_NE(con1.getConnect2(), con4.getConnect2());
	EXPECT_EQ(con1.getType(), con4.getType());

}

TEST_F(ConnectTest, GetAndSet)
{
	Connect con(f1, f2, ConnectType::Crosslink);
	EXPECT_EQ(con.getConnect1(), f1);
	EXPECT_EQ(con.getConnect2(), f2);
	EXPECT_EQ(con.getType(), ConnectType::Crosslink);

	con.setFragment(1, f3);
	EXPECT_NE(con.getConnect1(), f1);
	EXPECT_EQ(con.getConnect1(), f3);

	EXPECT_NO_THROW(con.setFragment(1024, f3));
	EXPECT_EQ(con.getConnect1(), f3);
	EXPECT_EQ(con.getConnect2(), f2);

	con.setType(ConnectType::Undefined);
	EXPECT_EQ(con.getType(), ConnectType::Undefined);

	con.setFragment(2, f4);
	EXPECT_EQ(con.getConnect2(), f4);

	Connect con_null;
	EXPECT_EQ(con_null.getConnect1(), nullptr);
	EXPECT_EQ(con_null.getConnect2(), nullptr);
	EXPECT_EQ(con_null.getType(), ConnectType::Undefined);
}

TEST_F(ConnectTest, Connect)
{
	Connect con(f1, f2, ConnectType::Crosslink);
	EXPECT_TRUE(con.isConnected(f1));
	EXPECT_TRUE(con.isConnected(f2));
	EXPECT_FALSE(con.isConnected(f3));

	con.setFragment(2, f4);
	EXPECT_TRUE(con.isConnected(f4));
	EXPECT_FALSE(con.isConnected(f2));

	Connect con_same(f1, f1, ConnectType::Crosslink);
	EXPECT_TRUE(con_same.isConnected(f1));

	EXPECT_EQ(con.getConnect12(f1), f4);
	EXPECT_EQ(con.getConnect12(f4), f1);
	EXPECT_EQ(con_same.getConnect12(f1), f1);

	EXPECT_EQ(con.getConnect12(f2), nullptr);
	EXPECT_EQ(con_same.getConnect12(f2), nullptr);

	Connect con_null;
	EXPECT_TRUE(con_null.isConnected(nullptr));
	// FIXME: Expected behavior?
	EXPECT_EQ(con_null.getConnect12(nullptr), nullptr);
}
