// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include "Fragment.h"
#include <iostream>

using namespace mcPolymer_ns;

class FragmentTest : public ::testing::Test {
public:
	FragmentTest() {
	}
	~FragmentTest() {
	}

protected:
	Fragment *frag;
	const std::string testfileIn = "src/gtest/Fragment_Test/infile.data";
	const std::string testfileTurnAround = "build/gtest/fragment_turnaround.data";

private:
	void SetUp() override {
		numberMonomers = 4000;
		frag = new Fragment();
		for (size_t i = 0; i < numberMonomers; i++) {
			frag->addM(i);
			if (i % 30 == 0) {
				frag->addSCB(45+i);
				frag->addM(i);
				frag->addM(i);
			}
		}
	}
	void TearDown() override {
		delete frag;
	}
};

TEST_F(FragmentTest, Construction)
{
	Fragment *f = new Fragment();
	EXPECT_NE(f, nullptr);
	delete f;
	f = new Fragment(size_t(0));
	EXPECT_NE(f, nullptr);
	delete f;
	f = new Fragment(15, 50);
	EXPECT_NE(f, nullptr);
	delete f;
}

TEST_F(FragmentTest, Copy)
{
	Fragment *f = new Fragment(15, 50);
	Fragment *copy = new Fragment(f);
	EXPECT_EQ(f->getSCB(), copy->getSCB());
	EXPECT_EQ(f->getMonomers()[3999], copy->getMonomers()[3999]);
	EXPECT_EQ(f->getLength(), copy->getLength());
	EXPECT_EQ(f->getM(15), copy->getM(15));
	delete f;
	delete copy;
}

TEST_F(FragmentTest, Split)
{
	EXPECT_EQ(frag->getLength(), 4268);
	EXPECT_EQ(frag->getM(30), 3);
	EXPECT_EQ(frag->getSCB(), 11216);
	EXPECT_EQ(frag->getMonomers()[100], 1);

	Fragment *f_05 = new Fragment(frag, 0.5);
	EXPECT_EQ(f_05->getLength(), 4134);
	EXPECT_EQ(f_05->getM(30), 2);
	EXPECT_EQ(f_05->getSCB(), 5608);
	EXPECT_EQ(f_05->getMonomers()[100], 1);

	EXPECT_EQ(frag->getLength(), 134);
	EXPECT_EQ(frag->getM(30), 1);
	EXPECT_EQ(frag->getSCB(), 5608);
	EXPECT_EQ(frag->getMonomers()[100], 0);

	Fragment *f_02 = new Fragment(f_05, 0.2);
	EXPECT_EQ(f_02->getLength(), 0);
	EXPECT_EQ(f_02->getM(30), 0);
	EXPECT_EQ(f_02->getSCB(), 1122);
	EXPECT_EQ(f_02->getMonomers()[100], 0);

	// Carry over everthing from frag
	Fragment *f_10 = new Fragment(frag, 1.0);
	EXPECT_EQ(f_10->getLength(), 134);
	EXPECT_EQ(f_10->getM(30), 1);
	EXPECT_EQ(f_10->getSCB(), 5608);
	EXPECT_EQ(f_10->getMonomers()[100], 0);

	// What happens if we take more?
	Fragment *f_20 = new Fragment(f_10, 2.0);
	EXPECT_EQ(f_20->getLength(), 268);
	EXPECT_EQ(f_20->getM(30), 2);
	EXPECT_EQ(f_20->getSCB(), 11216);
	EXPECT_EQ(f_20->getMonomers()[100], 0);

	// Overflow!
	EXPECT_EQ(f_10->getLength(), 4294967162);
	EXPECT_EQ(f_10->getM(30), -1);
	EXPECT_EQ(f_10->getSCB(), 59928);
	EXPECT_EQ(f_10->getMonomers()[100], 0);

	delete f_05;
	delete f_02;
	delete f_10;
	delete f_20;
}

TEST_F(FragmentTest, AddDeleteIncrease)
{
	EXPECT_EQ(frag->getLength(), 4268);
	EXPECT_EQ(frag->getM(30), 3);
	EXPECT_EQ(frag->getSCB(), 11216);
	EXPECT_EQ(frag->getMonomers()[100], 1);

	frag->addM(30);
	EXPECT_EQ(frag->getM(30), 4);
	EXPECT_EQ(frag->getLength(), 4269);
	frag->deleteM(30);
	EXPECT_EQ(frag->getM(30), 3);
	frag->deleteM(30, 2);
	EXPECT_EQ(frag->getM(30), 1);
	EXPECT_EQ(frag->getLength(), 4266);

	frag->incrSCB();
	EXPECT_EQ(frag->getSCB(), 11217);
	frag->addSCB(25);
	EXPECT_EQ(frag->getSCB(), 11242);
	frag->deleteSCB(11000);
	EXPECT_EQ(frag->getSCB(), 242);
}

TEST_F(FragmentTest, ImportExport)
{
	std::ifstream is(testfileIn, std::ios::binary);
	ASSERT_TRUE(is.is_open());
	Fragment f;
	f.importFragment(&is);
	is.close();

	EXPECT_EQ(frag->getLength(), f.getLength());
	EXPECT_EQ(frag->getM(30), f.getM(30));
	EXPECT_EQ(frag->getSCB(), f.getSCB());
	EXPECT_EQ(frag->getMonomers()[100], f.getMonomers()[100]);

	frag->addM(30);
	EXPECT_EQ(frag->getLength(), 4269);
	EXPECT_EQ(frag->getSCB(), 11216);
	frag->deleteSCB(11000);
	EXPECT_EQ(frag->getSCB(), 216);

	std::ofstream os(testfileTurnAround, std::ios::binary);
	ASSERT_TRUE(os.is_open());
	EXPECT_NO_THROW(frag->exportFragment(&os));
	os.close();

	is.open(testfileTurnAround, std::ios::binary);
	ASSERT_TRUE(is.is_open());
	f.importFragment(&is);
	is.close();

	EXPECT_EQ(f.getLength(), 4269);
	EXPECT_EQ(f.getSCB(), 216);
}
