// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include "Histogram.h"
#include "MersenneTwister.h"

using namespace mcPolymer_ns;

class HistogramTest : public ::testing::Test {
public:
	HistogramTest() {
	}
	~HistogramTest() {
	}

protected:
		Histogram single;
		Histogram triple;
		Histogram single_small;
		Histogram triple_small;


private:
	void SetUp() override {
		MTRand rand(10);
		for (unsigned i = 0; i < 500; i++) {
			int rv1 = rand.randInt(100)-50;
			int rv2 = rand.randInt(200)-100;
			int rv3 = rand.randInt(500)-100;
			single.addValue(i, rv1);
			triple.addValue(i, rv1, rv2, rv3);
		}
		for (unsigned i = 0; i < 10; i++) {
			int rv1 = rand.randInt(100)-50;
			int rv2 = rand.randInt(200)-100;
			int rv3 = rand.randInt(500)-100;
			single_small.addValue(i+10, rv1);
			triple_small.addValue(i+10, rv1, rv2, rv3);
		}
	}
	void TearDown() override {
	}
};

TEST_F(HistogramTest, Construction)
{
	Histogram *h = new Histogram();
	EXPECT_NE(h, nullptr);
	delete h;
}

TEST_F(HistogramTest, GetSize)
{
	EXPECT_EQ(single.getSize(), 500);
	EXPECT_EQ(triple.getSize(), 500);

	EXPECT_EQ(single.getSx(), 124750);
	EXPECT_EQ(triple.getSx(), 124750);

	// FIXME: Overflow!
	EXPECT_EQ(single.getSy1(), 18446744073709550399ull);
	EXPECT_EQ(triple.getSy1(), 18446744073709550399ull);

	EXPECT_EQ(single.getSy2(), 0);
	EXPECT_EQ(triple.getSy2(), 679);

	EXPECT_EQ(single.getSy3(), 0);
	EXPECT_EQ(triple.getSy3(), 74313);
}

TEST_F(HistogramTest, GetValues)
{
	EXPECT_EQ(single.getX(50), 50);
	EXPECT_EQ(triple.getX(50), 50);
	EXPECT_EQ(single.getX(70), 70);
	EXPECT_EQ(triple.getX(70), 70);
	EXPECT_EQ(single.getX(470), 470);
	EXPECT_EQ(triple.getX(470), 470);
	EXPECT_EQ(single.getX(9999), 0);
	EXPECT_EQ(triple.getX(9999), 0);

	EXPECT_EQ(single.getY1(50), 12);
	EXPECT_EQ(triple.getY1(50), 12);
	EXPECT_EQ(single.getY1(70), 43);
	EXPECT_EQ(triple.getY1(70), 43);
	EXPECT_EQ(single.getY1(470), 13);
	EXPECT_EQ(triple.getY1(470), 13);
	EXPECT_EQ(single.getY1(9999), 0);
	EXPECT_EQ(triple.getY1(9999), 0);

	EXPECT_EQ(single.getY2(50), 0);
	EXPECT_EQ(triple.getY2(50), -23);
	EXPECT_EQ(single.getY2(70), 0);
	EXPECT_EQ(triple.getY2(70), -16);
	EXPECT_EQ(single.getY2(470), 0);
	EXPECT_EQ(triple.getY2(470), -86);
	EXPECT_EQ(single.getY2(9999), 0);
	EXPECT_EQ(triple.getY2(9999), 0);

	EXPECT_EQ(single.getY3(50), 0);
	EXPECT_EQ(triple.getY3(50), 263);
	EXPECT_EQ(single.getY3(70), 0);
	EXPECT_EQ(triple.getY3(70), 332);
	EXPECT_EQ(single.getY3(470), 0);
	EXPECT_EQ(triple.getY3(470), -24);
	EXPECT_EQ(single.getY3(9999), 0);
	EXPECT_EQ(triple.getY3(9999), 0);
}

TEST_F(HistogramTest, AddValues)
{
	EXPECT_EQ(single.getX(50), 50);
	EXPECT_EQ(single.getY1(50), 12);
	single.addValue(50, 100);
	EXPECT_EQ(single.getX(50), 50);
	EXPECT_EQ(single.getY1(50), 112);
	single.addValue(50, -113);
	EXPECT_EQ(single.getY1(50), -1);

	EXPECT_EQ(triple.getX(50), 50);
	EXPECT_EQ(triple.getY1(50), 12);
	EXPECT_EQ(triple.getY2(50), -23);
	EXPECT_EQ(triple.getY3(50), 263);
	triple.addValue(50, 10, 10, 10);
	EXPECT_EQ(triple.getX(50), 50);
	EXPECT_EQ(triple.getY1(50), 22);
	EXPECT_EQ(triple.getY2(50), -13);
	EXPECT_EQ(triple.getY3(50), 273);

	// FIXME: Input parameter should be unsigned?
	single.addValue(-5, 100);
	EXPECT_EQ(single.getX(-5), 0);
	EXPECT_EQ(single.getY1(-5), 0);
	triple.addValue(-150, 10, 10, 10);
	EXPECT_EQ(triple.getX(-150), 0);
	EXPECT_EQ(triple.getY1(-150), 0);
	EXPECT_EQ(triple.getY2(-150), 0);
	EXPECT_EQ(triple.getY3(-150), 0);

	// Spacious Histgram
	single_small.addValue(0, 111);
	EXPECT_EQ(single_small.getX(0), 0);
	EXPECT_EQ(single_small.getY1(0), 111);
	triple_small.addValue(0, 111, 222, 333);
	EXPECT_EQ(triple_small.getX(0), 0);
	EXPECT_EQ(triple_small.getY1(0), 111);
	EXPECT_EQ(triple_small.getY2(0), 222);
	EXPECT_EQ(triple_small.getY3(0), 333);
}

TEST_F(HistogramTest, GetPn)
{
	Histogram empty;
	EXPECT_NEAR(empty.getPn(), 0, 0.00001);

	// Affects y1 only
	EXPECT_NEAR(single.getPn(), -1.72009e-14, 1e-18);
	EXPECT_NEAR(triple.getPn(), -1.72009e-14, 1e-18);

	single.addValue(25, 8000000);
	EXPECT_NEAR(single.getPn(), 24.964, 1e-3);
}

TEST_F(HistogramTest, GetPw)
{
	Histogram empty;
	EXPECT_NEAR(empty.getPw(), 0, 0.00001);

	// Affects y1 only
	EXPECT_NEAR(single.getPw(), 434.98, 0.01);
	EXPECT_NEAR(triple.getPw(), 434.98, 0.01);

	single.addValue(25, 8000000);
	EXPECT_NEAR(single.getPw(), 24.3485, 0.001);
}
