// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
#include <gtest/gtest.h>
#include <string>
#include <valarray>
#include "modelObjects.h"
#include "Polymer.h"
#include "polymerSpecies.h"

using namespace mcPolymer_ns;

class PolymerSpeciesTest : public ::testing::Test {
public:
	PolymerSpeciesTest() {
	}
	~PolymerSpeciesTest() {
	}

protected:
	polymerSpecies *psTest;
	polymerSpecies *psTestPopulated;
	polymerSpecies *psTestPopulatedTransfer;
	
private:
	void SetUp() override {
		psTest = new polymerSpecies(1);
		Polymer* p1 = new Polymer(1, 10);
		psTest->addPolymer(p1);
		psTestPopulated = new polymerSpecies(2);
		Polymer* p2 = new Polymer(1, 20);
		p2->activateECR();
		psTestPopulated->addPolymer(p2);
		Polymer* p3 = new Polymer(1, 30);
		p3->activateECR();
		psTestPopulated->addPolymer(p3);
		psTestPopulatedTransfer = new polymerSpecies(3);
		Polymer* p4 = new Polymer(0, 40);
		p4->activateECR();
		psTestPopulatedTransfer->setTransferMonomer(0);
		psTestPopulatedTransfer->addPolymer(p4);
		Polymer* p5 = new Polymer(0, 50);
		p5->activateECR();
		psTestPopulatedTransfer->addPolymer(p5);
		Polymer* p6 = new Polymer(1, 60);
		p6->activateECR();
		psTestPopulatedTransfer->addPolymer(p6);
		
	}
	void TearDown() override {
		delete psTest;
		delete psTestPopulated;
		delete psTestPopulatedTransfer;
	}
};
TEST_F(PolymerSpeciesTest, Construction)
{
	//Polymer *p = new Polymer();
	polymerSpecies* pS = new polymerSpecies(1);
	EXPECT_NE(pS, nullptr);
	delete pS;
}

TEST_F(PolymerSpeciesTest, General)
{
	EXPECT_EQ(psTest->getSize(), 1);
	polymerSpeciesProperties psTestP;
	psTestP = psTest->getPolymerSpeciesProperties();
	EXPECT_EQ(psTestP.numPolymers, 1);
	Polymer* p1 = psTest->deleteP();
	EXPECT_NE(p1, nullptr);
	EXPECT_EQ(p1->getLength(), 10);
	p1 = psTest->deleteP();
	EXPECT_EQ(p1, nullptr);
	EXPECT_EQ(psTest->getSize(), 0);
	psTest->getPolymerSpeciesProperties();
	
	
	EXPECT_EQ(psTestPopulated->getSize(), 2);
	Polymer* px = psTestPopulated->deleteP();
	EXPECT_NE(px, nullptr);
	EXPECT_EQ(psTestPopulated->getSize(), 1);
	Polymer* py	= psTestPopulated->deleteP();
	EXPECT_NE(py, nullptr);
	EXPECT_EQ(psTestPopulated->getSize(), 0);
	if (px->getLength() < py->getLength()) {
		EXPECT_EQ(px->getLength(), 20);
		EXPECT_EQ(py->getLength(), 30);
	}
	else {
		EXPECT_EQ(px->getLength(), 30);
		EXPECT_EQ(py->getLength(), 20);
	}
}
TEST_F(PolymerSpeciesTest, newP1)
{
	psTestPopulated->newP1(1);
	psTestPopulated->newP1(1);
	psTestPopulated->newP1(1);
	psTestPopulated->newP1(0);
	psTestPopulated->newP1(0);
	polymerSpeciesProperties psTestP;
	psTestP = psTestPopulated->getPolymerSpeciesProperties();
	EXPECT_EQ(psTestP.numMonomers[0], 2);
	EXPECT_EQ(psTestP.numMonomers[1], 53);
	EXPECT_EQ(psTestP.numPolymers, 7);
	
}
TEST_F(PolymerSpeciesTest, Propagation)
{
	psTestPopulated->incrP(1);
	psTestPopulated->incrP(1);
	polymerSpeciesProperties psTestP;
	psTestP = psTestPopulated->getPolymerSpeciesProperties();
	// TODO: next line correct assumption?
	EXPECT_EQ(psTestP.numMonomers[1], 54);
}
TEST_F(PolymerSpeciesTest, Transfer)
{
	EXPECT_EQ(psTestPopulatedTransfer->getNumTransferMonomers(), 90);
	Polymer* pt = psTestPopulatedTransfer->deletePolymerForTransfer(0);
	if (pt->getLength() > 40) {
		EXPECT_EQ(pt->getLength(), 50);
		EXPECT_EQ(psTestPopulatedTransfer->getNumTransferMonomers(), 40);
		Polymer* pt1 = psTestPopulatedTransfer->deletePolymerForTransfer(0);
		EXPECT_EQ(pt1->getLength(), 40);
		EXPECT_EQ(psTestPopulatedTransfer->getNumTransferMonomers(), 0);
	}
	else {
		EXPECT_EQ(pt->getLength(), 40);
		EXPECT_EQ(psTestPopulatedTransfer->getNumTransferMonomers(), 50);
		Polymer* pt1 = psTestPopulatedTransfer->deletePolymerForTransfer(0);
		EXPECT_EQ(pt1->getLength(), 50);
		EXPECT_EQ(psTestPopulatedTransfer->getNumTransferMonomers(), 0);
	}
	polymerSpeciesProperties psTestT;
	psTestT = psTestPopulatedTransfer->getPolymerSpeciesProperties();
	EXPECT_EQ(psTestT.numMonomers[1], 60);
}
