// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include <string>
#include <valarray>
#include "RandomBuffer.h"
#include "Polymer.h"

using namespace mcPolymer_ns;

class PolymerTest : public ::testing::Test {
public:
	PolymerTest() {
	}
	~PolymerTest() {
	}

protected:
	Polymer *ptest;
	Polymer *populated;

	Fragment *f1, *f2, *f3, *f4;
	Connect *c1, *c2;

	const std::string importFile = "src/gtest/Polymer_Test/polymerimport.data";
	const std::string exportFile = "build/gtest/polymerexport.data";

private:
	void SetUp() override {
		numberMonomers = 1000;
		ptest = new Polymer(2, 40);

		f1 = new Fragment(1, 20);
		f2 = new Fragment(2, 25);
		f3 = new Fragment(3, 30);
		f4 = new Fragment(4, 35);
		populated = new Polymer(1, 20);
		populated->addFragment(f1);
		populated->addFragment(f2);
		populated->addFragment(f3);
		populated->addFragment(f4);
		populated->addConnect(f1, f2, ConnectType::Termination);
		populated->addConnect(f3, f4, ConnectType::LCB);
	}
	void TearDown() override {
		delete ptest;
		delete populated;
	}
};

TEST_F(PolymerTest, Construction)
{
	numberMonomers = 100;
	Polymer *p = new Polymer();
	EXPECT_NE(p, nullptr);
	delete p;
	p = new Polymer(1);
	EXPECT_NE(p, nullptr);
	delete p;

	p = new Polymer(1, 50);
	EXPECT_NE(p, nullptr);
	delete p;

	//  Exceed number of monomers causes coredump
	/*
	p = new Polymer(-21, 120);
	EXPECT_NE(p, nullptr);
	EXPECT_EQ(numberMonomers, 100);
	delete p;
	*/

	Fragment *f = new Fragment(1, 5);
	p = new Polymer(f);
	EXPECT_NE(p, nullptr);
	delete p;
}

TEST_F(PolymerTest, General)
{
	const std::valarray<uint32_t> monomers = populated->getMonomers();
	EXPECT_EQ(monomers[0], 0);
	EXPECT_EQ(monomers[1], 40);
	EXPECT_EQ(populated->getLength(), 130);
	EXPECT_EQ(populated->getM(1), 40);

	// FIXME: Produces coredump
	// populated->getM(101110), std::exception);
	EXPECT_EQ(populated->getSCB(), 0);

}

TEST_F(PolymerTest, ImportExport)
{
	Polymer *pi = new Polymer();
	std::ifstream is(importFile, std::ios::binary);
	ASSERT_TRUE(is.is_open());
	ASSERT_TRUE(pi->importPolymer(&is));

	EXPECT_EQ(populated->getLength(), pi->getLength());
	EXPECT_EQ(populated->getM(1), pi->getM(1));
	// FIXME: causes coredump
	// EXPECT_EQ(populated->getLengthPropagating(), pi->getLengthPropagating());
	EXPECT_EQ(populated->getSCB(), pi->getSCB());

	std::ofstream os(exportFile, std::ios::binary);
	ASSERT_TRUE(os.is_open());
	EXPECT_TRUE(populated->exportPolymer(&os));

	delete pi;
}

TEST_F(PolymerTest, Selection)
{
	EXPECT_FALSE(ptest->getSelection());
	ptest->deselect();
	EXPECT_FALSE(ptest->getSelection());
	ptest->select();
	EXPECT_TRUE(ptest->getSelection());
	ptest->deselect();
	EXPECT_FALSE(ptest->getSelection());
}

TEST_F(PolymerTest, FragmentManipulation)
{
	Polymer *p = new Polymer();
	EXPECT_EQ(p->getFragmentSize(), 0);
	EXPECT_EQ(ptest->getFragmentSize(), 1);

	EXPECT_NE(ptest->getFragment(0), nullptr);
	EXPECT_THROW(ptest->getFragment(1), std::exception);
	EXPECT_EQ(ptest->getFragment(0), ptest->getLastFragment());
	Fragment f;
	ptest->addFragment(&f);
	EXPECT_NE(ptest->getFragment(0), ptest->getLastFragment());
	EXPECT_NO_THROW(ptest->getFragment(1));
	EXPECT_EQ(ptest->getFragmentSize(), 2);
	EXPECT_EQ(ptest->deleteFragment(), &f);
	EXPECT_EQ(ptest->getFragmentSize(), 1);

	// Default seems to be an activated fragment
	EXPECT_EQ(ptest->getPropagatingFragment(), ptest->getFragment(0));
}

TEST_F(PolymerTest, ConnectManipulation)
{
	c1 = new Connect(f1, f4, ConnectType::Termination);
	EXPECT_EQ(populated->getConnects(), 2);
	populated->addConnect(c1);
	EXPECT_EQ(populated->getConnect(99, 1), nullptr);
	EXPECT_EQ(populated->getConnect(0, 3), nullptr);
	EXPECT_EQ(populated->getConnect(0, 1), f1);

	EXPECT_EQ(populated->getConnectType(0), ConnectType::Termination);
	EXPECT_EQ(populated->getConnectType(1), ConnectType::LCB);
	EXPECT_EQ(populated->getConnectType(2), ConnectType::Termination);
	EXPECT_EQ(populated->getConnectType(5), ConnectType::Undefined);

	EXPECT_EQ(populated->getConnects(), 3);
	EXPECT_EQ(populated->getConnect(2), c1);
	EXPECT_EQ(populated->deleteConnect(), c1);
	EXPECT_EQ(populated->getConnects(), 2);
	populated->deleteLastConnect();
	EXPECT_EQ(populated->getConnects(), 1);

}
