// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include "RandomBuffer.h"
#include "Reaction.h"
#include <iostream>
#include <vector>

using namespace mcPolymer_ns;

struct SimulationData {
	std::valarray<uint64_t> moleculeCounter;
	std::valarray<double> reactionRateCoefficientsMC;
	std::valarray<polymerSpecies*> macromolSpecies;
	uint64_t solidsContentMoleculeCounter;
	uint64_t solidsContentMoleculeCounterEduct;
	uint64_t styreneMoleculeCounter;
};

class ReactionTest : public ::testing::Test {
public:
	ReactionTest() {
	}
	~ReactionTest() {
	}

protected:
	SimulationData data;
	ElementaryReaction* instantiateElementaryReaction() { return new ElementaryReaction(&data.reactionRateCoefficientsMC[0],
																std::vector<uint64_t*> { &data.moleculeCounter[0], &data.moleculeCounter[3] },
																std::vector<uint64_t*> { &data.moleculeCounter[1] }); }
	Initiation* instantiateInitiation() { return new Initiation(&data.reactionRateCoefficientsMC[1],
																&data.moleculeCounter[3], &data.moleculeCounter[4],
																&data.moleculeCounter[0], data.macromolSpecies[0], 55, &data.solidsContentMoleculeCounter); }
	InitiatorDecomposition* instantiateInitiatorDecomposition() { return new InitiatorDecomposition(&data.reactionRateCoefficientsMC[0],
																&data.moleculeCounter[4], &data.moleculeCounter[1],
																0.78); }
	Propagation_SCB* instantiatePropagation_SCB() { return new Propagation_SCB(&data.reactionRateCoefficientsMC[2],
																&data.moleculeCounter[2], &data.moleculeCounter[1], &data.moleculeCounter[3],
																data.macromolSpecies[2], data.macromolSpecies[1],
																55, &data.solidsContentMoleculeCounter); }
	Propagation_PnPnp1* instantiatePropagation_PnPnp1() { return new Propagation_PnPnp1(&data.reactionRateCoefficientsMC[3],
																&data.moleculeCounter[0], &data.moleculeCounter[2], &data.moleculeCounter[4],
																data.macromolSpecies[1], data.macromolSpecies[0],
																60, &data.solidsContentMoleculeCounter); }
	Propagation_PnQnp1* instantiatePropagation_PnQnp1() { return new Propagation_PnQnp1(&data.reactionRateCoefficientsMC[3],
																&data.moleculeCounter[2], &data.moleculeCounter[3], &data.moleculeCounter[4],
																data.macromolSpecies[2], data.macromolSpecies[0],
																4, &data.solidsContentMoleculeCounter); }
	Termination_PnPmDnDm* instantiateTermination_PnPmDnDm(size_t _init = 2) { return new Termination_PnPmDnDm(&data.reactionRateCoefficientsMC[1],
																&data.moleculeCounter[0], &data.moleculeCounter[_init], &data.moleculeCounter[4], &data.moleculeCounter[3],
																data.macromolSpecies[2], data.macromolSpecies[1], data.macromolSpecies[3], data.macromolSpecies[0]); }
	Termination_PnPmDnpm* instantiateTermination_PnPmDnpm(size_t _init = 2) { return new Termination_PnPmDnpm(&data.reactionRateCoefficientsMC[0],
																&data.moleculeCounter[0], &data.moleculeCounter[_init], &data.moleculeCounter[3],
																data.macromolSpecies[2], data.macromolSpecies[3], data.macromolSpecies[0]); }
	Transfer* instantiateTransfer() { return new Transfer(&data.reactionRateCoefficientsMC[3],
																std::vector<uint64_t*> { &data.moleculeCounter[2], &data.moleculeCounter[3] },
																std::vector<uint64_t*> { &data.moleculeCounter[1], &data.moleculeCounter[0] },
																data.macromolSpecies[0], data.macromolSpecies[3]); }
	TransferToMonomer* instantiateTransferToMonomer() { return new TransferToMonomer(&data.reactionRateCoefficientsMC[3],
																&data.moleculeCounter[1], &data.moleculeCounter[3], &data.moleculeCounter[4], &data.moleculeCounter[2],
																data.macromolSpecies[3], data.macromolSpecies[1], data.macromolSpecies[2],
																6, &data.solidsContentMoleculeCounter); }
	TransferToPolymer* instantiateTransferToPolymer() { return new TransferToPolymer(&data.reactionRateCoefficientsMC[2],
																&data.moleculeCounter[2], &data.moleculeCounter[3], &data.moleculeCounter[4], &data.moleculeCounter[1],
																data.macromolSpecies[0], data.macromolSpecies[3], data.macromolSpecies[1], data.macromolSpecies[2],
																6, 5, &data.solidsContentMoleculeCounterEduct,  &data.solidsContentMoleculeCounter,
																PropagatingRadicalType::MCRL); }
private:
	void SetUp() override {
		data.moleculeCounter.resize(5);
		data.moleculeCounter[0] = 10;
		data.moleculeCounter[1] = 400;
		data.moleculeCounter[2] = 1000;
		data.moleculeCounter[3] = 15000;
		data.moleculeCounter[4] = 300000;

		data.reactionRateCoefficientsMC.resize(4);
		data.reactionRateCoefficientsMC[0] = 1.24;
		data.reactionRateCoefficientsMC[1] = 5.0e-4;
		data.reactionRateCoefficientsMC[2] = 3.3e+3;
		data.reactionRateCoefficientsMC[3] = 10e-10;

		data.macromolSpecies.resize(4);
		data.macromolSpecies[0] = new polymerSpecies(4);
		data.macromolSpecies[0]->newP1(5);
		data.macromolSpecies[1] = new polymerSpecies(5);
		data.macromolSpecies[1]->newP1(3);
		data.macromolSpecies[2] = new polymerSpecies(6);
		data.macromolSpecies[2]->newP1(2);
		data.macromolSpecies[3] = new polymerSpecies(7);
		data.macromolSpecies[3]->newP1(1);

		data.solidsContentMoleculeCounter = 10;
		data.solidsContentMoleculeCounterEduct = 50;
		data.styreneMoleculeCounter = 5;
	}
	void TearDown() override {
		delete data.macromolSpecies[0];
		delete data.macromolSpecies[1];
		delete data.macromolSpecies[2];
		delete data.macromolSpecies[3];
	}
};

TEST_F(ReactionTest, ElementaryReaction_Construction)
{
	ElementaryReaction* r = instantiateElementaryReaction();
	EXPECT_NE(r, nullptr);
	delete r;
}

TEST_F(ReactionTest, ElementaryReaction_PartialProbability)
{
	ElementaryReaction* r = instantiateElementaryReaction();
	EXPECT_NEAR(r->calculatePartialProbability(), 186000.0, 1.0);
	delete r;

	// Same educt case
	r = new ElementaryReaction(&data.reactionRateCoefficientsMC[1],
							std::vector<uint64_t*> { &data.moleculeCounter[0], &data.moleculeCounter[0] },
							std::vector<uint64_t*> { &data.moleculeCounter[1] });
	EXPECT_NEAR(r->calculatePartialProbability(), 0.0225, 1.0e-5);
	delete r;
}

TEST_F(ReactionTest, ElementaryReaction_Execution)
{
	ElementaryReaction* r = instantiateElementaryReaction();
	r->execute();
	EXPECT_EQ(data.moleculeCounter[0], 9);
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[1], 401);
	delete r;
}

TEST_F(ReactionTest, Initiation_Construction)
{
	Initiation* i = instantiateInitiation();
	EXPECT_NE(i, nullptr);
	delete i;
}

TEST_F(ReactionTest, Initiation_PartialProbability)
{
	Initiation* i = instantiateInitiation();
	EXPECT_NEAR(i->calculatePartialProbability(), 2250000.0, 1.0);
	delete i;
}

TEST_F(ReactionTest, Initiation_Execution)
{
	Initiation* i = instantiateInitiation();
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	i->execute();
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[4], 299999);
	EXPECT_EQ(data.moleculeCounter[0], 11);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 2);
	EXPECT_EQ(data.solidsContentMoleculeCounter, 11);
	delete i;
}

TEST_F(ReactionTest, InitiatorDecomposition_Construction)
{
	InitiatorDecomposition* i = instantiateInitiatorDecomposition();
	EXPECT_NE(i, nullptr);
	delete i;
}

TEST_F(ReactionTest, InitiatorDecomposition_PartialProbability)
{
	InitiatorDecomposition* i = instantiateInitiatorDecomposition();
	EXPECT_NEAR(i->calculatePartialProbability(), 372000.0, 1.0);
	delete i;
}

TEST_F(ReactionTest, InitiatorDecomposition_Execution)
{
	InitiatorDecomposition* i = instantiateInitiatorDecomposition();
	randBufferInstance = new RandomBuffer(1);
	randBufferInstance->fillRands(1); // use seed for testing
	EXPECT_EQ(data.moleculeCounter[4], 300000);
	EXPECT_EQ(data.moleculeCounter[1], 400);
	i->execute();
	EXPECT_EQ(data.moleculeCounter[4], 299999);
	EXPECT_EQ(data.moleculeCounter[1], 401);
	for (unsigned s = 0; s < 140; s++)
		i->execute();
	EXPECT_EQ(data.moleculeCounter[4], 299859);
	EXPECT_EQ(data.moleculeCounter[1], 624);
	delete i;
	delete randBufferInstance;
}

TEST_F(ReactionTest, Propagation_SCB_Construction)
{
	Propagation_SCB* s = instantiatePropagation_SCB();
	EXPECT_NE(s, nullptr);
	delete s;
}

TEST_F(ReactionTest, Propagation_SCB_PartialProbability)
{
	Propagation_SCB* s = instantiatePropagation_SCB();
	EXPECT_NEAR(s->calculatePartialProbability(), 1.32e+9, 1.0e+3);
	delete s;
}

TEST_F(ReactionTest, Propagation_SCB_Execution)
{
	Propagation_SCB* s = instantiatePropagation_SCB();
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	s->execute();
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[1], 399);
	EXPECT_EQ(data.moleculeCounter[3], 15001);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 2);
	delete s;
}

TEST_F(ReactionTest, Propagation_PnPnp1_Construction)
{
	Propagation_PnPnp1* p = instantiatePropagation_PnPnp1();
	EXPECT_NE(p, nullptr);
	delete p;
}

TEST_F(ReactionTest, Propagation_PnPnp1_PartialProbability)
{
	Propagation_PnPnp1* p = instantiatePropagation_PnPnp1();
	EXPECT_NEAR(p->calculatePartialProbability(), 1.0e-5, 1.0e-8);
	delete p;
}

TEST_F(ReactionTest, Propagation_PnPnp1_Execution)
{
	Propagation_PnPnp1* p = instantiatePropagation_PnPnp1();
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	p->execute();
	EXPECT_EQ(data.moleculeCounter[0], 10);
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[4], 300000);
	// FIXME: macroEduct[0] unaffected?
	// FIXME: test addM
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	EXPECT_EQ(data.solidsContentMoleculeCounter, 11);
	delete p;
}

TEST_F(ReactionTest, Propagation_PnQnp1_Construction)
{
	Propagation_PnQnp1* q = instantiatePropagation_PnQnp1();
	EXPECT_NE(q, nullptr);
	delete q;
}

TEST_F(ReactionTest, Propagation_PnQnp1_PartialProbability)
{
	Propagation_PnQnp1* q = instantiatePropagation_PnQnp1();
	EXPECT_NEAR(q->calculatePartialProbability(), 0.015, 1.0e-4);
	delete q;
}

TEST_F(ReactionTest, Propagation_PnQnp1_Execution)
{
	Propagation_PnQnp1* q = instantiatePropagation_PnQnp1();
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	q->execute();
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[4], 300001);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 2);
	EXPECT_EQ(data.solidsContentMoleculeCounter, 11);
	// FIXME: macroEduct[0] useless?
	// FIXME: test addM
	delete q;
}

TEST_F(ReactionTest, Termination_PnPmDnDm_Construction)
{
	Termination_PnPmDnDm* tp = instantiateTermination_PnPmDnDm();
	EXPECT_NE(tp, nullptr);
	delete tp;
}

TEST_F(ReactionTest, Termination_PnPmDnDm_PartialProbability)
{
	Termination_PnPmDnDm* tp = instantiateTermination_PnPmDnDm();
	EXPECT_NEAR(tp->calculatePartialProbability(), 5.0, 1.0e-3);
	delete tp;
	tp = instantiateTermination_PnPmDnDm(0);
	EXPECT_NEAR(tp->calculatePartialProbability(), 0.0225, 1.0e-7);
	delete tp;
}

TEST_F(ReactionTest, Termination_PnPmDnDm_Execution)
{
	Termination_PnPmDnDm* tp = instantiateTermination_PnPmDnDm();
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	tp->execute();
	EXPECT_EQ(data.moleculeCounter[0], 9);
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[4], 300001);
	EXPECT_EQ(data.moleculeCounter[3], 15001);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 2);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 2);
	delete tp;
}

TEST_F(ReactionTest, Termination_PnPmDnpm_Construction)
{
	Termination_PnPmDnpm* tp = instantiateTermination_PnPmDnpm();
	EXPECT_NE(tp, nullptr);
	delete tp;
}

TEST_F(ReactionTest, Termination_PnPmDnpm_PartialProbability)
{
	Termination_PnPmDnpm* tp = instantiateTermination_PnPmDnpm();
	EXPECT_NEAR(tp->calculatePartialProbability(), 12400.0, 1.0);
	delete tp;
	tp = instantiateTermination_PnPmDnpm(0);
	EXPECT_NEAR(tp->calculatePartialProbability(), 55.8, 1.0e-3);
	delete tp;
}

TEST_F(ReactionTest, Termination_PnPmDnpm_Execution)
{
	Termination_PnPmDnpm* tp = instantiateTermination_PnPmDnpm();
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	tp->execute();
	EXPECT_EQ(data.moleculeCounter[0], 9);
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[3], 15001);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 2);
	delete tp;
}

TEST_F(ReactionTest, Transfer_Construction)
{
	Transfer* t = instantiateTransfer();
	EXPECT_NE(t, nullptr);
	delete t;
}

TEST_F(ReactionTest, Transfer_PartialProbability)
{
	Transfer* t = instantiateTransfer();
	EXPECT_NEAR(t->calculatePartialProbability(), 0.015, 1.0e-5);
	delete t;
}

TEST_F(ReactionTest, Transfer_Execution)
{
	Transfer* t = instantiateTransfer();
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 1);
	t->execute();
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[1], 401);
	EXPECT_EQ(data.moleculeCounter[0], 11);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 2);
	delete t;
}

TEST_F(ReactionTest, TransferToMonomer_Construction)
{
	TransferToMonomer* tm = instantiateTransferToMonomer();
	EXPECT_NE(tm, nullptr);
	delete tm;
}

TEST_F(ReactionTest, TransferToMonomer_PartialProbability)
{
	TransferToMonomer* tm = instantiateTransferToMonomer();
	EXPECT_NEAR(tm->calculatePartialProbability(), 0.006, 1.0e-6);
	delete tm;
}

TEST_F(ReactionTest, TransferToMonomer_Execution)
{
	TransferToMonomer* tm = instantiateTransferToMonomer();
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	tm->execute();
	EXPECT_EQ(data.moleculeCounter[1], 399);
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[4], 300001);
	EXPECT_EQ(data.moleculeCounter[2], 1001);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 2);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 2);
	EXPECT_EQ(data.solidsContentMoleculeCounter, 11);
	delete tm;
}

TEST_F(ReactionTest, TransferToPolymer_Construction)
{
	TransferToPolymer* tp = instantiateTransferToPolymer();
	EXPECT_NE(tp, nullptr);
	delete tp;
}

TEST_F(ReactionTest, TransferToPolymer_PartialProbability)
{
	TransferToPolymer* tp = instantiateTransferToPolymer();
	data.macromolSpecies[3]->setTransferMonomer(1);
	EXPECT_NEAR(tp->calculatePartialProbability(), 0.0, 1.0e-5);
	delete tp;
}

TEST_F(ReactionTest, TransferToPolymer_Execution)
{
	TransferToPolymer* tp = instantiateTransferToPolymer();
	randBufferInstance = new RandomBuffer(1);
	randBufferInstance->fillRands(1); // use seed for testing
	data.macromolSpecies[3]->setTransferMonomer(1);
	// FIXME: Transfer monomer implicitly == 0?
	data.macromolSpecies[3]->addPolymer(new Polymer(0, 15));
	data.macromolSpecies[3]->addPolymer(new Polymer(0, 25));
	data.macromolSpecies[3]->addPolymer(new Polymer(0, 50));
	data.macromolSpecies[3]->addPolymer(new Polymer(0, 75));

	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 4);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 1);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 1);
	tp->execute();
	EXPECT_EQ(data.moleculeCounter[2], 999);
	EXPECT_EQ(data.moleculeCounter[3], 14999);
	EXPECT_EQ(data.moleculeCounter[4], 300001);
	EXPECT_EQ(data.moleculeCounter[1], 401);
	EXPECT_EQ(data.macromolSpecies[0]->getSize(), 0);
	EXPECT_EQ(data.macromolSpecies[3]->getSize(), 3);
	EXPECT_EQ(data.macromolSpecies[1]->getSize(), 2);
	EXPECT_EQ(data.macromolSpecies[2]->getSize(), 2);
	EXPECT_EQ(data.solidsContentMoleculeCounter, 11);
	EXPECT_EQ(data.solidsContentMoleculeCounterEduct, 49);
	delete tp;
}
