// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <gtest/gtest.h>
#include "binaryTree.h"
#include "Polymer.h"
#include <algorithm>

using namespace mcPolymer_ns;

class BinaryTreeTest : public ::testing::Test {
public:
	BinaryTreeTest() {
	}
	~BinaryTreeTest() {
	}

protected:
	Polymer *p1;
	Polymer *p2;

private:
	void SetUp() override {
		p1 = new Polymer();
		p2 = new Polymer();
	}
	void TearDown() override {
		delete p1;
		delete p2;
	}
};

TEST_F(BinaryTreeTest, TreeConstruction)
{
	binTree *tree = new binTree(4);
	EXPECT_NE(tree, nullptr);
	delete tree;
}

TEST_F(BinaryTreeTest, TreeAddPolymer)
{
	binTree tree(1);
	EXPECT_NO_THROW(tree.addPolymer(p1));
	EXPECT_NO_THROW(tree.addPolymer(p2));
}

TEST_F(BinaryTreeTest, TreeGetPolymer)
{
	binTree tree(16);
	numberMonomers = 100;
	tree.addPolymer(p1);
	tree.addPolymer(p2);
	Polymer *p3 = new Polymer(0, 25);
	Polymer *p4 = new Polymer(0, 44);

	tree.addPolymer(p3);
	tree.addPolymer(p4);
	tree.addPolymer(p4);
	tree.addPolymer(p4);

	EXPECT_NE(tree.getPolymer(0), nullptr);
	EXPECT_NE(tree.getPolymer(1), nullptr);
	EXPECT_NE(tree.getPolymer(33), nullptr);
	EXPECT_NE(tree.getPolymer(66), nullptr);
	EXPECT_EQ(tree.getPolymer(77), nullptr);
	EXPECT_EQ(tree.getPolymer(120), nullptr);

	delete p3;
	delete p4;
}

TEST_F(BinaryTreeTest, TreeGetSumMonomers)
{
	binTree tree(5);
	tree.addPolymer(p1);
	tree.addPolymer(p2);
	EXPECT_EQ(tree.getSumMonomers(), 0);
}

TEST_F(BinaryTreeTest, TreeAnalyzeAllPolymers)
{
	binTree tree(5);
	EXPECT_EQ(tree.analyzeAllPolymers().size(), 0);
	tree.addPolymer(p1);
	tree.addPolymer(p2);
	EXPECT_EQ(tree.analyzeAllPolymers().size(), 2);
	tree.addPolymer(p1);
	tree.addPolymer(p1);
	EXPECT_EQ(tree.analyzeAllPolymers().size(), 4);
	auto vpolymers = tree.analyzeAllPolymers();
	EXPECT_NE(std::find(vpolymers.begin(), vpolymers.end(), p1), vpolymers.end());
	EXPECT_NE(std::find(vpolymers.begin(), vpolymers.end(), p2), vpolymers.end());
}

TEST_F(BinaryTreeTest, NodeConstruction)
{
	Node *node = new Node(nullptr, nullptr, nullptr);
	EXPECT_NE(node, nullptr);
	delete node;
}

TEST_F(BinaryTreeTest, NodeFather)
{
	Node parent(nullptr, nullptr, nullptr);
	Node node(nullptr, nullptr, nullptr);
	node.connectFather(&parent);
	EXPECT_EQ(&parent, node.getFather());
}

TEST_F(BinaryTreeTest, NodeChildren)
{
	Node parent(nullptr, nullptr, nullptr);
	Node child1(nullptr, nullptr, nullptr);
	Node child2(nullptr, nullptr, nullptr);
	parent.connectChilds(&child1, &child2);
	EXPECT_EQ(&child1, parent.getLeftChild());
	EXPECT_EQ(&child2, parent.getrightChild());
}

TEST_F(BinaryTreeTest, NodeMonomers)
{
	Node node(nullptr, nullptr, nullptr);
	EXPECT_EQ(node.getSumMonomers(), 0);
	node.addMonomers(25);
	EXPECT_EQ(node.getSumMonomers(), 25);
	node.delMonomers(19);
	EXPECT_EQ(node.getSumMonomers(), 6);
	node.delMonomers(7);
	EXPECT_EQ(node.getSumMonomers(), (uint64_t)-1);
}
