// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
 * mcPolymer.h
 *
 *  Created on: 24.08.2008
 */

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <stdint.h>
#include <math.h>
#include "Fragment.h"
#include "logging.h"

using namespace std;
namespace mcPolymer_ns {

// Only used during file import
Fragment::Fragment() : SCB(0) {
	monomerCount.resize(numberMonomers);
}

Fragment::Fragment(SpeciesID _id) {
	monomerCount.resize(numberMonomers, 0);
	monomerCount[_id] = 1;
	SCB = 0;
}

Fragment::Fragment(SpeciesID _id, uint32_t _length) {
	monomerCount.resize(numberMonomers, 0);
	monomerCount[_id] = _length;
	SCB = 0;
}
const valarray<uint32_t> Fragment::getMonomers() const {
	return monomerCount;
}
Fragment::Fragment(Fragment *_Fragment) {
	monomerCount = _Fragment->getMonomers();
	SCB = _Fragment->SCB;
}

// FIXME: Look for improvements
Fragment::Fragment(Fragment *_Fragment, double _randr) {
	// split
	monomerCount.resize(numberMonomers);
	for (size_t i = 0; i < numberMonomers; i++) {
		int nTr = lround(_Fragment->monomerCount[i] * _randr);
		monomerCount[i] = nTr;
		_Fragment->monomerCount[i] = _Fragment->monomerCount[i] - nTr;
	}
	// copy SCB
	int nTr = lround(_Fragment->SCB * _randr);
	SCB = static_cast<uint16_t>(nTr);
	_Fragment->SCB = _Fragment->SCB - static_cast<uint16_t>(nTr);
}

Fragment::~Fragment() {
	monomerCount.resize(0);
}

void Fragment::exportFragment(ofstream* _os) const {
	for (auto& elem : monomerCount) {
		_os->write((const char*)& elem, sizeof(uint32_t));
	}
	_os->write((const char*)&SCB, sizeof(uint16_t));
}

void Fragment::importFragment(ifstream* _is) {
	monomerCount.resize(numberMonomers);
	for (auto& elem : monomerCount) {
		_is->read((char*)& elem, sizeof(uint32_t));
	}
	_is->read((char *)&SCB, sizeof(uint16_t));
}

void Fragment::addM(SpeciesID _id) {
	++monomerCount[_id];
}

void Fragment::deleteM(SpeciesID _id) {
	// FIXME: handle overflow, logfile?
	if (monomerCount[_id])
		monomerCount[_id]--;
}

void Fragment::deleteM(SpeciesID _id, uint32_t _count) {
	if (monomerCount[_id] < _count) {
		string logString = "Fragment::deleteM() _id: " + to_string(_id) + " _count: " + to_string(_count) + " overflow";
		logError(logString);
	}
	monomerCount[_id] -= _count;
}

void Fragment::incrSCB() {
	++SCB;
}

void Fragment::deleteSCB(uint16_t _count) {
	// FIXME: handle overflow, logfile?
	SCB -= _count;
}

void Fragment::addSCB(uint16_t _count) {
	// FIXME: handle overflow, logfile?
	SCB += _count;
}

size_t Fragment::getLength() const {
	uint32_t cl = 0;
	for (const auto& elem : monomerCount) {
		cl += elem;
	}
	return cl;
}

int Fragment::getM(SpeciesID _id) const {
	return monomerCount[_id];
}

int Fragment::getSCB() const {
	return SCB;
}

} // namespace
