// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 * Plugin.h
 *
 *  Created on: 04.08.2017
 *
*/

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <math.h>
#include "Plugin.h"
#include "modelObjects.h"
//#include "Reactor.h"

using namespace std;
namespace mcPolymer_ns {

Plugin::Plugin(int _id, const string& _outputDir) {
	id = _id;
	outputDir = _outputDir;
}
int Plugin::getID() {
	return id;
}
string Plugin::getOutputDir() {
	return outputDir;
}
double calcSumPowij(const std::vector<int>& _cli, const std::vector<int>& _clj, double _as, double _al, double _ic) {
	double sumpowij = 0.0;
	for (size_t i = 0; i < _cli.size(); i++) {
		int cli = _cli.at(i);
		int clj = _clj.at(i);
		double powii;
		if (cli < _ic) {
			powii = pow(cli, (-1.0 * _as));
		}
		else {
			powii = pow(_ic, (-1.0 * _as + _al)) * pow(cli, (-1.0 * _al));
		}
		double powjj;
		if (clj < _ic) {
			powjj = pow(clj, (-1.0 * _as));
		}
		else {
			powjj = pow(_ic, (-1.0 * _as + _al)) * pow(clj, (-1.0 * _al));
		}
		sumpowij += (powii + powjj) / 2.0;
	}
	return sumpowij;
}

TerminationHamielec::TerminationHamielec(int _id, const string& _outputDir) :Plugin(_id, _outputDir) {
}
void TerminationHamielec::setCountStyrene0(uint64_t _count) {
	countStyrene0 = _count;
}
uint64_t TerminationHamielec::getCountStyrene0() {
	return countStyrene0;
}
double TerminationHamielec::getScaleFactor(double _temperature, double _conversion) {
	double temperatureKelvin = _temperature + 273.15;
	double a1 = parameter.a10 + parameter.a11 * temperatureKelvin;
	double a2 = parameter.a20 + parameter.a21 * temperatureKelvin;
	double a3 = parameter.a30 + parameter.a31 * temperatureKelvin;
	double f = exp(-2.0 * (a1 * _conversion + a2 * _conversion * _conversion + a3 * _conversion * _conversion * _conversion));
	return f;
}
TerminationChainLengthDependent::TerminationChainLengthDependent(int _id, const string& _outputDir) :Plugin(_id, _outputDir) {
}
void TerminationChainLengthDependent::setParameter(int _ic, double _as, double _al, double _as2, double _as3, double _as4, double _kt11p1, double _kt11p2, double _kt11p3) {
	// FIXME: check parameter initialization
	if (_ic != 0) {
		parameter.ic = (_ic * 1.0);
	}
	if (_as != 0) {
		parameter.as = _as;
	}
	if (_al != 0) {
		parameter.al = _al;
	}
	parameter.as2 = _as2;
	parameter.as3 = _as3;
	parameter.as4 = _as4;
	parameter.kt11p1 = _kt11p1;
	parameter.kt11p2 = _kt11p2;
	parameter.kt11p3 = _kt11p3;
	if ((parameter.as2 != 0.0) && (parameter.as3 != 0.0) && (parameter.as4 != 0.0) && (parameter.kt11p1 != 0.0) && (parameter.kt11p2 != 0.0) && (parameter.kt11p3 != 0.0)) {
		parameter.polymerContentDependent = true;
		parameter.kt11Offset = 1.0 - 1.0 / (1.0 + exp(parameter.kt11p2 / parameter.kt11p3));
	}
	else {
		parameter.polymerContentDependent = false;
	}
}

double TerminationChainLengthDependent::calcFactorCliClj(const vector<int>& _cli, const vector<int>& _clj) {
	return calcSumPowij(_cli, _clj, parameter.as, parameter.al, parameter.ic) / static_cast<double>(_cli.size()) * parameter.kt11Offset;
}
} // namespace
