// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 * class Polymer
 *
 *  Created on: 24.08.2008
 */

#include <iostream>
#include <fstream>
#include <string>
#include <stdint.h>
#include <valarray>
#include <vector>
#include "modelObjects.h"
#include "RandomBuffer.h"
#include "Histogram.h"
#include "Fragment.h"
#include "Connect.h"
#include "Polymer.h"
#include "logging.h"

using namespace std;

namespace mcPolymer_ns {

Polymer::Polymer() {
	selectPolymer = false;
}
const valarray<uint32_t> Polymer::getMonomers() {
	valarray<uint32_t> sumMonomerCount;
	sumMonomerCount.resize(numberMonomers);
	for (auto& elem : sumMonomerCount) {
		elem = 0;
	}
	for (size_t i = 0; i < polymerFragment.size(); i++) {
		valarray<uint32_t> monomerCountFragment = polymerFragment.at(i)->getMonomers();
		for (size_t j = 0; j < numberMonomers; j++) {
			sumMonomerCount[j] += monomerCountFragment[j];
		}
	}
	return sumMonomerCount;
}

Polymer::Polymer(int id) {
	Fragment* frag;
	frag = new Fragment(id);
	polymerFragment.push_back(frag);
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	propagatingRadical.MCRS = nullptr;
	propagatingRadical.activeFragment = polymerFragment.back();
	selectPolymer = false;
}
Polymer::Polymer(int id, int _length) {
	Fragment* frag;
	frag = new Fragment(id, _length);
	polymerFragment.push_back(frag);
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	propagatingRadical.MCRS = nullptr;
	propagatingRadical.activeFragment = polymerFragment.back();
	selectPolymer = false;
}
Polymer::Polymer(Fragment* _Fragment) {
	polymerFragment.push_back(_Fragment);
	selectPolymer = false;
}
uint32_t Polymer::getNumTransferMonomers() {
	return getM(idTransferMonomer);
}

Polymer::~Polymer() {
	for (size_t i = 0; i < polymerFragment.size(); i++) {
		delete polymerFragment.at(i);
	}
	for (size_t i = 0; i < fragmentConnects.size(); i++) {
		delete fragmentConnects.at(i);
	}
	polymerFragment.clear();
	fragmentConnects.clear();
}

// FIXME: returning int casted to unsigned, which is size_t is somehow stupid
int Polymer::getFragmentSize() {
	return unsigned(polymerFragment.size());
}

// FIXME: propagatingRadical not being exported
bool Polymer::exportPolymer(ofstream* _os) {

	// write number of fragments
	size_t fragmentSize = polymerFragment.size();
	_os->write((const char*)&fragmentSize, sizeof(size_t));
	// write fragments
	for (auto& elem : polymerFragment) {
		elem->exportFragment(_os);
	}
	// write number of connects
	size_t connectSize = fragmentConnects.size();
	_os->write((const char*)&connectSize, sizeof(size_t));
	// write connects
	for (auto& elem : fragmentConnects) {
		bool fragmentFound = false;
		uint32_t indexFragment1 = 0;
		uint32_t indexFragment2 = 0;
		Fragment* fragment1 = elem->getConnect1();
		Fragment* fragment2 = elem->getConnect2();
		// identify fragments --> index
		for (size_t i = 0; i < polymerFragment.size(); i++) {
			if (polymerFragment.at(i) == fragment1) {
				indexFragment1 = static_cast<uint32_t>(i);
				if (fragmentFound) {
					// all fragments identified
					break;
				}
				fragmentFound = true;
			}
			if (polymerFragment.at(i) == fragment2) {
				indexFragment2 = static_cast<uint32_t>(i);
				if (fragmentFound) {
					// all fragments identified
					break;
				}
				fragmentFound = true;
			}
		}
		// write fragments as index
		_os->write((const char*)&indexFragment1, sizeof(int32_t));
		_os->write((const char*)&indexFragment2, sizeof(int32_t));
		ConnectType cType = elem->getType();
		_os->write((const char*)&cType, sizeof(ConnectType));
	}
	return true;
}


bool Polymer::importPolymer(ifstream* _is) {

	// read number of fragments
	size_t fragmentSize;
	_is->read((char*)&fragmentSize, sizeof(size_t));
	// read fragments
	polymerFragment.resize(fragmentSize);
	for (size_t i = 0; i < fragmentSize; i++) {
		polymerFragment.at(i) = new Fragment();
		polymerFragment.at(i)->importFragment(_is);
	}

	// read number of connects
	size_t connectSize;
	_is->read((char*)&connectSize, sizeof(size_t));
	// read connects
	for (size_t i = 0; i < connectSize; i++) {
		uint32_t indexFragment1;
		uint32_t indexFragment2;
		ConnectType cType;
		_is->read((char*)&indexFragment1, sizeof(int32_t));
		_is->read((char*)&indexFragment2, sizeof(int32_t));
		_is->read((char*)&cType, sizeof(ConnectType));
		Connect* newConnect = new Connect(polymerFragment.at(indexFragment1), polymerFragment.at(indexFragment2), cType);
		fragmentConnects.push_back(newConnect);
	}
	return true;
}

void Polymer::select()
{
	selectPolymer = true;
}

void Polymer::deselect()
{
	selectPolymer = false;
}

bool Polymer::getSelection() {
	return selectPolymer;
}

void Polymer::addM1SCB(int _id) {
	propagatingRadical.activeFragment->addM(_id);
	propagatingRadical.activeFragment->incrSCB();
	propagatingRadical.MCRS->blockCurrentPosition();
	return;
}

int Polymer::getConnects() {
	return unsigned(fragmentConnects.size());
}
Connect* Polymer::getConnect(size_t _index) {
	return fragmentConnects.at(_index);
}
Connect* Polymer::deleteConnect() {
	Connect* cend;
	cend = fragmentConnects.back();
	fragmentConnects.pop_back();
	return cend;
}
void Polymer::deleteLastConnect() {
	delete fragmentConnects.back();
	fragmentConnects.pop_back();
}
Fragment* Polymer::getConnect(size_t _index, int _number) {
	if (_index >= fragmentConnects.size()) {
		return nullptr;
	}
	return fragmentConnects[_index]->getConnect(_number);
}
ConnectType Polymer::getConnectType(size_t _index) {
	if (_index >= fragmentConnects.size()) {
		return ConnectType::Undefined;
	}
	return fragmentConnects[_index]->getType();
}
bool Polymer::setConnectType(size_t _index, ConnectType _type) {
	if (_index >= fragmentConnects.size()) {
		return false;
	}
	fragmentConnects[_index]->setType(_type);
	return true;
}
void Polymer::addConnect(Fragment* _frg1, Fragment* _frg2, ConnectType _type) {
	fragmentConnects.push_back(new Connect(_frg1, _frg2, _type));
}
Polymer* Polymer::splitPolymerAtConnectType2() {
	Polymer* p2 = new Polymer();
	size_t connectType2Index = 0;
	size_t fragmentMacromonomerIndex = 0;

	// identify macromonomer
	for (size_t i = 0; i < fragmentConnects.size(); i++) {
		if (fragmentConnects.at(i) == propagatingRadical.activeConnect) {
			connectType2Index = i;
			break;
		}
	}
	for (size_t i = 0; i < polymerFragment.size(); i++) {
		if (fragmentConnects.at(connectType2Index)->getConnect2() == polymerFragment.at(i)) {
			// split --> potential ECR
			p2->activateECR(polymerFragment.at(i));
			fragmentMacromonomerIndex = i;
			break;
		}
	}
	// copy macromonomer
	size_t nFragments = polymerFragment.size();
	for (size_t i = fragmentMacromonomerIndex; i < nFragments; i++) {
		p2->addFragment(polymerFragment.back());
		polymerFragment.pop_back();
	}
	size_t nConnects = getConnects();
	for (size_t i = (connectType2Index + 1); i < nConnects; i++) {
		p2->addConnect(fragmentConnects.back());
		fragmentConnects.pop_back();
	}
	// remove Type 2 connect
	fragmentConnects.pop_back();
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	return p2;
}

void Polymer::addConnect(Connect* _c) {
	fragmentConnects.push_back(_c);
}

int Polymer::getM(int id) {
	int sum = 0;
	for (size_t i = 0; i < polymerFragment.size(); i++) {
		sum += polymerFragment.at(i)->getM(id);
	}
	return sum;
}

void Polymer::activateFragment(int _idS, int _idD) {
	propagatingRadical.activeFragment = polymerFragment.back();
	if (polymerFragment.size() > 1) {
		int sum = 0;
		for (size_t i = 0; i < polymerFragment.size(); i++) {
			sum += polymerFragment.at(i)->getM(_idS);
		}
		double randr = randBufferInstance->getRand();
		int tr1 = int(sum * randr);
		sum = 0;
		for (size_t i = 0; i < polymerFragment.size(); i++) {
			sum += polymerFragment.at(i)->getM(_idS);
			if (tr1 < sum) {
				propagatingRadical.activeFragment = polymerFragment.at(i);
				break;
			}
		}
	}
	propagatingRadical.radicalType = PropagatingRadicalType::MCRL;
	propagatingRadical.activeFragment->deleteM(_idS);
	propagatingRadical.activeFragment->addM(_idD);
}

Fragment* Polymer::getFragment(int ind) {
	return polymerFragment.at(ind);
}

Fragment* Polymer::getLastFragment() {
	return polymerFragment.back();
}

void Polymer::addFragment(Fragment* c1) {
	polymerFragment.push_back(c1);
}

Fragment* Polymer::deleteFragment() {
	Fragment* cend;
	cend = polymerFragment.back();
	polymerFragment.pop_back();
	return cend;
}

int Polymer::getLength() {
	int l = 0;
	vector<Fragment*>::iterator I;
	Fragment* cp;
	for (I = polymerFragment.begin(); I != polymerFragment.end(); ++I) {
		cp = *I;
		l += cp->getLength();
	}
	return l;
}

int Polymer::getSCB() {
	int SCB = 0;
	vector<Fragment*>::iterator I;
	Fragment* cp;
	for (I = polymerFragment.begin(); I != polymerFragment.end(); ++I) {
		cp = *I;
		SCB += cp->getSCB();
	}
	return SCB;
}

size_t Polymer::getNumLCB() {
	size_t cnt = 0;
	for (const auto& connect : fragmentConnects) {
		switch (connect->getType()) {
		case ConnectType::LCB:
			cnt++;
			continue;
		default:
			continue;
		}
	}
	return cnt;
}

void Polymer::activateECR() {
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	propagatingRadical.activeFragment = polymerFragment.back();
}
void Polymer::activateECR(Fragment* _activeFragment) {
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	propagatingRadical.activeFragment = _activeFragment;
}

void Polymer::addM(int _id) {
	switch (propagatingRadical.radicalType) {
		case PropagatingRadicalType::ECR:
			// ECR(n) --> ECR(n+1)
			propagatingRadical.activeFragment->addM(_id);
			return;
		case PropagatingRadicalType::MCR:
			// MCR(n)--> ECR(n + 1) + SCB
			propagatingRadical.activeFragment->addM(_id);
			propagatingRadical.radicalType = PropagatingRadicalType::ECR;
			if (propagatingRadical.MCRS->SCBPosition() == false) {
				propagatingRadical.activeFragment->incrSCB();
			}
			delete propagatingRadical.MCRS;
			propagatingRadical.MCRS = nullptr;
			return;
		case PropagatingRadicalType::MCRL:
			// MCRL(n) --> ECR(n+1) + connect
			// MCRL(n,macroradical) --> ECR(n+1) + connect

			polymerFragment.push_back(new Fragment(_id));
			fragmentConnects.push_back(new Connect(propagatingRadical.activeFragment, polymerFragment.back(), ConnectType::LCB));
			propagatingRadical.radicalType = PropagatingRadicalType::ECR;
			propagatingRadical.activeConnect = nullptr;
			propagatingRadical.activeFragment = polymerFragment.back();
			return;
		case PropagatingRadicalType::Undefined:
			string logString = "Polymer::addM(): PropagatingRadicalType::Undefined";
			logError(logString);;
			return;
	}
	string logString = "Polymer::addM(): Undefined";
	logError(logString);
}

void Polymer::addSCB() {
	propagatingRadical.activeFragment->incrSCB();
}

void Polymer::removeRadical() {
	propagatingRadical.activeConnect = nullptr;
	propagatingRadical.activeFragment = nullptr;
	if (propagatingRadical.MCRS != nullptr) {
		delete propagatingRadical.MCRS;
		propagatingRadical.MCRS = nullptr;
	}
	propagatingRadical.radicalType = PropagatingRadicalType::Undefined;
}
Fragment* Polymer::getPropagatingFragment() {
	return propagatingRadical.activeFragment;
}
PropagatingRadicalType Polymer::getPropagatingRadicalType() {
	return propagatingRadical.radicalType;
}
void Polymer::backbiting(uint32_t radicalPosition) {
	propagatingRadical.radicalType = PropagatingRadicalType::MCR;
	propagatingRadical.MCRS = new(environmentMCR);
	propagatingRadical.MCRS->setNewRadicalPosition(radicalPosition);
}
void Polymer::activateMCR(int _eductID, int _productID, PropagatingRadicalType _RadicalType) {
	propagatingRadical.radicalType = _RadicalType;
	if (polymerFragment.size() == 1) {
		propagatingRadical.activeFragment = polymerFragment.back();
	}
	else {
		uint64_t sumEductMonomers = 0;
		for (auto& elem : polymerFragment) {
			sumEductMonomers += elem->getM(_eductID);
		}
		double randr = randBufferInstance->getRand();
		uint64_t targetMonomer = static_cast<uint64_t>(sumEductMonomers * randr);
		sumEductMonomers = 0;
		for (auto& elem : polymerFragment) {
			sumEductMonomers += elem->getM(_eductID);
			if (targetMonomer <= sumEductMonomers) {
				propagatingRadical.activeFragment = elem;
				break;
			}
		}
	}
	propagatingRadical.activeFragment->deleteM(_eductID);
	propagatingRadical.activeFragment->addM(_productID);
	return;
}
void Polymer::randomizeRadicalPosition() {
	// FIXME: Implicitly checking for SCB.at() > 0
	if (propagatingRadical.MCRS->SCB.at(propagatingRadical.MCRS->radicalPosition)) {
		if (propagatingRadical.MCRS->radicalPosition == 2) {
			if (propagatingRadical.MCRS->SCB.at(2)) {
				// pos 3 occupied by SCB --> no migrition possible
				return;
			}
			// 2 --> 3
			propagatingRadical.MCRS->setNewRadicalPosition(3);
			return;
		}
		// check -->
		bool decrOK = false;
		if (!propagatingRadical.MCRS->SCB.at(propagatingRadical.MCRS->radicalPosition - 1)) {
			decrOK = true;
		}
		// check <--
		bool incrOK = false;
		uint32_t shortfragmentSize = propagatingRadical.MCRS->getShortFragmentSize();
		if ((propagatingRadical.activeFragment->getLength() - shortfragmentSize) > 1) {
			if (!propagatingRadical.MCRS->blockedPositionPlus()) {
				incrOK = true;
			}
		}
		if (incrOK && decrOK) {
			double randr = randBufferInstance->getRand();
			if (randr < 0.5) {
				propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition + 1);
			}
			else {
				propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition - 1);
			}
			return;
		}
		if (incrOK) {
			propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition + 1);
			return;
		}
		if (decrOK) {
			propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition - 1);
			return;
		}
		return;
	}
	else {
		if (propagatingRadical.MCRS->radicalPosition == 3) {
			// 3 --> 5
			propagatingRadical.MCRS->setNewRadicalPosition(5);
			return;
		}
		if (propagatingRadical.MCRS->radicalPosition == 2) {
			// 2 --> 4
			propagatingRadical.MCRS->setNewRadicalPosition(4);
			return;
		}
		if ((propagatingRadical.MCRS->radicalPosition + 2) >= propagatingRadical.activeFragment->getLength()) {
			// Y <-- X
			propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition - 2);
			return;
		}
		double randr = randBufferInstance->getRand();
		if (randr < 0.5) {
			// Y <-- X
			propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition - 2);
		}
		else {
			// X --> Y
			propagatingRadical.MCRS->setNewRadicalPosition(propagatingRadical.MCRS->radicalPosition + 2);
		}
	}
	return;

}

Polymer* Polymer::fragmentation() {
	// returns polymer formed by fragmentation reaction 
	Polymer* p = new Polymer();
	double randr = randBufferInstance->getRand();
	if (randr < 0.5) {
		p->addFragment(polymerFragment.back());
		polymerFragment.pop_back();
	}
	else {
		p->addFragment(polymerFragment.at(0));
		polymerFragment.at(0) = polymerFragment.back();
		polymerFragment.pop_back();
	}
	fragmentConnects.pop_back();
	propagatingRadical.activeFragment = polymerFragment.at(0);
	propagatingRadical.radicalType = PropagatingRadicalType::ECR;
	p->activateECR();
	return p;
}

size_t Polymer::getFragmentIndex(Fragment* _c) {
	for (size_t i = 0; i < polymerFragment.size(); i++) {
		if (polymerFragment.at(i) == _c) {
			return i;
		}
	}
	return -1;
}


} // namespace
