// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 *  Created on: 01.03.2021
 */

#include "MersenneTwister.h"
#include "RandomBuffer.h"
#include "logging.h"

#include <vector>

namespace mcPolymer_ns {

RandomBuffer* randBufferInstance;

RandomBuffer::RandomBuffer(uint16_t _sizeMBytes, uint16_t _partitioning)
    : partitioning(_partitioning) {
    if (_sizeMBytes == 0)
        _sizeMBytes = 1;
    buffer.resize(size_t(1024 * 1024 * _sizeMBytes) / sizeof(double), 0.0);
    for (unsigned i = 0; i < partitioning; i++) {
        buffer.at((buffer.size() / partitioning) * i) = REFILL;
    }
    nextPos = buffer.begin();
}

double RandomBuffer::getRand() {
    if (nextPos == buffer.end()) {
        size_t idxPartition = buffer.size() - (buffer.size() / partitioning);
        buffer.at(idxPartition) = REFILL;
        nextPos = buffer.begin();
    }

    if (*nextPos == REFILL) {
        MTRand rand;
        return rand.randExc();
    }
    if (*nextPos == READY) {
        if (nextPos != buffer.begin()) {
            size_t currentIdx = (nextPos - buffer.begin());
            buffer.at(currentIdx - (buffer.size() / partitioning)) = REFILL;
        }
        nextPos++;
    }
    return *nextPos++;
}

bool RandomBuffer::fillRands(uint32_t _oneseed) {
    if (buffer.at(0) == EXIT)
        return true;
    MTRand rand;
    if (_oneseed != UINT32_MAX)
        rand.seed(_oneseed);
    for (unsigned i = 0; i < partitioning; i++) {
        size_t idxPartition = (buffer.size() / partitioning) * i;
        if (buffer.at(idxPartition) == REFILL) {
            for (unsigned idx = idxPartition+1;
                idx < idxPartition + (buffer.size() / partitioning);
                idx++) {
                buffer[idx] = rand.randExc();
            }
            buffer.at(idxPartition) = READY;
        }
    }
    return false;
}


void RandomBuffer::setExit() {
    buffer.at(0) = EXIT;
}


} // namespace
