// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 *  Created on: 06.03.2021
 */
#include "RandomBuffer.h"
#include "Reaction.h"


namespace mcPolymer_ns {

Reaction::Reaction(size_t _numEducts, size_t _numProducts, const double *_kmc,
	               size_t _numMacroEducts, size_t _numMacroProducts)
	: kmc(_kmc) {
	eductCounter.resize(_numEducts);
	productCounter.resize(_numProducts);
	macroEduct.resize(_numMacroEducts);
	macroProduct.resize(_numMacroProducts);
}
Reaction::~Reaction() {}

Backbiting_RadPos3::Backbiting_RadPos3(const double* _kmc, std::vector <uint64_t*> _eductc, std::vector<uint64_t*> _productc,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1)
	: Reaction(_eductc.size(), _productc.size(), _kmc, 1, 1) {
	for (size_t i = 0; i < _eductc.size(); i++) {
		eductCounter[i] = _eductc[i];
	}
	for (size_t i = 0; i < _productc.size(); i++) {
		productCounter[i] = _productc[i];
	}
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
}
void Backbiting_RadPos3::execute() {
	for (size_t i = 0; i < eductCounter.size(); i++) {
		--* eductCounter[i];
	}
	for (size_t i = 0; i < productCounter.size(); i++) {
		++* productCounter[i];
	}
	Polymer* p = macroEduct[0]->deleteP();
	p->backbiting(3);
	macroProduct[0]->addPolymer(p);
}
double Backbiting_RadPos3::calculatePartialProbability() const {
	double Probability = *kmc * *eductCounter[0];
	for (size_t i = 1; i < eductCounter.size(); i++) {
		Probability *= *eductCounter[i];
	}
	return Probability;
}

ElementaryReaction::ElementaryReaction(const double* _kmc, std::vector<uint64_t*> _eductc, std::vector<uint64_t*> _productc)
	: Reaction(_eductc.size(), _productc.size(), _kmc) {
	for (size_t i = 0; i < _eductc.size(); i++) {
		eductCounter[i] = _eductc[i];
	}
	for (size_t i = 0; i < _productc.size(); i++) {
		productCounter[i] = _productc[i];
	}
}
void ElementaryReaction::execute() {
	for (size_t i = 0; i < eductCounter.size(); i++) {
		--* eductCounter[i];
	}
	for (size_t i = 0; i < productCounter.size(); i++) {
		++* productCounter[i];
	}
}
double ElementaryReaction::calculatePartialProbability() const {
	if (eductCounter.size() == 2) {
		if (eductCounter[0] == eductCounter[1]) {
			return *kmc * *eductCounter[0] * (*eductCounter[1] - 1) / 2.0;
		}
	}
	double Probability = *kmc * *eductCounter[0];
	for (size_t i = 1; i < eductCounter.size(); i++) {
		Probability *= *eductCounter[i];
	}
	return Probability;
}

Fragmentation::Fragmentation(const double* _kmc, uint64_t* _educt1c, uint64_t* _product1c, uint64_t* _product2c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2)
	: Reaction(1, 2, _kmc, 1, 2) {
	eductCounter[0] = _educt1c;
	productCounter[0] = _product1c;
	productCounter[1] = _product2c;
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
	macroProduct[1] = _macroProduct2;
}
void Fragmentation::execute() {
	Polymer* p = macroEduct[0]->deleteP();
	Polymer* macroRAFT = p->fragmentation();
	--* eductCounter[0];
	++* productCounter[0];
	++* productCounter[1];
	macroProduct[0]->addPolymer(p);
	macroProduct[1]->addPolymer(macroRAFT);
}
double Fragmentation::calculatePartialProbability() const {
	return *kmc * *eductCounter[0];
}

Initiation::Initiation(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
	polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter)
	: Reaction(2, 1, _kmc, 0, 1), monomerID(_monomerID), solidsContentMoleculeCounter(_solidsContentMoleculeCounter) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	macroProduct[0] = _macroProduct1;
}
void Initiation::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	macroProduct[0]->addPolymer(new Polymer(monomerID));
	++* solidsContentMoleculeCounter;
}
double Initiation::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * *eductCounter[1];
}

InitiatorDecomposition::InitiatorDecomposition(const double* _kmc, uint64_t* _educt1c, uint64_t* _product1c, double _initiatorEfficiency)
	: Reaction(1, 1, _kmc), initiatorEfficiency(_initiatorEfficiency) {
	eductCounter[0] = _educt1c;
	productCounter[0] = _product1c;
}
void InitiatorDecomposition::execute() {
	--*eductCounter[0];
	// Has to be processed twice
	for (unsigned i = 0; i < 2; i++)
		if (randBufferInstance->getRand() < initiatorEfficiency)
			++ *productCounter[0];
}
double InitiatorDecomposition::calculatePartialProbability() const {
	return *kmc * *eductCounter[0];
}

Propagation_SCB::Propagation_SCB(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter)
	: Reaction(2, 1, _kmc, 1, 1), monomerID(_monomerID), solidsContentMoleculeCounter(_solidsContentMoleculeCounter) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
}
void Propagation_SCB::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	++* solidsContentMoleculeCounter;
	Polymer* p = macroEduct[0]->deleteP();
	p->addM(monomerID);
	p->addSCB();
	macroProduct[0]->addPolymer(p);
}
double Propagation_SCB::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * *eductCounter[1];
}

Propagation_PnPnp1::Propagation_PnPnp1(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter)
	: Reaction(2, 1, _kmc, 1, 1), monomerID(_monomerID), solidsContentMoleculeCounter(_solidsContentMoleculeCounter) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
}
void Propagation_PnPnp1::execute() {
	--* eductCounter[1];
	++* solidsContentMoleculeCounter;
	macroProduct[0]->incrP(monomerID);
}
double Propagation_PnPnp1::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * *eductCounter[1];
}

Propagation_PnQnp1::Propagation_PnQnp1(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter)
	: Reaction(2, 1, _kmc, 1, 1), monomerID(_monomerID), solidsContentMoleculeCounter(_solidsContentMoleculeCounter) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
}
void Propagation_PnQnp1::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	++* solidsContentMoleculeCounter;
	Polymer* p = macroEduct[0]->deleteP();
	p->addM(monomerID);
	macroProduct[0]->addPolymer(p);
}
double Propagation_PnQnp1::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * *eductCounter[1];
}

Termination_PnPmDnDm::Termination_PnPmDnDm(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2)
	: Reaction(2, 2, _kmc, 2, 2) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	productCounter[1] = _product2c;
	macroEduct[0] = _macroEduct1;
	macroEduct[1] = _macroEduct2;
	macroProduct[0] = _macroProduct1;
	macroProduct[1] = _macroProduct2;
}
void Termination_PnPmDnDm::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	++* productCounter[1];
	macroProduct[0]->addPolymer(macroEduct[0]->deleteP());
	macroProduct[1]->addPolymer(macroEduct[1]->deleteP());
}
double Termination_PnPmDnDm::calculatePartialProbability() const {
	if (eductCounter[0] == eductCounter[1]) {
		return *kmc * *eductCounter[0] * (*eductCounter[1] - 1) / 2.0;
	}
	else {
		return *kmc * *eductCounter[0] * *eductCounter[1];
	}
}

Termination_PnPmDnpm::Termination_PnPmDnpm(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1)
	: Reaction(2, 1, _kmc, 2, 1) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	macroEduct[0] = _macroEduct1;
	macroEduct[1] = _macroEduct2;
	macroProduct[0] = _macroProduct1;
}
void Termination_PnPmDnpm::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	Polymer* p1 = macroEduct[0]->deleteP();
	Polymer* p2 = macroEduct[1]->deleteP();
	macroProduct[0]->addPolymerFromCombination(p1, p2);
	// FIXME ??
	// p1.addPolymer(p2);
	// macroProduct[0]->addPolymer(p1);
}
double Termination_PnPmDnpm::calculatePartialProbability() const {
	if (eductCounter[0] == eductCounter[1]) {
		return *kmc * *eductCounter[0] * (*eductCounter[1] - 1) / 2.0;
	}
	else {
		return *kmc * *eductCounter[0] * *eductCounter[1];
	}
}

ThermalInitiation3::ThermalInitiation3(const double* _kmc, uint64_t* _educt1c, uint64_t* _product1c,
	polymerSpecies* _macroProduct1, int _monomerID)
	: Reaction(1, 1, _kmc, 0, 1), monomerID(_monomerID) {
	eductCounter[0] = _educt1c;
	productCounter[0] = _product1c;
	macroProduct[0] = _macroProduct1;
}
double ThermalInitiation3::calculatePartialProbability() const {
	return *kmc * pow(*eductCounter[0], 3.0);
}
void ThermalInitiation3::execute() {
	--* eductCounter[0];
	--* eductCounter[0];
	--* eductCounter[0];
	++* productCounter[0];
	++* productCounter[0];
	macroProduct[0]->addPolymer(new Polymer(monomerID));
	macroProduct[0]->addPolymer(new Polymer(monomerID,2));
}

Transfer::Transfer(const double* _kmc, std::vector <uint64_t*> _eductc, std::vector<uint64_t*> _productc,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1)
	: Reaction(_eductc.size(), _productc.size(), _kmc, 1, 1) {
	for (size_t i = 0; i < _eductc.size(); i++) {
		eductCounter[i] = _eductc[i];
	}
	for (size_t i = 0; i < _productc.size(); i++) {
		productCounter[i] = _productc[i];
	}
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
}
void Transfer::execute() {
	for (size_t i = 0; i < eductCounter.size(); i++) {
		--* eductCounter[i];
	}
	for (size_t i = 0; i < productCounter.size(); i++) {
		++* productCounter[i];
	}
	macroProduct[0]->addPolymer(macroEduct[0]->deleteP());
}
double Transfer::calculatePartialProbability() const {
	double Probability = *kmc * *eductCounter[0];
	for (size_t i = 1; i < eductCounter.size(); i++) {
		Probability *= *eductCounter[i];
	}
	return Probability;
}

TransferToMonomer::TransferToMonomer(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2, int _monomerID, uint64_t* _solidsContentMoleculeCounter)
	: Reaction(2, 2, _kmc, 1, 2), monomerID(_monomerID), solidsContentMoleculeCounter(_solidsContentMoleculeCounter) {
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	productCounter[1] = _product2c;
	macroEduct[0] = _macroEduct1;
	macroProduct[0] = _macroProduct1;
	macroProduct[1] = _macroProduct2;
}
void TransferToMonomer::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	++* productCounter[0];
	++* productCounter[1];
	++* solidsContentMoleculeCounter;
	macroProduct[0]->addPolymer(macroEduct[0]->deleteP());
	macroProduct[1]->addPolymer(new Polymer(monomerID));
}
double TransferToMonomer::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * *eductCounter[1];
}

TransferToPolymer::TransferToPolymer(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
	polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2,
	int _monomerIDEduct, int _monomerIDProduct, uint64_t* _solidsContentMoleculeCounterEduct, uint64_t* _solidsContentMoleculeCounterProduct,
	PropagatingRadicalType _RadicalTypeMCRL)
	: Reaction(2, 2, _kmc, 2, 2), monomerIDEduct(_monomerIDEduct), monomerIDProduct(_monomerIDProduct),
		solidsContentMoleculeCounterEduct(_solidsContentMoleculeCounterEduct), solidsContentMoleculeCounterProduct(_solidsContentMoleculeCounterProduct),
		RadicalTypeMCRL(_RadicalTypeMCRL){
	eductCounter[0] = _educt1c;
	eductCounter[1] = _educt2c;
	productCounter[0] = _product1c;
	productCounter[1] = _product2c;
	macroEduct[0] = _macroEduct1;
	macroEduct[1] = _macroEduct2;
	macroProduct[0] = _macroProduct1;
	macroProduct[1] = _macroProduct2;
}
void TransferToPolymer::execute() {
	--* eductCounter[0];
	--* eductCounter[1];
	--* solidsContentMoleculeCounterEduct;
	++* productCounter[0];
	++* productCounter[1];
	++* solidsContentMoleculeCounterProduct;
	macroProduct[0]->addPolymer(macroEduct[0]->deleteP());
	Polymer* p = macroEduct[1]->deletePolymerForTransfer(monomerIDEduct);
	p->activateMCR(monomerIDEduct, monomerIDProduct, RadicalTypeMCRL);
	macroProduct[1]->addPolymer(p);
}
double TransferToPolymer::calculatePartialProbability() const {
	return *kmc * *eductCounter[0] * macroEduct[1]->getNumTransferMonomers();
}
} // namespace
