// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 *  Created on: 06.03.2021
 */

#pragma once

#include "polymerSpecies.h"

#include <string>
#include <valarray>
#include <math.h>

namespace mcPolymer_ns {

class Reaction {
public:
	Reaction(size_t _numEducts, size_t _numProducts, const double* _kmc, 
		     size_t _numMacroEducts = 0, size_t _numMacroProducts = 0);
	virtual ~Reaction();
	virtual double calculatePartialProbability() const = 0;
	virtual void execute() = 0;
protected:
	std::valarray<uint64_t*> eductCounter;
	std::valarray<uint64_t*> productCounter;
	std::valarray<polymerSpecies*> macroEduct;
	std::valarray<polymerSpecies*> macroProduct;
	const double* kmc;
};

class Backbiting_RadPos3 : public Reaction {
public:
	Backbiting_RadPos3(const double* _kmc, std::vector <uint64_t*> _eductc, std::vector<uint64_t*> _productc,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
};

class ElementaryReaction : public Reaction {
public:
	ElementaryReaction(const double* _kmc, std::vector <uint64_t*> _eductc, std::vector<uint64_t*> _productc);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
};

class Fragmentation : public Reaction {
public:
	Fragmentation(const double* _kmc, uint64_t* _educt1c, uint64_t* _product1c, uint64_t* _product2c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
};

class Initiation : public Reaction {
public:
	Initiation(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
				polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
	uint64_t* solidsContentMoleculeCounter;
};

class InitiatorDecomposition : public Reaction {
public:
	InitiatorDecomposition(const double* _kmc, uint64_t* _educt1c, uint64_t* _product1c, double _initiatorEfficiency);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const double initiatorEfficiency;
};

class Propagation_SCB : public Reaction {
public:
	Propagation_SCB(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
	uint64_t* solidsContentMoleculeCounter;
};

class Propagation_PnPnp1 : public Reaction {
public:
	Propagation_PnPnp1(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
	uint64_t* solidsContentMoleculeCounter;
};

class Propagation_PnQnp1 : public Reaction {
public:
	Propagation_PnQnp1(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, 
		               polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, int _monomerID, uint64_t* _solidsContentMoleculeCounter);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
	uint64_t* solidsContentMoleculeCounter;
};

class Termination_PnPmDnDm : public Reaction {
public:
	Termination_PnPmDnDm(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
};

class Termination_PnPmDnpm : public Reaction {
public:
	Termination_PnPmDnpm(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
};

class ThermalInitiation3 : public Reaction {
public:
	ThermalInitiation3(const double* _kmc, uint64_t* _educt1c,uint64_t* _product1c,
		polymerSpecies* _macroProduct1, int _monomerID);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
};

class Transfer : public Reaction {
public:
	Transfer(const double* _kmc, std::vector <uint64_t*> _eductc, std::vector<uint64_t*> _productc,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
};

class TransferToMonomer : public Reaction {
public:
	TransferToMonomer(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2, 
		int _monomerID, uint64_t* _solidsContentMoleculeCounter);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerID;
	uint64_t* solidsContentMoleculeCounter;
};

class TransferToPolymer : public Reaction {
public:
	TransferToPolymer(const double* _kmc, uint64_t* _educt1c, uint64_t* _educt2c, uint64_t* _product1c, uint64_t* _product2c,
		polymerSpecies* _macroEduct1, polymerSpecies* _macroEduct2, polymerSpecies* _macroProduct1, polymerSpecies* _macroProduct2, 
		int _monomerIDEduct, int _monomerIDProduct, uint64_t* _solidsContentMoleculeCounterEduct, uint64_t* _solidsContentMoleculeCounterProduct,
		PropagatingRadicalType _RadicalTypeMCRL);
	virtual double calculatePartialProbability() const override;
	virtual void execute() override;
private:
	const int monomerIDEduct;
	const int monomerIDProduct;
	uint64_t* solidsContentMoleculeCounterEduct;
	uint64_t* solidsContentMoleculeCounterProduct;
	PropagatingRadicalType RadicalTypeMCRL;
};
} // namespace
