// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* Created on: 13.02.2021
*/

#include "cmdlineOptions.h"
#include "logging.h"
#include "modelObjects.h"
#include "modelXML.h"
#include "filesXML.h"
#include "modelInterpreter.h"
#include "simulation.h"
#include "SimulationController.h"
#include "RandomBuffer.h"

#include <chrono>
#include <iostream>
#include <filesystem>
#include <fstream>
#include <string>
#include <time.h>
#include <thread>
#include <vector>

namespace mcPolymer_ns {

using namespace std;

void setupThreads(SimulationController &simController) {
	randBufferInstance = new RandomBuffer(256);
	std::thread simThread(startSimulation, std::ref(simController));
	std::thread randomThread(startRandomThread);
	simThread.join();
	randomThread.join();
}

void prepareSimulation(int argc, char* argv[]) {
	mcPolymerParameter simulationParameter = parseCmdlineOptions(argc, argv);
	simulationParameter.setOutputPrefix();
	auto simController = mcPolymer_ns::SimulationController(simulationParameter);
	setupThreads(simController);
}

void startSimulation(SimulationController &simController) {
	simController.runSimulation();
}

void startRandomThread() {
	for (;;) {
		if (randBufferInstance->fillRands())
			break;
		std::this_thread::sleep_for(std::chrono::milliseconds(10));
	}
}


SimulationController::SimulationController(const mcPolymerParameter &_simulationParameter)
	: simulationParameter(_simulationParameter) {
	initCurrentModel();
	kMC = new Simulation(&currentModel, simulationParameter.outputdir);
}

SimulationController::~SimulationController() {
	delete kMC;
}

void SimulationController::exportResults(double _reactionTime, bool _initial) const {
	ofstream concentrationfile;
	ofstream numMoleculesfile;
	if (_initial) {
		concentrationfile.open(simulationParameter.modelFile_export + ".Concentrations.dat", ofstream::out);
		numMoleculesfile.open(simulationParameter.modelFile_export + ".numMolecules.dat", ofstream::out);
		concentrationfile << "time\t";
		numMoleculesfile << "time\t";
		for (const auto& elem : currentModel.getSpecies()) {
			concentrationfile << elem.name << "\t";
			numMoleculesfile << elem.name << "\t";
		}
		concentrationfile << endl;
		numMoleculesfile << endl;
	} else {
		concentrationfile.open(simulationParameter.modelFile_export + ".Concentrations.dat", ofstream::out | ofstream::app);
		numMoleculesfile.open(simulationParameter.modelFile_export + ".numMolecules.dat", ofstream::out | ofstream::app);
	}
	concentrationfile << _reactionTime << "\t";
	numMoleculesfile << _reactionTime << "\t";
	for (const auto& elem : currentModel.getSpecies()) {
		concentrationfile << kMC->getMolecules(elem.id) * currentModel.getRatioConcMolecules() << "\t";
		numMoleculesfile << kMC->getMolecules(elem.id) << "\t";
	}
	concentrationfile << endl;
	numMoleculesfile << endl;
	concentrationfile.close();
	numMoleculesfile.close();
}

void SimulationController::exportComplete(time_t _timerStart) const {
	string filename = simulationParameter.outputdir + "/simulation.complete";
	ofstream logFile(filename, ofstream::out);
	time_t timerEnd = time(&timerEnd);
	logFile << "simulation time: " << difftime(timerEnd, _timerStart) << " s";
	logFile.close();
}

void SimulationController::exportPolymers(double _reactionTime) const {
		for (const auto& species : currentModel.getSpecies()) {
			if (species.isPolymer) {
				cout << "export macromolecules, species: " << species.name << endl;
				kMC->exportPolymerSpecies(species.id, simulationParameter.outputdir + "/" + species.name + "." + to_string(_reactionTime), FileTypePolymerSpecies::CLD, &currentModel);
				if (simulationParameter.exportPolymers) {
					kMC->exportPolymerSpecies(species.id, simulationParameter.outputdir + "/" + species.name + "." + to_string(_reactionTime), FileTypePolymerSpecies::Binary, &currentModel);
				}
			}
		}
		kMC->exportPolymers(simulationParameter.outputdir + "/" + "Polymers." + to_string(_reactionTime), FileTypePolymerSpecies::CLD, &currentModel);
		if (simulationParameter.exportPolymers) {
			kMC->exportPolymers(simulationParameter.outputdir + "/" + "Polymers." + to_string(_reactionTime), FileTypePolymerSpecies::Binary, &currentModel);
		}
}

void SimulationController::initCurrentModel() {
	modelInterface mcPolymerInterface;
	mcPolymerInterface.readModelFromFile(simulationParameter.modelFile_import);
	mcPolymerInterface.initInterface(uint64_t(simulationParameter.numMolecules), simulationParameter.temperature, simulationParameter.pressure);
	mcPolymerInterface.writeObjectsFile(simulationParameter.modelFile_import + ".xml");
	importModelFromXML(simulationParameter.modelFile_import + ".xml", &currentModel);
	currentModel.initModel(uint64_t(simulationParameter.numMolecules), simulationParameter.temperature, simulationParameter.pressure);
}

void SimulationController::runSimulation() {
	if (simulationParameter.reactionTime <= 0.0) {
		cout << "error: reaction time undefined" << endl;
		exit(1);
	}
	if (!std::filesystem::create_directories(std::filesystem::path(simulationParameter.outputdir))) {
		logError("Could not create output directory " + simulationParameter.outputdir);
	}
	exportResults(0.0, true);
	time_t timerStart = time(&timerStart);
	for (size_t i = 0; i < simulationParameter.timeline.size(); i++) {
		kMC->runTo(simulationParameter.timeline.at(i).time, &currentModel);
		if (simulationParameter.timeline.at(i).exportConcentrations) {
			exportResults(simulationParameter.timeline.at(i).time, false);
			cout << "time: " << simulationParameter.timeline.at(i).time << endl;
		}
		if (simulationParameter.timeline.at(i).exportCLD) {
			exportPolymers(simulationParameter.timeline.at(i).time);
		}
	}
	exportComplete(timerStart);
	randBufferInstance->setExit();
}

} // namespace
