// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
 * binaryTree.cpp
 *
 * Created on: 23.09.2020
 */

#include <iostream>
#include "binaryTree.h"
#include "Polymer.h"

using namespace std;
namespace mcPolymer_ns {

Node::Node(Node* _father, Node* _leftChild, Node* _rightChild) {
	father = _father;
	leftChild = _leftChild;
	rightChild = _rightChild;
}
lastLevelNode::lastLevelNode(Node* _father, Node* _leftChild, Node* _rightChild) : Node(_father, _leftChild, _rightChild) {
}
void Node::connectFather(Node* _father) {
	father = _father;
}
void Node::connectChilds(Node* _leftChild, Node* _rightChild) {
	leftChild = _leftChild;
	rightChild = _rightChild;
}
Node* Node::getFather() const {
	return father;
}
Node* Node::getLeftChild() const {
	return leftChild;
}
Node* Node::getrightChild() const {
	return rightChild;
}
uint64_t Node::getSumMonomers() const {
	return sumMonomers;
}
void lastLevelNode::addPolymer(Polymer* _polymer) {
	Polymers.push_back(_polymer);
}
uint64_t lastLevelNode::getNumPolymers() const {
	return Polymers.size();
}
Polymer* lastLevelNode::getPolymer(size_t _indexPolymer) {
	Polymer* polymer = Polymers.at(_indexPolymer);
	Polymers.at(_indexPolymer) = Polymers.back();
	Polymers.pop_back();
	return polymer;
}
void lastLevelNode::deletePolymer(Polymer *_polymer) {
	size_t idx;
	for (idx = 0; idx < Polymers.size(); idx++) {
		if (Polymers.at(idx) == _polymer) {
			break;
		}
	}
	Polymers.at(idx) = Polymers.back();
	Polymers.pop_back();
}
vector<Polymer*> lastLevelNode::analyzeAllPolymers() const {
	return Polymers;
}
void Node::addMonomers(uint64_t _monomers) {
	sumMonomers += _monomers;
}
void Node::delMonomers(uint64_t _monomers) {
	sumMonomers -= _monomers;
}
binTree::binTree(int _bits) {
	root = new Node(nullptr, nullptr, nullptr);
	allNodes.push_back(root);
	int bit = 0;
	while (true) {
		++bit;
		if (bit < _bits) {
			vector<Node*> NodeTmp;
			for (auto& currentNode : allNodes) {
				if (currentNode->getLeftChild() == nullptr) {
					Node* leftChild = new Node(nullptr, nullptr, nullptr);
					Node* rightChild = new Node(nullptr, nullptr, nullptr);
					leftChild->connectFather(currentNode);
					rightChild->connectFather(currentNode);
					currentNode->connectChilds(leftChild, rightChild);
					NodeTmp.push_back(leftChild);
					NodeTmp.push_back(rightChild);
				}
			}
			for (auto& currentNode : NodeTmp) {
				allNodes.push_back(currentNode);
			}
		}
		else {
			for (auto& currentNode : allNodes) {
				if (currentNode->getLeftChild() == nullptr) {
					lastLevelNode* leftChild = new lastLevelNode(nullptr, nullptr, nullptr);
					lastLevelNode* rightChild = new lastLevelNode(nullptr, nullptr, nullptr);
					leftChild->connectFather(currentNode);
					rightChild->connectFather(currentNode);
					currentNode->connectChilds(leftChild, rightChild);
					lastLevelNodes.push_back(leftChild);
					lastLevelNodes.push_back(rightChild);
				}
			}
		}
		if (bit == _bits) {
			break;
		}
	}
}
void binTree::addPolymer(Polymer* _polymer) {
	uint64_t cl = _polymer->getNumTransferMonomers();
	lastLevelNode* currentLastLevelNode = lastLevelNodes.at(cl);
	currentLastLevelNode->addPolymer(_polymer);
	currentLastLevelNode->addMonomers(cl);
	Node* currentNode = currentLastLevelNode->getFather();
	while (currentNode != nullptr) {
		currentNode->addMonomers(cl);
		currentNode = currentNode->getFather();
	}
}
Polymer* binTree::getPolymerLastLevel(uint64_t _targetPolymer) const {
	uint64_t analyzedPolymers = 0;
	lastLevelNode* targetedNode = lastLevelNodes.at(0);
	for (const auto& currentNode : lastLevelNodes) {
		if (currentNode->getNumPolymers() == 0) {
			continue;
		}
		if ((analyzedPolymers + currentNode->getNumPolymers()) >= _targetPolymer) {
			targetedNode = currentNode;
			break;
		}
		else {
			analyzedPolymers += currentNode->getNumPolymers();
		}
	}
	size_t numPolymers = targetedNode->getNumPolymers();
	size_t targetP = 0;
	Polymer* polymer = nullptr;
	while (true) {
		double randr = mtrandr();
		targetP = static_cast<size_t>(numPolymers * randr);
		if (targetP < numPolymers) {
			break;
		}
	}
	polymer = targetedNode->getPolymer(targetP);
	uint64_t cl = polymer->getNumTransferMonomers();
	Node* currentNode = targetedNode;
	currentNode->delMonomers(cl);
	currentNode = currentNode->getFather();
	while (currentNode != nullptr) {
		currentNode->delMonomers(cl);
		currentNode = currentNode->getFather();
	}
	return polymer;
}
Polymer* binTree::getPolymer(uint64_t _numMonomer) const {
	if (_numMonomer > allNodes.at(0)->getSumMonomers()) {
		return nullptr;
	}
	uint64_t numMonomer = _numMonomer;
	Node* currentNode = allNodes.at(0);
	Node* leftChild = allNodes.at(0)->getLeftChild();
	Node* rightChild = allNodes.at(0)->getrightChild();
	while (leftChild != nullptr) {
		if (numMonomer <= leftChild->getSumMonomers()) {
			currentNode = leftChild;
			leftChild = currentNode->getLeftChild();
			rightChild = currentNode->getrightChild();
		}
		else {
			numMonomer -= leftChild->getSumMonomers();
			currentNode = rightChild;
			leftChild = currentNode->getLeftChild();
			rightChild = currentNode->getrightChild();
		}
	}
	lastLevelNode* targetedNode = (static_cast<lastLevelNode*>(currentNode));
	size_t numPolymers = targetedNode->getNumPolymers();
	size_t targetP = 0;
	Polymer* polymer = nullptr;
	while (true) {
		double randr = mtrandr();
		targetP = static_cast<size_t>(numPolymers * randr);
		if (targetP < numPolymers) {
			break;
		}
	}
	polymer = targetedNode->getPolymer(targetP);
	uint64_t cl = polymer->getNumTransferMonomers();
	currentNode->delMonomers(cl);
	currentNode = currentNode->getFather();
	while (currentNode != nullptr) {
		currentNode->delMonomers(cl);
		currentNode = currentNode->getFather();
	}
	return polymer;
}

void binTree::deletePolymer(Polymer* _polymer) const {
	uint64_t cl = _polymer->getNumTransferMonomers();
	lastLevelNode* currentLastLevelNode = lastLevelNodes.at(cl);
	currentLastLevelNode->deletePolymer(_polymer);
	Node* currentNode = (static_cast<Node*>(currentLastLevelNode));
	currentNode->delMonomers(cl);
	currentNode = currentNode->getFather();
	while (currentNode != nullptr) {
		currentNode->delMonomers(cl);
		currentNode = currentNode->getFather();
	}
}
vector<Polymer*> binTree::analyzeAllPolymers() const {
	vector<Polymer*> polymers;
	for (const auto& currentNode : lastLevelNodes) {
		vector<Polymer*> polymersAtNode = currentNode->analyzeAllPolymers();
		if (polymersAtNode.size() > 0) {
			polymers.insert(end(polymers), begin(polymersAtNode), end(polymersAtNode));
		}
	}
	return polymers;
}
uint64_t binTree::getSumMonomers() const {
	return allNodes.at(0)->getSumMonomers();
}
void binTree::exportPolymers(ofstream* _os) const {
	for (const auto& currentNode : lastLevelNodes) {
		vector<Polymer*> polymersAtNode = currentNode->analyzeAllPolymers();
		if (polymersAtNode.size() > 0) {
			for (const auto polymer : polymersAtNode) {
				polymer->exportPolymer(_os);
			}
		}
	}
}
size_t binTree::getNumPolymers() const {
	size_t numPolymers = 0;
	for (const auto& currentNode : lastLevelNodes) {
		numPolymers += currentNode->getNumPolymers();
	}
	return numPolymers;
}
} // namespace
