// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
/*
 *  Created on: 13.02.2021
 */

#include <iostream>
#include <fstream>
#include "cxxopts.hpp"
#include "cmdlineOptions.h"
#include "parsing.h"

using namespace std;

static bool filenameValid(const string& _filename) {
	if (_filename.size() < 1) {
		cout << "filename invalid" << endl;
		return false;
	}
	ifstream is;
	is.open(_filename, ifstream::in);
	if (!is.is_open()) {
		cout << "error open file " << _filename << endl;
		return false;
	}
	is.get();
	if (!is.good()) {
		cout << "error reading file " << _filename << endl;
		return false;
	}
	return true;
}

static mcPolymerParameter getParameterFromModelFile(const string& _filename) {
	mcPolymerParameter simulationParameter;
	ifstream is(_filename, ios::in);
	if (!is.is_open())
		return simulationParameter;
	string line;
	cout << "model file: " << _filename << endl;
	while (getline(is, line)) {
		if (line.find("time", 0) != string::npos) {
			vector<string> specLine = parsing::splitline(line);
			if (specLine.size() < 2)
				continue;
			simulationParameter.reactionTime = stod(specLine.at(2));
			continue;
		}
		if (line.find("temperature", 0) != string::npos) {
			vector<string> specLine = parsing::splitline(line);
			if (specLine.size() < 2)
				continue;
			simulationParameter.ModelFileSetTemperature = true;
			simulationParameter.temperature = stod(specLine.at(2));
			continue;
		}
	}

	return simulationParameter;
}

mcPolymerParameter parseCmdlineOptions(int argc, char* argv[]) {
    if (argc == 1) {
		cout << "mcPolymer - Monte Carlo simulation program of radical polymerization, Version 4.0" << endl;
		cout << "Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl" << endl;
		cout << "Usage:" << endl;
		cout << "  mcPolymer - m modelFile [OPTION..]" << endl;
		cout << "  mcPolymer --help" << endl;
		exit(0);
	}
    mcPolymerParameter simulationParameter;
	try {
		cxxopts::Options options(argv[0], "Example:\n  mcPolymer -m Styrene.mcPolymer --time=3600 --dt=60 --temperature=70.0 --molecules=10000000000");
	    options.positional_help("[optional args]");

		options.add_options()
			("m, model", "Model", cxxopts::value<std::vector<std::string>>(), "modelFILE")
			("outputDir", "result directory", cxxopts::value<string>(), "outputDir")
			("help", "Print help")
			("license", "Print license")
			("time", "reaction time / s", cxxopts::value<double>(), "double")
			("dt", "time interval for output concentrations / s", cxxopts::value<double>(), "double")
			("dtCLD", "time interval for output cld | mmd / s", cxxopts::value<double>(), "double")
			("definedCLDTimes", "times for CLD export", cxxopts::value<std::vector<double>>(), "double, double..")
			("exportPolymers", "export all Polymers")
			("timeline", "time steps for output", cxxopts::value<string>(), "fileName")
			("temperature", "temperature / centigrade", cxxopts::value<double>(), "double")
			("pressure", "pressure / bar", cxxopts::value<double>(), "double")
			("molecules", "num molecules", cxxopts::value<uint64_t>(), "unsigned int")
			;

		auto result = options.parse(argc, argv);

		if (result.count("help")) {
			std::cout << options.help({ "", "Group" }) << std::endl;
			exit(0);
		}

		if (result.count("m")) {
			auto& fileVec = result["m"].as<std::vector<std::string>>();
			if (!filenameValid(fileVec.at(0))) {
				exit(1);
			}
			simulationParameter.modelFile_import = fileVec.at(0);
		}
		if (result.count("outputDir")) {
			simulationParameter.outputdir = result["outputDir"].as<string>();
		}
		if (result.count("time")) {
			double timeOpt = result["time"].as<double>();
			if (timeOpt < 0.0) {
				cout << "error: --time= " << timeOpt << " - not valid" << endl;
				exit(1);
			}
			simulationParameter.reactionTime = timeOpt;
			simulationParameter.CmdlSetReactionTime = true;
		}
		if (result.count("dt")) {
			double dt = result["dt"].as<double>();
			if (dt < 0.0) {
				cout << "error: --dt= " << dt << " - not valid" << endl;
				exit(1);
			}
			simulationParameter.dt = dt;
			simulationParameter.CmdlSetdt = true;
		}
		if (result.count("dtCLD")) {
			double dtCLD = result["dtCLD"].as<double>();
			if (dtCLD < 0.0) {
				cout << "error: --dtCLD= " << dtCLD << " - not valid" << endl;
				exit(1);
			}
			simulationParameter.dtCLD = dtCLD;
			simulationParameter.CmdlSetdtCLD = true;
		}
		if (result.count("definedCLDTimes"))
		{
			const auto values = result["definedCLDTimes"].as<std::vector<double>>();
			simulationParameter.CmdlSetdefinedCLDTimes = true;
			simulationParameter.definedCLDTimes.insert(simulationParameter.definedCLDTimes.begin(), values.begin(), values.end());
		}
		if (result.count("exportPolymers")) {
			simulationParameter.exportPolymers = true;
		}
		if (result.count("timeline")) {
			string filename = result["timeline"].as<string>();
			ifstream testOpenFile(filename);
			if (!testOpenFile) {
				cout << "error: --timeline, file " << filename << " not found" << endl;
				exit(1);
			}
			ifstream is(filename, ios::in);
			string line;
			while (getline(is, line)) {
				vector<string> specLine = parsing::splitline(line);
				if (specLine.size() > 0) {
					exportEvent eE;
					eE.exportCLD = true;
					eE.exportConcentrations = true;
					eE.time = stod(specLine.at(0));
					simulationParameter.timeline.push_back(eE);
				}
			}
			simulationParameter.reactionTime = simulationParameter.timeline.back().time;
			simulationParameter.CmdlSetReactionTime = true;
		}
		if (result.count("temperature")) {
			double temperature = result["temperature"].as<double>();
			if (temperature < -273.15) {
				cout << "error: --temperature= " << temperature << " - not valid" << endl;
				exit(1);
			}
			simulationParameter.temperature = temperature;
			simulationParameter.CmdlSetTemperature = true;
		}
		if (result.count("pressure")) {
			double pressure = result["pressure"].as<double>();
			if (pressure < 0.0) {
				cout << "error: --pressure= " << pressure << " - not valid" << endl;
				exit(1);
			}
			simulationParameter.pressure = pressure;
			simulationParameter.CmdlSetPressure = true;
		}
		if (result.count("molecules")) {
			simulationParameter.numMolecules = result["molecules"].as<uint64_t>();
			simulationParameter.CmdlSetnumMolecules = true;
		}
		if (result.count("license")) {
			cout << "mcPolymer - Monte Carlo simulation program of radical polymerization" << endl;
			cout << "Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl" << endl;
			cout << endl;
			cout << "This program is free software: you can redistribute it and/or modify" << endl;
			cout << "it under the terms of the GNU General Public License as published by" << endl;
			cout << "the Free Software Foundation, either version 3 of the License, or" << endl;
			cout << "(at your option) any later version." << endl;
			cout << endl;
			cout << "This program is distributed in the hope that it will be useful," << endl;
			cout << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << endl;
			cout << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << endl;
			cout << "GNU General Public License for more details." << endl;
			cout << endl;
			cout << "You should have received a copy of the GNU General Public License" << endl;
			cout << "along with this program.  If not, see <http://www.gnu.org/licenses/>." << endl;
			exit(0);
		}
	} catch (const cxxopts::OptionException& e) {
		std::cout << "error parsing options: " << e.what() << std::endl;
		exit(1);
	}

	auto simulationParameterFile = getParameterFromModelFile(simulationParameter.modelFile_import);
	if (simulationParameter.CmdlSetnumMolecules) {
		cout << "model dimension: " << simulationParameter.numMolecules << " ( " << scientific << double(simulationParameter.numMolecules) << " ) molecules" << endl;
		cout.unsetf(std::ios::scientific);
	} else {
		cout << "error: model dimension undefined (numMolecules = 0)" << endl;
		exit(1);
	}
	if (simulationParameter.CmdlSetReactionTime) {
		cout << "reaction time " << simulationParameter.reactionTime << " s" << endl;
	} else {
		if (simulationParameterFile.reactionTime > 0.0) {
			simulationParameter.reactionTime = simulationParameterFile.reactionTime;
			cout << "reaction time " << simulationParameter.reactionTime << ", defined in model file" << endl;
		}
	}
	if (simulationParameter.timeline.empty()) {
		if (simulationParameter.CmdlSetdt) {
			cout << "concentration output dt " << simulationParameter.dt << " s" << endl;
		} else {
			simulationParameter.dt = simulationParameter.reactionTime / 60.0;
			cout << "concentration output dt " << simulationParameter.dt << " s, set to 1/60 reaction time by default" << endl;
		}
		if (simulationParameter.CmdlSetdtCLD) {
			cout << "cld | mmd output dtCLD " << simulationParameter.dtCLD << " s" << endl;
		} else {
			simulationParameter.dtCLD = simulationParameter.reactionTime;
			cout << "cld | mmd output dtCLD " << simulationParameter.dtCLD << " s, set to reaction time by default" << endl;
		}
	} else {
		cout << simulationParameter.timeline.size() << " concentration and cld output times defined" << endl;
	}
	if (simulationParameter.CmdlSetTemperature) {
		cout << "reaction temperature " << simulationParameter.temperature;
		wcout << " degree Celsius" << endl;
	} else {
		if (simulationParameterFile.ModelFileSetTemperature) {
			simulationParameter.temperature = simulationParameterFile.temperature;
			cout << "temperature " << simulationParameter.temperature;
			cout << " degree Celsius, defined in model file" << endl;;
		} else {
			cout << "error: reaction temperature undefined" << endl;
			exit(1);
		}
	}
	if (simulationParameter.CmdlSetPressure) {
		cout << "pressure " << simulationParameter.pressure << " bar" << endl;
	} else {
		cout << "pressure " << simulationParameter.pressure << " bar, set by default" << endl;
	}
	if (simulationParameter.timeline.empty()) {
		double time = simulationParameter.dt;
		while (time < simulationParameter.reactionTime) {
			exportEvent eE;
			eE.exportCLD = false;
			eE.exportConcentrations = true;
			eE.time = time;
			simulationParameter.timeline.push_back(eE);
			time += simulationParameter.dt;
		}
		exportEvent eE;
		eE.exportCLD = false;
		eE.exportConcentrations = true;
		eE.time = simulationParameter.reactionTime;
		simulationParameter.timeline.push_back(eE);
	}
	if (simulationParameter.CmdlSetdefinedCLDTimes) {
		if (simulationParameter.definedCLDTimes.back() < simulationParameter.reactionTime) {
			simulationParameter.definedCLDTimes.push_back(simulationParameter.reactionTime);
		}
	}
	if (simulationParameter.CmdlSetdtCLD || simulationParameter.CmdlSetdefinedCLDTimes) {
		vector <exportEvent> timelineCopy;
		double timeCLDExport;
		size_t indexCLD = 0;
		if (simulationParameter.CmdlSetdefinedCLDTimes) {
			timeCLDExport = simulationParameter.definedCLDTimes.at(indexCLD);
		}
		else {
			timeCLDExport = simulationParameter.dtCLD;
		}
		for (size_t i = 0; i < simulationParameter.timeline.size(); i++) {
			if (timeCLDExport == simulationParameter.timeline.at(i).time) {
				simulationParameter.timeline.at(i).exportCLD = true;
				timelineCopy.push_back(simulationParameter.timeline.at(i));
				if (simulationParameter.CmdlSetdefinedCLDTimes) {
					++indexCLD;
					if (indexCLD < simulationParameter.definedCLDTimes.size()) {
						timeCLDExport = simulationParameter.definedCLDTimes.at(indexCLD);
					}
					else {
						break;
					}
				}
				else {
					timeCLDExport += simulationParameter.dtCLD;
					if (timeCLDExport > simulationParameter.reactionTime) {
						break;
					}
				}
			}
			else {
				if (simulationParameter.timeline.at(i).time < timeCLDExport) {
					timelineCopy.push_back(simulationParameter.timeline.at(i));
				}
				else {
					if (simulationParameter.timeline.at(i).time > timeCLDExport) {
						exportEvent eE;
						eE.exportCLD = true;
						eE.exportConcentrations = false;
						eE.time = timeCLDExport;
						timelineCopy.push_back(eE);
						timelineCopy.push_back(simulationParameter.timeline.at(i));
						if (simulationParameter.CmdlSetdefinedCLDTimes) {
							++indexCLD;
							if (indexCLD < simulationParameter.definedCLDTimes.size()) {
								timeCLDExport = simulationParameter.definedCLDTimes.at(indexCLD);
							}
							else {
								break;
							}
						}
						else {
							timeCLDExport += simulationParameter.dtCLD;
							if (timeCLDExport > simulationParameter.reactionTime) {
								break;
							}
						}
					}
				}
			}
		}
		simulationParameter.timeline.clear();
		simulationParameter.timeline.insert(simulationParameter.timeline.begin(), timelineCopy.begin(), timelineCopy.end());
	}
    return simulationParameter;
}
