// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* filesXML.cpp
*
*  Created on: 16.06.2021
*/
#include <string>
#include <fstream>
#include <iostream>
#include "parsing.h"
#include "filesXML.h"
#include "modelObjects.h"
#include "modelXML.h"
using namespace std;
namespace mcPolymer_ns {

static vector<speciesObject> parseSpecies(ifstream* _is) {
	vector<speciesObject> speciesObjectList;
	speciesObject speciesObj;
	string line;
	while (getline(*_is, line)) {
		vector<string> specLine = parsing::splitline(line);
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Species") {
				speciesObjectList.push_back(speciesObj);
				speciesObject speciesObjNew;
				speciesObj = speciesObjNew;
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Specieslist") {
				break;
			}
		}
		if (specLine.size() > 4) {
			if (specLine.at(1) == "ID") {
				speciesObj.id = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "NAME") {
				speciesObj.name = specLine.at(3);
				continue;
			}
			if (specLine.at(1) == "TYPE") {
				if (specLine.at(3) == "monomer") {
					speciesObj.isMonomer = true;
					continue;
				}
				if (specLine.at(3) == "polymer") {
					speciesObj.isPolymer = true;
					continue;
				}
				if (specLine.at(3) == "initiator") {
					speciesObj.isInitiator = true;
					continue;
				}
			}
			if (specLine.at(1) == "CONCENTRATION") {
				speciesObj.concentration = stod(specLine.at(3));
			}
			if (specLine.at(1) == "MOLARMASS") {
				speciesObj.optionalParameterMolarMass = stod(specLine.at(3));
			}
			if (specLine.at(1) == "INITIATOREFFICIENCY") {
				speciesObj.optionalParameterF = stod(specLine.at(3));
			}
		}
	}
	return speciesObjectList;
}

static vector<coefficientObject> parseCoefficients(ifstream* _is) {
	coefficientObject coefficientObj;
	vector<coefficientObject> coefficientObjectList;
	string line;
	while (getline(*_is, line)) {
		vector<string> specLine = parsing::splitline(line);
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Coefficient") {
				coefficientObjectList.push_back(coefficientObj);
				coefficientObject coefficientObjNew;
				coefficientObj = coefficientObjNew;
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Coefficientlist") {
				break;
			}
		}
		if (specLine.size() > 4) {
			if (specLine.at(1) == "ID") {
				coefficientObj.id = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "NAME") {
				coefficientObj.name = specLine.at(3);
				continue;
			}
			if (specLine.at(1) == "A0") {
				coefficientObj.A0 = stod(specLine.at(3));
			}
			if (specLine.at(1) == "EA") {
				coefficientObj.Ea = stod(specLine.at(3));
				coefficientObj.isTemperatureDependent = true;
			}
			if (specLine.at(1) == "CONSTANTVALUE") {
				coefficientObj.value = stod(specLine.at(3));
			}
		}
	}
	return coefficientObjectList;
}

static vector<reactionObject> parseReactions(ifstream* _is) {
	reactionObject reactionObj;
	vector<reactionObject> reactionObjectList;
	string line;
	while (getline(*_is, line)) {
		vector<string> specLine = parsing::splitline(line);
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Reaction") {
				reactionObjectList.push_back(reactionObj);
				reactionObject reactionObjNew;
				reactionObj = reactionObjNew;
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Reactionlist") {
				break;
			}
		}
		if (specLine.size() > 4) {
			if (specLine.at(1) == "ID") {
				reactionObj.id = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "TEMPLATE") {
				reactionObj.rktTemplate = StringToReactionTemplate(specLine.at(3));;
				continue;
			}
			if (specLine.at(1) == "COEFFICIENTID") {
				reactionObj.coefficientID = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "EDUCTID") {
				int eductID = stoi(specLine.at(3));
				reactionObj.eductIdList.push_back(eductID);
				continue;
			}
			if (specLine.at(1) == "PRODUCTID") {
				int productID = stoi(specLine.at(3));
				reactionObj.productIDList.push_back(productID);
				continue;
			}
		}
	}
	return reactionObjectList;
}

static vector<pluginObject> parsePlugin(ifstream* _is) {
	pluginObject pluginObj;
	vector<pluginObject> pluginObjectList;
	string line;
	while (getline(*_is, line)) {
		vector<string> specLine = parsing::splitline(line);
		if (specLine.size() > 2) {
			if (specLine.at(1) == "/Plugin") {
				pluginObjectList.push_back(pluginObj);
				pluginObject pluginObjNew;
				pluginObj = pluginObjNew;
			}
		}
		if (specLine.size() > 4) {
			if (specLine.at(1) == "ID") {
				pluginObj.id = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "TYPE") {
				pluginObj.type = StringToPluginType(specLine.at(3));
				pluginObj.templateLabel = PluginTypeToString(pluginObj.type);
				pluginObj.templateLabelFilename = pluginObj.templateLabel;
				continue;
			}
			if (specLine.at(1) == "TEMPLATE") {
				continue;
			}
			if (specLine.at(1) == "FILENAME") {
				pluginObj.filename = specLine.at(3);
				continue;
			}
			if (specLine.at(1) == "COEFFICIENTID") {
				pluginObj.coefficientID = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "SPECIESID") {
				int speciesID = stoi(specLine.at(3));
				pluginObj.speciesIdList.push_back(speciesID);
				continue;
			}
			if (specLine.at(1) == "IC") {
				pluginObj.chainLengthDependenceParameter.ic = stoi(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "AS") {
				pluginObj.chainLengthDependenceParameter.as = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "AL") {
				pluginObj.chainLengthDependenceParameter.al = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "AS2") {
				pluginObj.chainLengthDependenceParameter.as2 = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "AS3") {
				pluginObj.chainLengthDependenceParameter.as3 = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "AS4") {
				pluginObj.chainLengthDependenceParameter.as4 = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "KT11P1") {
				pluginObj.chainLengthDependenceParameter.kt11p1 = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "KT11P2") {
				pluginObj.chainLengthDependenceParameter.kt11p2 = stod(specLine.at(3));
				continue;
			}
			if (specLine.at(1) == "KT11P3") {
				pluginObj.chainLengthDependenceParameter.kt11p3 = stod(specLine.at(3));
				continue;
			}
		}
	}
	return pluginObjectList;
}

void importModelFromXML(const string &_filename, modelXML* _mcPolymerModel) {
	ifstream testOpenFile(_filename);
	if (!testOpenFile)
	{
		cout << "error: model file " << _filename << " not found" << endl;
		exit(1);
	}
	ifstream is(_filename, ios::in);
	string line;
	vector <speciesObject> species;
	vector <coefficientObject> coefficients;
	vector <reactionObject> reactions;
	vector <pluginObject> plugins;

	while (getline(is, line)) {
		vector<string> specLine = parsing::splitline(line);
		if (specLine.size() > 2) {
			if (specLine.at(1) == "Species") {
				species = parseSpecies(&is);
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "Coefficient") {
				coefficients = parseCoefficients(&is);
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "Reaction") {
				reactions = parseReactions(&is);
			}
		}
		if (specLine.size() > 2) {
			if (specLine.at(1) == "Plugin") {
				plugins = parsePlugin(&is);
			}
		}
	}
	_mcPolymerModel->constructModel(species, coefficients, reactions, plugins);
}

void exportModelToXML(const string &_filename, modelXML* _mcPolymerModel) {
	ofstream outfile(_filename);
	string outstr = "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>";
	outfile << outstr << endl;
	outstr = "<!DOCTYPE mcPolymer-kineticModel[";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Specieslist (Species)*>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Species (ID, NAME, TYPE? , CONCENTRATION? , MOLARMASS? , INITIATOREFFICIENCY)>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Coefficientlist (Coefficient)*>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Coefficient (ID, NAME, CONSTANTVALUE? , A0? , EA? )>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Reactionlist (Reaction)*>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Reaction (ID, TEMPLATE, COEFFICIENTID, EDUCTID*, PRODUCTID*)>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Pluginlist (Plugin)*>";
	outfile << outstr << endl;
	outstr = "\t<!ELEMENT Plugin (ID, TYPE, TEMPLATE, FILENAME?, COEFFICIENTID?, SPECIESID?, KT11A0?, KT11EA?, IC?, AS?, AL?, AS2?, AS3?, AS4?, KT11P1?, KT11P2?, KT11P3?)>";
	outfile << outstr << endl;
	outstr = "\t] >";
	outfile << outstr << endl;
	outstr = "<mcPolymer-kineticModel>";
	outfile << outstr << endl;
	outfile << "<Specieslist>" << endl;
	vector<speciesObject> speciesObjectList = _mcPolymerModel->getSpecies();
	for (auto& elem : speciesObjectList) {
		outstr = "\t<Species>";
		outfile << outstr << endl;
		outstr = "\t\t<ID>" + to_string(elem.id) + "</ID>";
		outfile << outstr << endl;
		outstr = "\t\t<NAME>" + elem.name + "</NAME>";
		outfile << outstr << endl;
		if (elem.concentration > 0.0) {
			outstr = "\t\t<CONCENTRATION>" + to_string(elem.concentration) + "</CONCENTRATION>";
			outfile << outstr << endl;
		}
		if (elem.isMonomer) {
			outstr = "\t\t<TYPE>monomer</TYPE>";
			outfile << outstr << endl;
			outstr = "\t\t<MOLARMASS>" + to_string(elem.optionalParameterMolarMass) + "</MOLARMASS>";
			outfile << outstr << endl;
			outfile << "\t</Species>" << endl;
			continue;
		}
		if (elem.isInitiator) {
			outstr = "\t\t<TYPE>initiator</TYPE>";
			outfile << outstr << endl;
			outstr = "\t\t<INITIATOREFFICIENCY>" + to_string(elem.optionalParameterF) + "</INITIATOREFFICIENCY>";
			outfile << outstr << endl;
			outfile << "\t</Species>" << endl;
			continue;
		}
		if (elem.isPolymer) {
			outstr = "\t\t<TYPE>polymer</TYPE>";
			outfile << outstr << endl;
			outfile << "\t</Species>" << endl;
			continue;
		}
		outfile << "\t</Species>" << endl;
	}
	outstr = "</Specieslist>";
	outfile << outstr << endl;
	outfile << "<Coefficientlist>" << endl;
	vector<coefficientObject> coefficientObjectList = _mcPolymerModel->calculateAndGetCoefficients();
	for (auto& elem : coefficientObjectList) {
		outstr = "\t<Coefficient>";
		outfile << outstr << endl;
		outstr = "\t\t<ID>" + to_string(elem.id) + "</ID>";
		outfile << outstr << endl;
		outstr = "\t\t<NAME>" + elem.name + "</NAME>";
		outfile << outstr << endl;
		if ((elem.A0 > 0.0) && (elem.Ea > 0.0)) {
			outstr = "\t\t<A0>" + to_string(elem.A0) + "</A0>";
			outfile << outstr << endl;
			outstr = "\t\t<EA>" + to_string(elem.Ea) + "</EA>";
			outfile << outstr << endl;
		}
		else {
			outstr = "\t\t<CONSTANTVALUE>" + to_string(elem.value) + "</CONSTANTVALUE>";
			outfile << outstr << endl;
		}
		outstr = "\t</Coefficient>";
		outfile << outstr << endl;
	}
	outfile << "</Coefficientlist>" << endl;
	outfile << "<Reactionlist>" << endl;
	vector<reactionObject> reactionObjectList = _mcPolymerModel->getReactions();
	for (auto& elem : reactionObjectList) {
		outstr = "\t<Reaction>";
		outfile << outstr << endl;
		outstr = "\t\t<ID>" + to_string(elem.id) + "</ID>";
		outfile << outstr << endl;
		outstr = "\t\t<TEMPLATE>" + ReactionTemplateToString(elem.rktTemplate) + "</TEMPLATE>";
		outfile << outstr << endl;
		outstr = "\t\t<COEFFICIENTID>" + to_string(elem.coefficientID) + "</COEFFICIENTID>";
		outfile << outstr << endl;
		for (auto& elemReactants : elem.eductIdList) {
			outstr = "\t\t<EDUCTID>" + to_string(elemReactants) + "</EDUCTID>";
			outfile << outstr << endl;
		}
		for (auto& elemReactants : elem.productIDList) {
			outstr = "\t\t<PRODUCTID>" + to_string(elemReactants) + "</PRODUCTID>";
			outfile << outstr << endl;
		}
		outfile << "\t</Reaction>" << endl;
	}
	outfile << "</Reactionlist>" << endl;
	outfile << "<Pluginlist>" << endl;
	vector<pluginObject> pluginObjectList = _mcPolymerModel->getPlugins();
	for (auto& elem : pluginObjectList) {
		outstr = "\t<Plugin>";
		outfile << outstr << endl;
		outstr = "\t\t<ID>" + to_string(elem.id) + "</ID>";
		outfile << outstr << endl;
		outstr = "\t\t<TYPE>" + PluginTypeToString(elem.type) + "</TYPE>";

		outfile << outstr << endl;
		outstr = "\t\t<TEMPLATE>" + elem.templateLabel + "</TEMPLATE>";
		outfile << outstr << endl;
		if (elem.filename != "") {
			outstr = "\t\t<FILENAME>" + elem.filename + "</FILENAME>";
			outfile << outstr << endl;
		}
		if (elem.coefficientID != -1) {
			outstr = "\t\t<COEFFICIENTID>" + to_string(elem.coefficientID) + "</COEFFICIENTID>";
			outfile << outstr << endl;
		}
		if (elem.speciesIdList.size() > 0) {
			for (auto& elemEductID : elem.speciesIdList) {
				outstr = "\t\t<SPECIESID>" + to_string(elemEductID) + "</SPECIESID>";
				outfile << outstr << endl;
			}
		}
		if (elem.chainLengthDependenceParameter.ic != 0.0) {
			outstr = "\t\t<IC>" + to_string(elem.chainLengthDependenceParameter.ic) + "</IC>";
			outfile << outstr << endl;
		}
		if (elem.chainLengthDependenceParameter.as != 0.0) {
			outfile << scientific;
			outfile << "\t\t<AS>" << elem.chainLengthDependenceParameter.as << "</AS>" << endl;
		}
		if (elem.chainLengthDependenceParameter.al != 0.0) {
			outfile << scientific;
			outfile << "\t\t<AL>" << elem.chainLengthDependenceParameter.al << "</AL>" << endl;
		}
		if (elem.chainLengthDependenceParameter.as2 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<AS2>" << elem.chainLengthDependenceParameter.as2 << "</AS2>" << endl;
		}
		if (elem.chainLengthDependenceParameter.as3 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<AS3>" << elem.chainLengthDependenceParameter.as3 << "</AS3>" << endl;
		}
		if (elem.chainLengthDependenceParameter.as4 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<AS4>" << elem.chainLengthDependenceParameter.as4 << "</AS4>" << endl;
		}
		if (elem.chainLengthDependenceParameter.kt11p1 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<KT11P1>" << elem.chainLengthDependenceParameter.kt11p1 << "</KT11P1>" << endl;
		}
		if (elem.chainLengthDependenceParameter.kt11p2 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<KT11P2>" << elem.chainLengthDependenceParameter.kt11p2 << "</KT11P2>" << endl;
		}
		if (elem.chainLengthDependenceParameter.kt11p3 != 0.0) {
			outfile << scientific;
			outfile << "\t\t<KT11P3>" << elem.chainLengthDependenceParameter.kt11p3 << "</KT11P3>" << endl;
		}
		outfile << "\t</Plugin>" << endl;
	}

	outfile << "</Pluginlist>" << endl;
	outstr = "</mcPolymer-kineticModel>";
	outfile << outstr << endl;
	return;
}

} // namespace