// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* modelInterpreter.cpp
*
* Created on: 07.06.2017
*/
#include "modelObjects.h"
#include "modelXML.h"
#include "filesXML.h"
#include "modelInterpreter.h"
#include <iostream>
#include <fstream>
#include <math.h>


using namespace std;
namespace mcPolymer_ns {

modelInterface::modelInterface() {
	initialized = false;
mcPolymerXML = new modelXML();
}
vector<modelInterface::species> modelInterface::importSpeciesFromModelFile(const string& _filename) {
	vector<species> specieslistRaw;
	ifstream inFile;
	inFile.open(_filename);
	if (inFile.bad()) {
		cout << "model file\t" << _filename << " not found" << endl;
		exit(1);
	}
	string line;
	while (getline(inFile, line)) {
		if (line.empty() == true) {
			continue;
		}
		if ((line.at(0) == '/') && (line.at(1) == '/')) {
			continue;
		}
		if (line.at(0) == '#') {
			continue;
		}
		if (line.at(line.size() - 1) == '\r') {
			// Remove \r (Carriage Return) from line
			line.erase(line.size() - 1);
		}
		std::vector<std::string> splitstring = splitline(line);
		if (splitstring.size() == 0) {
			continue;
		}
		if (splitstring.size() >= 6) {
			if ((splitstring.at(0) == "M") && (splitstring.at(1) == "(") && (splitstring.at(3) == ")") && (splitstring.at(4) == "=")) {
				string monomerName = splitstring.at(2);
				bool monomerInList = false;
				for (auto& species : specieslistRaw) {
					if (species.name == monomerName) {
						monomerInList = true;
						break;
					}
				}
				if (!monomerInList) {
					species newMonomer;
					newMonomer.monomer = true;
					newMonomer.name = monomerName;
					newMonomer.molarMass = stod(splitstring.at(5));
					specieslistRaw.push_back(newMonomer);
				}
			}
		}
		// two prerequisites: a) "--" and b) "-->"
		size_t pos1 = 0;
		size_t pos2 = 0;
		for (size_t i = 0; i < splitstring.size(); i++) {
			if (splitstring.at(i).compare("--") == 0) {
				pos1 = i;
			}
			if (splitstring.at(i).compare("-->") == 0) {
				pos2 = i;
			}
		}
		if ((pos1 == 0) || (pos2 == 0)) {
			continue;
		}
		if ((pos2 - pos1) != 2) {
			cout << "model file: " << _filename << " line: " << line << endl;
			cout << "synatx error: no coefficient defined" << endl;
			exit(1);
		}
		if ((pos1 == 0) || (pos2 == splitstring.size() - 1)) {
			cout << "model file: " << _filename << " line: " << line << endl;
			cout << "synatx error: no educts / products defined" << endl;
			exit(1);
		}
		bool macroProperties = false;
		for (size_t i = 0; i < splitstring.size(); i++) {
			if ((i >= pos1) && (i <= pos2)) {
				continue;
			}
			if (splitstring.at(i) == "+") {
				continue;
			}
			if (splitstring.at(i) == "(") {
				macroProperties = true;
				continue;
			}
			if (splitstring.at(i) == ")") {
				macroProperties = false;
				continue;
			}
			if (macroProperties) {
				continue;
			}
			string speciesName = splitstring.at(i);
			bool speciesInList = false;
			for (auto& species : specieslistRaw) {
				if (species.name == speciesName) {
					speciesInList = true;
					break;
				}
			}
			if (!speciesInList) {
				species newSpecies;
				newSpecies.name = speciesName;
				if ((i + 1) < splitstring.size()) {
					if (splitstring.at(i + 1) == "(") {
						newSpecies.macromolecular = true;
					}
				}
				specieslistRaw.push_back(newSpecies);
			}
		}
	}
	vector<species> specieslistMonomersFirst;
	for (auto& species : specieslistRaw) {
		if (species.monomer) {
			specieslistMonomersFirst.push_back(species);
		}
	}
	for (auto& species : specieslistRaw) {
		if (!species.monomer) {
			specieslistMonomersFirst.push_back(species);
		}
	}
	return specieslistMonomersFirst;
}
bool modelInterface::readModelFile(const string& _filename) {
	string line;
	vector<string> splitstring;
	bool interpretationfinished;
	int linenumber;
	size_t oldposition;
	ifstream inFile;
	inFile.open(_filename);
	if (inFile.bad()) {
		cout << "model file\t" << _filename << " not found" << endl;
		exit(1);
	}
	linenumber = 0;

	while (getline(inFile, line)) {
		++linenumber;
		if (line.empty() == true) {
			continue;
		}
		if ((line.at(0) == '/') && (line.at(1) == '/')) {
			continue;
		}
		if (line.at(0) == '#') {
			continue;
		}
		if (line.at(line.size() - 1) == '\r') {
			// Remove \r (Carriage Return) from line
			line.erase(line.size() - 1);
		}
		interpretationfinished = false;
		splitstring.clear();
		std::vector<std::string> splitstring = splitline(line);
		if (splitstring.size() == 0) {
			continue;
		}
		
		interpretationfinished = detectPlugin(splitstring, pluginlist);
		if (interpretationfinished) {
			continue;
		}
		interpretationfinished = detectCoeficients(splitstring, coefficientlist, shadowlist);
		if (interpretationfinished) {
			continue;
		}
		interpretationfinished = detectReaction(splitstring, reactionlist, specieslist, coefficientlist);
		if (interpretationfinished) {
			continue;
		}
		interpretationfinished = detectConcentration(splitstring, specieslist);
		if (interpretationfinished) {
			continue;
		}
		interpretationfinished = detectInitiatorEfficiency(splitstring, specieslist);
		if (interpretationfinished) {
			continue;
		}
	}
	inFile.close();
	// allocation of shadow coefficients, if used in any reaction
	for (size_t i = 0; i < coefficientlist.size(); i++) {
		allocateCoefficients(&coefficientlist.at(i), shadowlist);
	}
	// detect species type
	// delete obsolete species from species list
	for (size_t i = 0; i < specieslist.size(); i++) {
		specieslist.at(i).usedFlag = false;
		for (size_t j = 0; j < reactionlist.size(); j++) {
			for (size_t k = 0; k < reactionlist.at(j).indexEducts.size(); k++) {
				if (reactionlist.at(j).indexEducts.at(k) == i) {
					specieslist.at(i).usedFlag = true;
					break;
				}
			}
			if (!specieslist.at(i).usedFlag) {
				for (size_t k = 0; k < reactionlist.at(j).indexProducts.size(); k++) {
					if (reactionlist.at(j).indexProducts.at(k) == i) {
						specieslist.at(i).usedFlag = true;
						break;
					}
				}
			}
		}
	}
	// do not delete monomers because macromolecular species refer to them
	for (size_t i = 0; i < specieslist.size(); i++) {
		if (specieslist.at(i).monomer == true) {
			for (size_t j = 0; j < specieslist.size(); j++) {
				// check only active, macromolecular species
				if ((specieslist.at(j).usedFlag) && (specieslist.at(j).macromolecular)) {
					if (specieslist.at(i).name == specieslist.at(j).assignedMonomer) {
						specieslist.at(i).usedFlag = true;
					}
				}
			}
		}
	}
	for (size_t i = 0; i < specieslist.size(); i++) {
		if (!specieslist.at(i).usedFlag) {
			// delete species
			if (i != (specieslist.size() - 1)) {
				while (1) {
					// delete obsolete species at the end
					if (!specieslist.back().usedFlag) {
						specieslist.pop_back();
					}
					else {
						break;
					}
				}
				// first, copy last species at free position
				specieslist.at(i).name = specieslist.back().name;
				specieslist.at(i).macromolecular = specieslist.back().macromolecular;
				specieslist.at(i).initiator = specieslist.back().initiator;
				specieslist.at(i).monomer = specieslist.back().monomer;
				specieslist.at(i).molarMass = specieslist.back().molarMass;
				specieslist.at(i).concentration = specieslist.back().concentration;
				specieslist.at(i).filename = specieslist.back().filename;
				specieslist.at(i).assignedMonomer = specieslist.back().assignedMonomer;
				specieslist.at(i).f = specieslist.back().f;
				specieslist.at(i).molecules = specieslist.back().molecules;
				oldposition = specieslist.size() - 1;
				for (size_t j = 0; j < reactionlist.size(); j++) {
					for (size_t k = 0; k < reactionlist.at(j).indexEducts.size(); k++) {
						if (reactionlist.at(j).indexEducts.at(k) == oldposition) {
							// set index to new position
							reactionlist.at(j).indexEducts.at(k) = i;
						}
					}
					for (size_t k = 0; k < reactionlist.at(j).indexProducts.size(); k++) {
						if (reactionlist.at(j).indexProducts.at(k) == oldposition) {
							// set index to new position
							reactionlist.at(j).indexProducts.at(k) = i;
						}
					}
				}
			}
			specieslist.pop_back();
		}
	}

	for (size_t i = 0; i < specieslist.size(); i++) {
		specieslist.at(i).indexSpecies = i;
	}
	for (size_t i = 0; i < reactionlist.size(); i++) {
		for (size_t j = 0; j < reactionlist.at(i).indexEducts.size(); j++) {
			for (size_t k = 0; k < specieslist.size(); k++) {
				if (reactionlist.at(i).indexEducts.at(j) == specieslist.at(k).indexSpecies) {
					reactionlist.at(i).indexEducts.at(j) = k;
				}
			}
		}
		for (size_t j = 0; j < reactionlist.at(i).indexProducts.size(); j++) {
			for (size_t k = 0; k < specieslist.size(); k++) {
				if (reactionlist.at(i).indexProducts.at(j) == specieslist.at(k).indexSpecies) {
					reactionlist.at(i).indexProducts.at(j) = k;
				}
			}
		}
	}
	for (size_t i = 0; i < reactionlist.size(); i++) {
		typeReaction(&reactionlist.at(i), specieslist);
	}
	for (size_t i = 0; i < pluginlist.size(); i++) {
		for (size_t j = 0; j < specieslist.size(); j++) {
			if ((size_t)pluginlist.at(i).speciesIndex == specieslist.at(j).indexSpecies) {
				pluginlist.at(i).speciesIndex = static_cast<int>(j);
			}
		}
	}
	for (size_t i = 0; i < pluginlist.size(); i++) {
		typePlugin(&pluginlist.at(i), specieslist, coefficientlist);
	}
	return true;
}

modelInterface::~modelInterface() {

}
bool modelInterface::initInterface(double _controlVolumeL, double _temperatureCelsius, double _pressureBar) {
	initialized = false;
	if (_controlVolumeL <= 0.0) {
		cout << "control volume " << _controlVolumeL << " L too low" << endl;
		return false;
	}
	if (_temperatureCelsius < -273.15) {
		cout << "temperature " << _temperatureCelsius << " �C too low" << endl;
		return false;
	}
	if (_pressureBar < 0.0) {
		cout << "pressure " << _pressureBar << " bar too low" << endl;
		return false;
	}
	// determine number of molecules
	double csum = 0;
	for (size_t i = 0; i < specieslist.size(); i++) {
		csum += specieslist.at(i).concentration;
	}
	uint64_t molecules = llround(_controlVolumeL * AvogadroConstantNA * csum);
	initialized = true;
	return initInterface(molecules, _temperatureCelsius, _pressureBar);
}
bool modelInterface::initInterface(uint64_t _numMolecules, double _temperatureCelsius, double _pressureBar) {
	initialized = false;
	if (_numMolecules == 0) {
		cout << "number of molecules is zero" << endl;
	}
	if (_temperatureCelsius < -273.15) {
		cout << "temperature " << _temperatureCelsius << " �C too low" << endl;
	}
	if (_pressureBar < 0.0) {
		cout << "pressure " << _pressureBar << " bar too low" << endl;
	}

	// set initial number of molecules
	double csum = 0.0;
	for (size_t i = 0; i < specieslist.size(); i++) {
		csum += specieslist.at(i).concentration;
	}
	if (csum > 0.0) {
		for (size_t i = 0; i < specieslist.size(); i++) {
			specieslist.at(i).molecules = llround(specieslist.at(i).concentration / csum * _numMolecules);
		}
	}
	else {
		for (size_t i = 0; i < specieslist.size(); i++) {
			specieslist.at(i).molecules = 0;
		}
	}
	controlVolume = static_cast <double>(_numMolecules) / AvogadroConstantNA  / csum;
	molecules = _numMolecules;
	for (size_t i = 0; i < reactionlist.size(); i++) {
		if (coefficientlist.at(reactionlist.at(i).indexk).value == 0.0) {
			if (coefficientlist.at(reactionlist.at(i).indexk).Va == 0.0) {
				coefficientlist.at(reactionlist.at(i).indexk).value = coefficientlist.at(reactionlist.at(i).indexk).A0 * exp(-1.0 * coefficientlist.at(reactionlist.at(i).indexk).Ea * 1000.0 / (_temperatureCelsius + 273.15) / 8.314);
			}
			else {
				coefficientlist.at(reactionlist.at(i).indexk).value = coefficientlist.at(reactionlist.at(i).indexk).A0 * exp(-1.0 * coefficientlist.at(reactionlist.at(i).indexk).Ea * 1000.0 / (_temperatureCelsius + 273.15) / 8.314) * exp(-1.0 * coefficientlist.at(reactionlist.at(i).indexk).Va / 10.0 * _pressureBar / (_temperatureCelsius + 273.15) / 8.314);
			}
		}
		int rorder = 0;

		rorder = static_cast <int>(reactionlist.at(i).indexEducts.size());
		if (reactionlist.at(i).rktTemplate == ReactionTemplate::TransferToPolymer) {
			rorder = 2;
		}
		if (rorder == 2) {
			if (reactionlist.at(i).indexEducts.at(0) == reactionlist.at(i).indexEducts.at(1)) {
				coefficientlist.at(reactionlist.at(i).indexk).valueMC = coefficientlist.at(reactionlist.at(i).indexk).value * csum / (_numMolecules * 1.0) * 2.0;
			}
			else {
				coefficientlist.at(reactionlist.at(i).indexk).valueMC = coefficientlist.at(reactionlist.at(i).indexk).value * csum / (_numMolecules * 1.0);
			}
			coefficientlist.at(reactionlist.at(i).indexk).reactionOrder = rorder;
			continue;
		}
		if (rorder == 1) {
			coefficientlist.at(reactionlist.at(i).indexk).valueMC = coefficientlist.at(reactionlist.at(i).indexk).value;
		}
		if (rorder == 3) {
			coefficientlist.at(reactionlist.at(i).indexk).valueMC = coefficientlist.at(reactionlist.at(i).indexk).value * csum / (_numMolecules * 1.0) * csum / (_numMolecules * 1.0);
		}
		coefficientlist.at(reactionlist.at(i).indexk).reactionOrder = rorder;
	}
	initialized = true;
	return true;
}
bool modelInterface::filenameValid(const string& _filename) const {
	if (_filename.size() < 1) {
		cout << "filename invalid" << endl;
		return false;
	}
	ifstream is;
	is.open(_filename, ifstream::in);
	if (!is.is_open()) {
		cout << "error open file " << _filename << endl;
		return false;
	}

	if (!is.good()) {
		cout << "error reading file " << _filename << endl;
		return false;
	}
	return true;
}
bool modelInterface::readModelFromFile(const string& _filename) {
	if (!filenameValid(_filename))
		return false;
	specieslist = importSpeciesFromModelFile(_filename);
	if (!readModelFile(_filename)) {
		return false;
	}
	return true;
}


bool modelInterface::isInitialized() const {
	return initialized;
}

std::vector<std::string>  modelInterface:: splitline(std::string& _line) {
	vector<std::string> splitstring;
	size_t pos, pos1, pos2;
	int start;
	string clearstring, substr, linecopy;
	bool sepmode;
	bool st1, st2;
	linecopy = _line;

	// trim string --> detect reaction definition
	// two prerequisites: a) "--" and b) "-->"
	st1 = false;
	st2 = false;
	pos1 = linecopy.find("--");
	if (pos1 != string::npos) {
		st1 = true;
		pos2 = linecopy.find("-->");
		if (pos2 != string::npos) {
			st2 = true;
			if (pos1 < pos2) {
				if ((pos2 + 3) < linecopy.length()) {
					linecopy.insert(pos2 + 3, " ");
				}
				linecopy.insert(pos2, " ");
				linecopy.insert(pos1 + 2, " ");
				linecopy.insert(pos1, " ");
			}
		}
	}
	// separate +,(,),[,],=
	for (size_t i = 0; i < linecopy.length(); i++) {
		if ((linecopy.at(i) == '+') || (linecopy.at(i) == '(') || (linecopy.at(i) == ')') || (linecopy.at(i) == '{') || (linecopy.at(i) == '}') || (linecopy.at(i) == '[') || (linecopy.at(i) == ']') || (linecopy.at(i) == '=') || (linecopy.at(i) == '/') || (linecopy.at(i) == '*') || (linecopy.at(i) == ',')) {
			// seperate "+" only in reaction strings
			if ((linecopy.at(i) == '+') && (!st1 && !st2)) {
				continue;
			}
			else {
				if (i < (linecopy.length() - 1)) {
					linecopy.insert((uint64_t)i + 1, " ");
				}
				else {
					linecopy.append(" ");
				}

				linecopy.insert(i, " ");
				++i;
				++i;
			}
		}
	}
	// replace separators
	pos = linecopy.find_first_of(" \t:;");
	while (pos != std::string::npos)
	{
		linecopy[pos] = ' ';
		pos = linecopy.find_first_of(" \t:;", pos + 1);
	}
	start = 0;
	// ignore leading separators
	if (linecopy.at(0) == ' ') {
		for (size_t i = 0; i < linecopy.length(); i++) {
			if (linecopy.at(i) == ' ') {
				++start;
			}
			else {
				break;
			}
		}
	}
	// delete duplicate seperators
	sepmode = false;
	substr = "";
	for (size_t i = start; i < linecopy.length(); i++) {
		if (linecopy.at(i) == ' ') {
			if (!sepmode) {
				sepmode = true;
				//close splitstring
				splitstring.push_back(substr);
				substr = "";
			}
		}
		else {
			substr.append(&linecopy.at(i), 1);
			sepmode = false;
		}
	}
	if (!sepmode) {
		if (substr != "") {
			splitstring.push_back(substr);
		}
	}
	if (splitstring.at(0).compare("#") == 0) {
		splitstring.resize(0);
		return splitstring;
	}
	if (splitstring.size() >= 2) {
		if ((splitstring.at(0).compare("/") == 0) && (splitstring.at(1).compare("/") == 0)) {
			splitstring.resize(0);
			return splitstring;
		}
	}
	return splitstring;
}


int modelInterface::findSpecies(vector<species>& _species, string& _searchstr) {
	for (size_t i = 0; i < _species.size(); i++) {
		if (_species.at(i).name.compare(_searchstr) == 0) {
			return static_cast<int>(i);
		}
	}
	return -1;
}
int modelInterface::findCoefficient(vector<coefficient>& _coefficient, string& _searchstr) {
	for (size_t i = 0; i < _coefficient.size(); i++) {
		if (_coefficient.at(i).name.compare(_searchstr) == 0) {
			return static_cast<int>(i);
		}
	}
	return -1;
}
bool modelInterface::detectCoeficients(vector<string>& _splitstring, vector<coefficient>& _coefficient, vector<coefficient>& _shadow) {
	int posCoefficient, pos1, flagsRT;
	coefficient newCoefficient;
	bool knownCoefficient, countinueSearch;
	double activationVolume, activationEnergy, A0;

	countinueSearch = false;
	for (size_t i = 0; i < _splitstring.size(); i++) {
		if (_splitstring.at(i).compare("Plugin") == 0) {
			// syntax does not fit
			return false;
		}
		if (_splitstring.at(i).compare("=") == 0) {
			if (i == 0) {
				// syntax does not fit
				return false;
			}
			if (_splitstring.at((uint64_t)i - 1).compare(")") == 0) {
				// syntax does not fit
				return false;
			}
			if (_splitstring.at((uint64_t)i - 1).compare("]") == 0) {
				// syntax does not fit
				return false;
			}

			countinueSearch = true;
			pos1 = static_cast<int>(i) - 1;
			knownCoefficient = false;
			posCoefficient = findCoefficient(_coefficient, _splitstring.at(pos1));
			if (posCoefficient == -1) {
				//coefficient unknown
				newCoefficient.name.append(_splitstring.at(pos1));
				newCoefficient.value = 0.0;
				newCoefficient.A0 = 0.0;
				newCoefficient.Ea = 0.0;
				newCoefficient.Va = 0.0;
				newCoefficient.valueMC = 0.0;

			}
			else {
				// coefficient found
				knownCoefficient = true;
			}
			break;
		}
	}
	if (!countinueSearch) {
		// = not found
		return false;
	}
	activationVolume = 0.0;
	activationEnergy = 0.0;
	flagsRT = 0;
	for (size_t i = 5; i < (_splitstring.size() - 1); i++) {
		if (_splitstring.at(i).compare("RT") == 0) {
			if (_splitstring.at((uint64_t)i + 1).compare(")") == 0) {
				if (_splitstring.at((uint64_t)i - 1).compare("/") == 0) {
					if (_splitstring.at((uint64_t)i - 2).compare("p") == 0) {
						// exp(value*p/RT)
						activationVolume = stod(_splitstring.at((uint64_t)i - 4)) * -1.0;
						++flagsRT;
					}
					else {
						// exp(value/RT)
						activationEnergy = stod(_splitstring.at((uint64_t)i - 2)) * -1.0;
						++flagsRT;
					}
				}
			}
		}
	}
	if (flagsRT > 0) {
		A0 = stod(_splitstring.at((uint64_t)pos1 + 2));
		if (knownCoefficient) {
			_coefficient.at(posCoefficient).A0 = A0;
			_coefficient.at(posCoefficient).Ea = activationEnergy;
			_coefficient.at(posCoefficient).Va = activationVolume;
			_coefficient.at(posCoefficient).value = 0.0;
		}
		else {
			newCoefficient.A0 = A0;
			newCoefficient.Ea = activationEnergy;
			newCoefficient.Va = activationVolume;
			newCoefficient.value = 0.0;
			_shadow.push_back(newCoefficient);
		}
	}
	else {
		if (knownCoefficient) {
			_coefficient.at(posCoefficient).value = stod(_splitstring.at((uint64_t)pos1 + 2));
		}
		else {

			newCoefficient.value = stod(_splitstring.at((uint64_t)pos1 + 2));
			_shadow.push_back(newCoefficient);
		}
	}
	return true;
}

bool modelInterface::detectConcentration(vector<string>& _splitstring, vector<species>& _species) {
	int ofs;
	species newSpecies;
	string filename;
	string assignedMonomer;
	ofs = 0;
	filename = "";
	assignedMonomer = "";
	// two prerequisites: a) "[" and b) "]"
	for (size_t i = 0; i < _splitstring.size(); i++) {
		if (_splitstring.at(i).compare("[") == 0) {
			if (i < (_splitstring.size() + 3)) {
				if (_splitstring.at((uint64_t)i + 2).compare(",") == 0) {
					filename = _splitstring.at((uint64_t)i + 3);
					assignedMonomer = _splitstring.at((uint64_t)i + 5);
					ofs = 4;
				}
			}
			if (i < (_splitstring.size() + 4 + ofs)) {
				if (_splitstring.at((uint64_t)i + 2 + ofs).compare("]") == 0) {
					if (_splitstring.at((uint64_t)i + 3 + ofs).compare("=") == 0) {
						// syntax ok: [XX] = value
						int j = findSpecies(_species, _splitstring.at((uint64_t)i + 1));
						if (j != -1) {
							//species already exist, set concentration
							_species.at(j).concentration = stod(_splitstring.at((uint64_t)i + 4 + ofs));
							_species.at(j).filename = filename;
							_species.at(j).assignedMonomer = assignedMonomer;
							return true;
						}
						newSpecies.name.append(_splitstring.at((uint64_t)i + 1));
						newSpecies.concentration = stod(_splitstring.at((uint64_t)i + 4 + ofs));
						newSpecies.filename = filename;
						newSpecies.assignedMonomer = assignedMonomer;
						newSpecies.macromolecular = false;
						newSpecies.initiator = false;
						newSpecies.monomer = false;
						newSpecies.molarMass = 0.0;
						newSpecies.f = 0.0;
						newSpecies.molecules = 0;
						_species.push_back(newSpecies);
						return true;
					}
				}
			}
		}
	}
	return false;
}
bool modelInterface::detectPlugin(vector<string>& _splitstring, vector<plugin>& _plugin) {
	plugin newPlugin;
	newPlugin.type = PluginType::Undefined;
	// three prerequisites: a) Plugin , b) name and c) "{"
	// check
	string pluginName;
	if (_splitstring.size() > 3) {
		if (_splitstring.at(0).compare("Plugin") == 0) {
			if (_splitstring.at(2).compare("{") == 0) {
				pluginName = _splitstring.at(1);
				if (pluginName.compare("TerminationHamielec") == 0) {
					newPlugin.type = PluginType::TerminationHamielec;
				}
				if (pluginName.compare("TerminationBuback") == 0) {
					newPlugin.type = PluginType::Undefined;
				}
				if (pluginName.compare("TerminationChainLengthDependent") == 0) {
					newPlugin.type = PluginType::TerminationChainLengthDependent;
				}
			}
			else {
				return false;
			}
		}
		else {
			return false;
		}
	}
	else {
		return false;
	}
	// syntax ok: Plugin name {

	for (size_t i = 3; i < _splitstring.size(); i++) {
		// label (position i) = value (position i + 2)
		// check, if possible
		if ((i + 2) >= _splitstring.size()) {
			break;
		}
		switch (newPlugin.type) {
		case PluginType::TerminationHamielec: {
			if (_splitstring.at(i).compare("speciesStyrene") == 0) {
				newPlugin.speciesName = _splitstring.at((uint64_t)i + 2);
			}
			if (_splitstring.at(i).compare("coefficient") == 0) {
				newPlugin.coefficientName = _splitstring.at((uint64_t)i + 2);
			}
			break;
		}
		case PluginType::TerminationChainLengthDependent: {
			if (_splitstring.at(i).compare("coefficient") == 0) {
				newPlugin.coefficientName = _splitstring.at((uint64_t)i + 2);
			}
			if (_splitstring.at(i).compare("ic") == 0) {
				newPlugin.chainLengthDependenceParameter.ic = stoi(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("as") == 0) {
				newPlugin.chainLengthDependenceParameter.as = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("al") == 0) {
				newPlugin.chainLengthDependenceParameter.al = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("as2") == 0) {
				newPlugin.chainLengthDependenceParameter.as2 = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("as3") == 0) {
				newPlugin.chainLengthDependenceParameter.as3 = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("as4") == 0) {
				newPlugin.chainLengthDependenceParameter.as4 = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("kt11p1") == 0) {
				newPlugin.chainLengthDependenceParameter.kt11p1 = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("kt11p2") == 0) {
				newPlugin.chainLengthDependenceParameter.kt11p2 = stod(_splitstring.at((uint64_t)i + 2));
			}
			if (_splitstring.at(i).compare("kt11p3") == 0) {
				newPlugin.chainLengthDependenceParameter.kt11p3 = stod(_splitstring.at((uint64_t)i + 2));
			}
			break;
		}
		case PluginType::Undefined: {
			break;
		}
		}

	}
	_plugin.push_back(newPlugin);
	return true;
}
bool modelInterface::detectMolarMass(vector<string>& _splitstring, vector<species>& _species) {

	species newSpecies;
	// three prerequisites: a) M , b) "(" and c) "]"
	for (size_t i = 0; i < _splitstring.size(); i++) {
		if (_splitstring.at(i).compare("M") == 0) {
			if (i < (_splitstring.size() + 5)) {
				if (_splitstring.at((uint64_t)i + 1).compare("(") == 0) {
					if (_splitstring.at((uint64_t)i + 3).compare(")") == 0) {
						if (_splitstring.at((uint64_t)i + 4).compare("=") == 0) {
							// syntax ok: M(XX) = value
							int j = findSpecies(_species, _splitstring.at((uint64_t)i + 2));
							if (j != -1) {
								//species already exist, set concentration
								_species.at(j).molarMass = stod(_splitstring.at((uint64_t)i + 5));
								_species.at(j).monomer = true;
								return true;
							}
							newSpecies.name.append(_splitstring.at((uint64_t)i + 2));
							newSpecies.concentration = 0.0;
							newSpecies.filename = "";
							newSpecies.macromolecular = false;
							newSpecies.initiator = false;
							newSpecies.monomer = true;
							newSpecies.molarMass = stod(_splitstring.at((uint64_t)i + 5));
							newSpecies.f = 0.0;
							newSpecies.molecules = 0;
							_species.push_back(newSpecies);
							return true;
						}
					}
				}
			}
		}
	}
	return false;
}

bool modelInterface::detectInitiatorEfficiency(vector<string>& _splitstring, vector<species>& _species) {

	species newSpecies;
	// three prerequisites: a) f , b) "(" and c) "]"
	for (size_t i = 0; i < _splitstring.size(); i++) {
		if (_splitstring.at(i).compare("f") == 0) {
			if (i < (_splitstring.size() + 5)) {
				if (_splitstring.at((uint64_t)i + 1).compare("(") == 0) {
					if (_splitstring.at((uint64_t)i + 3).compare(")") == 0) {
						if (_splitstring.at((uint64_t)i + 4).compare("=") == 0) {
							// syntax ok: M(XX) = value
							int j = findSpecies(_species, _splitstring.at((uint64_t)i + 2));
							if (j != -1) {
								//species already exist, set concentration
								_species.at(j).initiator = true;
								_species.at(j).f = stod(_splitstring.at((uint64_t)i + 5));
								return true;
							}
							newSpecies.name.append(_splitstring.at((uint64_t)i + 2));
							newSpecies.concentration = 0.0;
							newSpecies.filename = "";
							newSpecies.macromolecular = false;
							newSpecies.initiator = false;
							newSpecies.monomer = false;
							newSpecies.molarMass = 0.0;
							newSpecies.f = stod(_splitstring.at((uint64_t)i + 5));
							newSpecies.molecules = 0;
							_species.push_back(newSpecies);
							return true;
						}
					}
				}
			}
		}
	}
	return false;
}

bool modelInterface::detectReaction(vector<string>& _splitstring, vector<reaction>& _reaction, vector<species>& _species, vector<coefficient>& _coefficient) {
	int pos1, pos2, pos3;
	size_t index1, ic;
	bool plusexpected, eductsearch;
	bool shadowMonomerEduct, shadowMonomerProduct;
	string mename, mpname;
	size_t sizespecies, sizecoefficients;
	species newSpecies;
	coefficient newCoefficient;
	reaction newReaction;

	index1 = -1;
	shadowMonomerEduct = false;
	shadowMonomerProduct = false;
	sizespecies = _species.size();
	sizecoefficients = _coefficient.size();

	// two prerequisites: a) "--" and b) "-->"
	pos1 = -1;
	pos2 = -1;
	for (size_t i = 0; i < _splitstring.size(); i++) {
		if (_splitstring.at(i).compare("--") == 0) {
			pos1 = static_cast<int>(i);
		}
		if (_splitstring.at(i).compare("-->") == 0) {
			pos2 = static_cast<int>(i);
		}
	}
	if ((pos1 == -1) || (pos2 == -1)) {
		return false;
	}
	if ((pos2 - pos1) != 2) {
		// synatx error: no coefficient
		return false;
	}
	if ((pos1 == 0) || (pos2 == static_cast<int>(_splitstring.size() - 1))) {
		// syntax error: no educts / products
		return false;
	}
	pos3 = pos1 + 1;
	int j = findCoefficient(_coefficient, _splitstring.at(pos3));
	if (j == -1) {
		//coefficient unknown
		newCoefficient.name.append(_splitstring.at(pos3));
		newCoefficient.value = 0.0;
		newCoefficient.A0 = 0.0;
		newCoefficient.Ea = 0.0;
		newCoefficient.Va = 0.0;
		newCoefficient.valueMC = 0.0;
		_coefficient.push_back(newCoefficient);
		newReaction.indexk = static_cast<int>(_coefficient.size() - 1);
	}
	else {
		newReaction.indexk = j;
	}

	// educts + products
	plusexpected = false;
	eductsearch = true;
	for (size_t i = 0; i < _splitstring.size(); i++) {
		
		if (static_cast<int>(i) == pos1) {
			// jump to products
			i = pos2 + 1;
			eductsearch = false;
			plusexpected = false;
		}
		if (plusexpected) {
			if (_splitstring.at(i).compare("+") == 0) {
				plusexpected = false;
			}
			else {
				// syntax error
				// delete changes
				_coefficient.resize(sizecoefficients);
				_species.resize(sizespecies);
				return false;
			}
		}
		else {
			// species expected
			j = findSpecies(_species, _splitstring.at(i));
			if (j == -1) {
				//species unknown
				newSpecies.name = "";
				newSpecies.name.append(_splitstring.at(i));
				newSpecies.concentration = 0.0;
				newSpecies.filename = "";
				newSpecies.f = 0.0;
				newSpecies.initiator = false;
				newSpecies.monomer = false;
				newSpecies.molarMass = 0.0;
				newSpecies.molecules = 0;
				newSpecies.macromolecular = false;
				_species.push_back(newSpecies);
				j = static_cast<int>(_species.size() - 1);
			}
			if (eductsearch) {
				newReaction.indexEducts.push_back(j);
			}
			else {
				newReaction.indexProducts.push_back(j);
			}
			plusexpected = true;
			if (i < (_splitstring.size() - 3)) {
				// macro: (n)
				if ((_splitstring.at((uint64_t)i + 1).compare("(") == 0) && (_splitstring.at((uint64_t)i + 3).compare(")") == 0)) {
					if (eductsearch) {
						_species.at(newReaction.indexEducts.back()).macromolecular = true;
					}
					else {
						_species.at(newReaction.indexProducts.back()).macromolecular = true;
						if (_splitstring.at((uint64_t)i + 2).compare("1") == 0) {
							// position of product P(1)
							index1 = newReaction.indexProducts.size() - 1;
						}
					}
					// jump over macro flag
					i += 3;
					continue;
				}
			}
			if (i < (_splitstring.size() - 5)) {
				// macro: (n+m)
				if ((_splitstring.at((uint64_t)i + 1).compare("(") == 0) && (_splitstring.at((uint64_t)i + 3).compare("+") == 0) && (_splitstring.at((uint64_t)i + 5).compare(")") == 0)) {
					if (eductsearch) {
						_species.at(newReaction.indexEducts.back()).macromolecular = true;
					}
					else {
						_species.at(newReaction.indexProducts.back()).macromolecular = true;
					}
					// jump over macro flag
					i += 5;
				}
			}
			if (i < (_splitstring.size() - 5)) {
				// macro: (m, SCB)
				// macro: (m, LCB)
				// macro: (m,Monomer)
				if ((_splitstring.at((uint64_t)i + 1).compare("(") == 0) && (_splitstring.at((uint64_t)i + 3).compare(",") == 0) && (_splitstring.at((uint64_t)i + 5).compare(")") == 0)) {
					if (eductsearch) {
						_species.at(newReaction.indexEducts.back()).macromolecular = true;
						shadowMonomerEduct = true;
						mename = _splitstring.at((uint64_t)i + 4);
					}
					else {
						_species.at(newReaction.indexProducts.back()).macromolecular = true;
						if (_splitstring.at((uint64_t)i + 4).compare("TDB") == 0) {
							newReaction.addTDB = true;
						}
						if (_splitstring.at((uint64_t)i + 4).compare("SCB") == 0) {
							newReaction.addSCB = true;
						}
						if (_splitstring.at((uint64_t)i + 4).compare("SCB1") == 0) {
							newReaction.addSCB1 = true;
						}
						if (_splitstring.at((uint64_t)i + 4).compare("LCB") == 0) {
							newReaction.addLCB = true;
						}
						if (_splitstring.at((uint64_t)i + 4).compare("radPos3") == 0) {
							newReaction.addRadicalPosition3 = true;
						}
						if ((!newReaction.addTDB) && (!newReaction.addSCB) && (!newReaction.addLCB)) {
							shadowMonomerProduct = true;
							mpname = _splitstring.at((uint64_t)i + 4);
						}
					}
					// jump over macro flag
					i += 5;
				}
			}
			if (_splitstring.size() > 8) {
				if (i < (_splitstring.size() - 7)) {
					// macro: (n+m, SCB)
					// macro: (n+m, LCB)
					if ((_splitstring.at((uint64_t)i + 1).compare("(") == 0) && (_splitstring.at((uint64_t)i + 3).compare("+") == 0) && (_splitstring.at((uint64_t)i + 5).compare(",") == 0) && (_splitstring.at((uint64_t)i + 7).compare(")") == 0)) {
						if (eductsearch) {
							_species.at(newReaction.indexEducts.back()).macromolecular = true;
						}
						else {
							_species.at(newReaction.indexProducts.back()).macromolecular = true;
							if (_splitstring.at((uint64_t)i + 6).compare("SCB") == 0) {
								newReaction.addSCB = true;
							}
							if (_splitstring.at((uint64_t)i + 6).compare("LCB") == 0) {
								newReaction.addLCB = true;
							}
						}
						// jump over macro flag
						i += 7;
					}
				}
			}
		}
	}
	if (shadowMonomerEduct && shadowMonomerProduct) {
		// species expected
		j = findSpecies(_species, mename);
		if (j == -1) {
			//species unknown
			newSpecies.name = "";
			newSpecies.name.append(mename);
			newSpecies.concentration = 0.0;
			newSpecies.filename = "";
			newSpecies.f = 0.0;
			newSpecies.initiator = false;
			newSpecies.monomer = false;
			newSpecies.molarMass = 0.0;
			newSpecies.molecules = 0;
			newSpecies.macromolecular = false;
			_species.push_back(newSpecies);
			j = static_cast<int>(_species.size() - 1);
		}
		newReaction.indexEducts.push_back(j);
		// species expected
		j = findSpecies(_species, mpname);
		if (j == -1) {
			//species unknown
			newSpecies.name = "";
			newSpecies.name.append(mpname);
			newSpecies.concentration = 0.0;
			newSpecies.filename = "";
			newSpecies.f = 0.0;
			newSpecies.initiator = false;
			newSpecies.monomer = false;
			newSpecies.molarMass = 0.0;
			newSpecies.molecules = 0;
			newSpecies.macromolecular = false;
			_species.push_back(newSpecies);
			j = static_cast<int>(_species.size() - 1);
		}
		newReaction.indexProducts.push_back(j);
	}
	// sequence of educts / products
	// P(n) + X
	if (newReaction.indexEducts.size() == 2) {
		if ((!_species.at(newReaction.indexEducts.at(0)).macromolecular) && _species.at(newReaction.indexEducts.at(1)).macromolecular) {
			// swap educts
			ic = newReaction.indexEducts.at(0);
			newReaction.indexEducts.at(0) = newReaction.indexEducts.at(1);
			newReaction.indexEducts.at(1) = ic;
		}
	}
	if (newReaction.indexProducts.size() == 2) {
		if ((!_species.at(newReaction.indexProducts.at(0)).macromolecular) && _species.at(newReaction.indexProducts.at(1)).macromolecular) {
			// swap products
			ic = newReaction.indexProducts.at(0);
			newReaction.indexProducts.at(0) = newReaction.indexProducts.at(1);
			newReaction.indexProducts.at(1) = ic;
		}
	}
	// transfer to monomer - sequence: D(n) + P(1)
	if (newReaction.indexProducts.size() == 2) {
		if (_species.at(newReaction.indexProducts.at(0)).macromolecular && _species.at(newReaction.indexProducts.at(1)).macromolecular) {
			if (index1 == 0) {
				// swap products
				ic = newReaction.indexProducts.at(0);
				newReaction.indexProducts.at(0) = newReaction.indexProducts.at(1);
				newReaction.indexProducts.at(1) = ic;
			}
		}
	}
	newReaction.rktTemplate = ReactionTemplate::Undefined;
	_reaction.push_back(newReaction);

	return true;
}
void modelInterface::allocateCoefficients(coefficient* _coefficient, vector<coefficient>& _shadow) {

	for (size_t i = 0; i < _shadow.size(); i++) {
		if (_coefficient->name.compare(_shadow.at(i).name) == 0) {
			// hit in shadow list, copy values to used coefficient
			_coefficient->A0 = _shadow.at(i).A0;
			_coefficient->Ea = _shadow.at(i).Ea;
			_coefficient->Va = _shadow.at(i).Va;
			_coefficient->value = _shadow.at(i).value;
		}
	}
}
void modelInterface::typeReaction(reaction* _reaction, vector<species>& _species) {

	bool eEducts, eProducts, eMonomer;
	int mEducts, mProducts;

	// only low molecular species ?
	// monomer as educt ?
	// determine number of macromolecular educts / products

	eEducts = true;
	eMonomer = false;
	mEducts = 0;
	for (size_t i = 0; i < _reaction->indexEducts.size(); i++) {
		if (_species.at(_reaction->indexEducts.at(i)).macromolecular) {
			eEducts = false;
			++mEducts;
		}
		if (_species.at(_reaction->indexEducts.at(i)).monomer) {
			eMonomer = true;
		}
	}
	eProducts = true;
	mProducts = 0;
	for (size_t i = 0; i < _reaction->indexProducts.size(); i++) {
		if (_species.at(_reaction->indexProducts.at(i)).macromolecular) {
			eProducts = false;
			++mProducts;
		}
	}
	if (eEducts && eProducts) {
		if (_species.at(_reaction->indexEducts.at(0)).initiator) {
			if (_reaction->indexEducts.size() == 1) {
				if (_reaction->indexProducts.size() == 2) {
					if (_reaction->indexProducts.at(0) == _reaction->indexProducts.at(1)) {
						_reaction->rktTemplate = ReactionTemplate::InitiatorDecomposition;
						return;
					}
				}
			}
			else {
				_reaction->rktTemplate = ReactionTemplate::ElementaryReaction;
				return;
			}
		}
		else {
			_reaction->rktTemplate = ReactionTemplate::ElementaryReaction;
			return;
		}
	}
	if (_reaction->indexEducts.size() == 3) {

		if ((mEducts == 2) && eMonomer) {
			_reaction->rktTemplate = ReactionTemplate::TransferToPolymer;
			runmodeKineticModel = 2;
			return;
		}
		if (mProducts == 2) {
			if (_species.at(_reaction->indexEducts.at(0)).monomer && _species.at(_reaction->indexEducts.at(1)).monomer && _species.at(_reaction->indexEducts.at(2)).monomer) {
				_reaction->rktTemplate = ReactionTemplate::ThermalInitiation3;
				return;
			}
		}
	}
	if (_reaction->indexEducts.size() == 2) {
		if (_reaction->indexProducts.size() == 1) {
			if ((mEducts == 1) && eMonomer && (mProducts == 1)) {
				if (_reaction->indexEducts.at(0) == _reaction->indexProducts.at(0)) {
					_reaction->rktTemplate = ReactionTemplate::Propagation_PnPnp1;
					return;
				}
				else {
					if (_reaction->addSCB) {
						_reaction->rktTemplate = ReactionTemplate::Propagation_SCB;
						return;
					}
					if (_reaction->addSCB1) {
						//_reaction->name = "propagationSCB: P(n) + M --kp--> Q(n+1,SCB1)";
						// FIXME SCB1 ???
						_reaction->rktTemplate = ReactionTemplate::Propagation_SCB;
						return;
					}
					_reaction->rktTemplate = ReactionTemplate::Propagation_PnQnp1;
					return;

				}
			}
			if ((mEducts == 0) && eMonomer && (mProducts == 1)) {
				_reaction->rktTemplate = ReactionTemplate::Initiation;
				return;
			}
			if ((mEducts == 2) && (mProducts == 1)) {
				if (_reaction->indexEducts.at(0) == _reaction->indexEducts.at(1)) {
						_reaction->rktTemplate = ReactionTemplate::Termination_PnPmDnpm;
				}
				else {
					_reaction->rktTemplate = ReactionTemplate::Termination_PnPmDnpm;
				}
				return;
			}
			if ((mEducts == 1) && (mProducts == 1)) {
				_reaction->rktTemplate = ReactionTemplate::Transfer;
				return;
			}
		}
		if (_reaction->indexProducts.size() == 3) {
			if (_reaction->addTDB) {
				if ((mEducts == 1) && eMonomer && (mProducts == 2)) {
					// unklar, nicht implementiert
					//_reaction->name = "transfer to monomer: P(n) + M --km--> D(n) + Q(1,TDB) + PDB";
					//_reaction->type = 19;
					// _reaction->name = mcPolymerXML->getReactionTemplateLabel(_reaction->type);
					return;
				}
			}
		}
		if (_reaction->indexProducts.size() == 2) {
			if ((mEducts == 1) && eMonomer && (mProducts == 1)) {
				if (_reaction->indexEducts.at(0) == _reaction->indexProducts.at(0)) {
					_reaction->rktTemplate = ReactionTemplate::Propagation_PnPnp1;
					return;
				}
			}
			if ((mEducts == 1) && eMonomer && (mProducts == 2)) {
				_reaction->rktTemplate = ReactionTemplate::TransferToMonomer;
				return;
			}
			if ((mEducts == 1) && (mProducts == 1)) {
				_reaction->rktTemplate = ReactionTemplate::Transfer;
				return;
			}
			if ((mEducts == 2) && (mProducts == 2)) {
				if (_reaction->addTDB) {
					_reaction->rktTemplate = ReactionTemplate::Termination_PnPmDnDm;
					return;
				}
			}
		}
	}
	if (_reaction->indexEducts.size() == 1) {
		if (_reaction->indexProducts.size() == 1) {
			if ((mEducts == 1) && (mProducts == 1)) {
				if (_reaction->addRadicalPosition3) {
					_reaction->rktTemplate = ReactionTemplate::Backbiting_RadPos3;
					return;
				}
				else {
					_reaction->rktTemplate = ReactionTemplate::Transfer;
					return;
				}
			}
		}
		if (_reaction->indexProducts.size() == 2) {
			if ((mEducts == 1) && (mProducts == 1)) {
				_reaction->rktTemplate = ReactionTemplate::Transfer;
				return;
			}
			if ((mEducts == 1) && (mProducts == 2)) {
				//_reaction->name = "fragmentation: P(n) --kfrag--> Q(n-s) + D(s)";
				_reaction->rktTemplate = ReactionTemplate::Fragmentation;
				return;
			}
		}
	}
	if (_reaction->rktTemplate == ReactionTemplate::Undefined) {
		cout << "error: reaction template " << endl;
		exit(1);
	}
}

void modelInterface::typePlugin(plugin* _plugin, vector<species>& _species, vector<coefficient>& _coefficient) {
	if (_plugin->speciesName.length() != 0) {
		_plugin->speciesIndex = findSpecies(_species, _plugin->speciesName);
	}
	if (_plugin->coefficientName.length() != 0) {
		_plugin->coefficientIndex = findCoefficient(_coefficient, _plugin->coefficientName);
	}
}
vector<speciesObject> modelInterface::getSpeciesObjects() {
	vector<speciesObject> objList;
	int id = 0;
	// first register monomers
	for (auto& elem : specieslist) {
		if (elem.monomer) {
			speciesObject speciesObj;
			speciesObj.id = id;
			speciesObj.name = elem.name;
			speciesObj.concentration = elem.concentration;
			speciesObj.optionalParameterMolarMass = elem.molarMass;
			speciesObj.isMonomer = elem.monomer;
			objList.push_back(speciesObj);
			++id;
		}
	}
	for (auto& elem : specieslist) {
		if (!elem.monomer) {
			speciesObject speciesObj;
			speciesObj.id = id;
			speciesObj.name = elem.name;
			speciesObj.concentration = elem.concentration;
			speciesObj.isMonomer = elem.monomer;
			speciesObj.isInitiator = elem.initiator;
			if (speciesObj.isInitiator) {
				speciesObj.optionalParameterF = elem.f;
			}
			speciesObj.isPolymer = elem.macromolecular;
			objList.push_back(speciesObj);
			++id;
		}
	}
	return objList;
}
vector<coefficientObject> modelInterface::getCoefficientObjects() {
	vector<coefficientObject> objList;
	int id = 0;
	for (auto& elem : coefficientlist) {
		coefficientObject coefficientObj;
		coefficientObj.id = id;
		coefficientObj.name = elem.name;
		if (elem.A0 == 0.0) {
			coefficientObj.isTemperatureDependent = false;
			coefficientObj.value = elem.value;
		}
		else {
			coefficientObj.isTemperatureDependent = true;
			coefficientObj.A0 = elem.A0;
			coefficientObj.Ea = elem.Ea;
		}
		objList.push_back(coefficientObj);
		++id;
	}
	return objList;
}
vector<reactionObject> modelInterface::getReactionObjects() {
	vector<reactionObject> objList;
	int id = 0;
	for (auto& elem : reactionlist) {
		reactionObject reactionObj;

		reactionObj.rktTemplate = elem.rktTemplate;
		reactionObj.id = id;
		reactionObj.coefficientID = static_cast<int>(elem.indexk);
		for (auto& elemReactants : elem.indexEducts) {
			int speciesID = static_cast<int>(elemReactants);
			reactionObj.eductIdList.push_back(speciesID);
		}
		for (auto& elemReactants : elem.indexProducts) {
			int speciesID = static_cast<int>(elemReactants);
			reactionObj.productIDList.push_back(speciesID);
		}
		objList.push_back(reactionObj);
		++id;
	}
	return objList;
}
vector<pluginObject> modelInterface::getPluginObjects() {
	vector<pluginObject> objList;
	int id = 0;
	for (auto& elem : pluginlist) {
		pluginObject pluginObj;
		pluginObj.type = elem.type;
		pluginObj.id = id;
		switch (elem.type) {
		case PluginType::TerminationHamielec: {
			pluginObj.templateLabel = PluginTypeToString(elem.type);
			pluginObj.templateLabelFilename = PluginTypeToString(elem.type);
			if (elem.coefficientIndex != -1) {
				pluginObj.coefficientID = elem.coefficientIndex;
			}
			else {
				cout << "model error: plugin definition " << pluginObj.templateLabel;
				cout << " coefficient not found" << endl;
				exit(1);
			}
			if (elem.speciesIndex != -1) {
				int eductID = elem.speciesIndex;
				pluginObj.speciesIdList.push_back(eductID);
			}
			else {
				cout << "model error: plugin definition " << pluginObj.templateLabel;
				cout << " species styrene not found" << endl;
				exit(1);
			}
			break;
		}
		case PluginType::TerminationChainLengthDependent: {
			pluginObj.templateLabel = PluginTypeToString(elem.type);
			pluginObj.templateLabelFilename = PluginTypeToString(elem.type);
			if (elem.coefficientIndex != -1) {
				pluginObj.coefficientID = elem.coefficientIndex;
			}
			else {
				cout << "model error: plugin definition " << pluginObj.templateLabel;
				cout << " coefficient not found" << endl;
				exit(1);
			}
			pluginObj.chainLengthDependenceParameter = elem.chainLengthDependenceParameter;
			break;
		}
		case PluginType::Undefined: {
			cout << "model error: undefined plugin";
			exit(1);
		}
		}
		objList.push_back(pluginObj);
		++id;
	}
	return objList;
}
void modelInterface::writeObjectsFile(const string& _filename) {
	auto species = getSpeciesObjects();
	auto coefficients = getCoefficientObjects();
	auto reactions = getReactionObjects();
	auto plugins = getPluginObjects();
	if (mcPolymerXML->constructModel(species, coefficients, reactions, plugins))
		exportModelToXML(_filename, mcPolymerXML);
	else
		exit(-1);
}
modelXML* modelInterface::getModel() {
	return mcPolymerXML;
}

} // namespace
