// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* modelInterpreter.h
*
*  Created on: 07.06.2017
*/
#pragma once

#include <iostream>
#include <string>
#include <vector>
#include "modelObjects.h"
#include "modelXML.h"

namespace mcPolymer_ns {

class modelInterface {
public:
	modelInterface();
	~modelInterface();

	bool readModelFromFile(const std::string& _filename);
	bool initInterface(uint64_t _numMolecules, double _temperatureCelsius, double _pressureBar);
	bool initInterface(double _controlVolumeL, double _temperatureCelsius, double _pressureBar);
	bool isInitialized() const;
	void writeObjectsFile(const std::string& _filename);
	modelXML* getModel();

private:
	const double AvogadroConstantNA = 6.022140857e23;
	bool readModelFile(const std::string& _filename);
	struct species {
		size_t indexSpecies;
		std::string name;
		bool macromolecular = false;
		bool initiator = false;
		bool monomer = false;
		bool usedFlag = false;
		double molarMass = 0.0;
		double concentration = 0.0;
		std::string filename = "";
		std::string assignedMonomer = "";
		double f = 0.0;
		uint64_t molecules = 0;
	};
	struct coefficient {
		std::string name = "";
		double value = 0.0;
		double A0 = 0.0;
		double Ea = 0.0;
		double Va = 0.0;
		double valueMC = 0.0;
		int reactionOrder = -1;
	};
	struct reaction {
		std::vector<size_t> indexEducts;
		std::vector<size_t> indexProducts;
		int indexk = 0;
		ReactionTemplate rktTemplate = ReactionTemplate::Undefined;
		// side effect
		// short chain branch SCB
		// long chain branch LCB
		// terminal double bond (macromonomer) TDB
		// beta scission on chain end (SCB) - fragments wit length 2 or 3
		// beta scission (LCB) - fragments with random chain length

		bool addSCB = false;
		bool addSCB1 = false;
		bool addLCB = false;
		bool addTDB = false;
		bool addRadicalPosition3 = false;
		int line = 0;
	};

	struct plugin {
		PluginType type = PluginType::Undefined;
		std::string coefficientName = "";
		std::string speciesName = "";
		int coefficientIndex = -1;
		int speciesIndex = -1;
		std::string filename = "";
		chainLengthDependence chainLengthDependenceParameter;
	};
	struct reactorControl {
		std::vector<double> time;
		std::vector<double> volume;
		std::vector<double> temperature;
	};
	modelXML *mcPolymerXML;
	bool initialized;
	int runmodeKineticModel = 1;
	std::vector<species> specieslist;
	std::vector<coefficient> coefficientlist;
	std::vector<coefficient> shadowlist;
	std::vector<reaction> reactionlist;
	std::vector<plugin> pluginlist;
	uint64_t moleculesParameter;
	double controlVolume;
	uint64_t molecules;
	std::vector<std::string> splitline(std::string& _line);
	std::vector<species> importSpeciesFromModelFile(const std::string& _filename);
	bool detectReaction(std::vector<std::string>& _splitstring, std::vector<reaction>& _reaction, std::vector<species>& _species, std::vector<coefficient>& _coefficient);
	bool detectConcentration(std::vector<std::string>& _splitstring, std::vector<species>& _species);
	bool detectCoeficients(std::vector<std::string>& _splitstring, std::vector<coefficient>& _coefficient, std::vector<coefficient>& _shadow);
	bool detectMolarMass(std::vector<std::string>& _splitstring, std::vector<species>& _species);
	bool detectInitiatorEfficiency(std::vector<std::string>& _splitstring, std::vector<species>& _species);
	bool detectPlugin(std::vector<std::string>& _splitstring, std::vector<plugin>& _species);
	void allocateCoefficients(coefficient* _coefficient, std::vector<coefficient>& _shadow);
	void typeReaction(reaction* _reaction, std::vector<species>& _species);
	void typePlugin(plugin* _plugin, std::vector<species>& _species, std::vector<coefficient>& _coefficient);
	int findSpecies(std::vector<species>& _species, std::string& _searchstr);
	int findCoefficient(std::vector<coefficient>& _coefficient, std::string& _searchstr);
	bool filenameValid(const std::string& _filename) const;
	std::vector<speciesObject> getSpeciesObjects();
	std::vector<coefficientObject> getCoefficientObjects();
	std::vector<reactionObject> getReactionObjects();
	std::vector<pluginObject> getPluginObjects();
};

} // namespace
