// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* modelXML.cpp
*
* Created on: 15.02.2020
*/

#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <math.h>
#include "modelXML.h"
#include "logging.h"
#include "parsing.h"

using namespace std;
namespace mcPolymer_ns {

modelXML::modelXML() {
}
modelXML::~modelXML() {
}

bool modelXML::constructModel(std::vector<speciesObject> _speciesObjects,
	std::vector<coefficientObject> _coefficientObjects,
	std::vector<reactionObject> _reactionObjects,
	std::vector<pluginObject> _pluginObjects) {
	addSpeciesObjects(_speciesObjects);
	addCoefficientObjects(_coefficientObjects);
	if (!addReactionObjects(_reactionObjects))
		return false;
	addPluginObjects(_pluginObjects);
	return true;
}

void modelXML::addSpeciesObjects(std::vector<speciesObject>& _speciesObjects) {
	speciesObjectList = _speciesObjects;
}
void modelXML::addCoefficientObjects(std::vector<coefficientObject>& _coefficientObjects) {
	coefficientObjectList = _coefficientObjects;
}
bool modelXML::addReactionObjects(std::vector<reactionObject>& _reactionObjects) {
	for (auto& reactionObj : _reactionObjects) {
		if (int(coefficientObjectList.size()) <= reactionObj.coefficientID) {
			logError("modelXML::addReactionObjects: coefficientObjectList size smaller than coefficientID");
			return false;
		}
		coefficientObjectList.at(reactionObj.coefficientID).reactionOrder = reactionObj.eductIdList.size();
		if (reactionObj.rktTemplate == ReactionTemplate::TransferToPolymer) {
			coefficientObjectList.at(reactionObj.coefficientID).reactionOrder = 2;
		}
		if (coefficientObjectList.at(reactionObj.coefficientID).reactionOrder == 2) {
			if (reactionObj.eductIdList.size() < 2) {
				logError("modelXML::addReactionObjects: eductIdList too small");
				return false;
			}
			if (reactionObj.eductIdList.at(0) == reactionObj.eductIdList.at(1)) {
				coefficientObjectList.at(reactionObj.coefficientID).order2EE = true;
			}
		}
	}
	reactionObjectList = _reactionObjects;
	return true;
}
void modelXML::addPluginObjects(std::vector<pluginObject>& _pluginObjects) {
	pluginObjectList = _pluginObjects;
}

void modelXML::initCoefficients() {
	for (auto& elem : reactionObjectList) {
		coefficientObjectList.at(elem.coefficientID).initialized = false;
	}
	for (auto& elem : reactionObjectList) {
		if (coefficientObjectList.at(elem.coefficientID).initialized) {
			continue;
		}
		coefficientObjectList.at(elem.coefficientID).order2EE = false;
		if (elem.eductIdList.size() == 1) {
			coefficientObjectList.at(elem.coefficientID).reactionOrder = 1;
		}
		if (elem.eductIdList.size() == 2) {
			coefficientObjectList.at(elem.coefficientID).reactionOrder = 2;
			if (elem.eductIdList.at(0) == elem.eductIdList.at(1)) {
				coefficientObjectList.at(elem.coefficientID).order2EE = true;
			}
		}
		if (elem.eductIdList.size() == 3) {
			if (elem.rktTemplate == ReactionTemplate::TransferToPolymer) {
				coefficientObjectList.at(elem.coefficientID).reactionOrder = 2;
			}
			else {
				coefficientObjectList.at(elem.coefficientID).reactionOrder = 3;
			}
		}
		if (coefficientObjectList.at(elem.coefficientID).value == 0.0) {
			coefficientObjectList.at(elem.coefficientID).isTemperatureDependent = true;
			if (coefficientObjectList.at(elem.coefficientID).deltaV != 0.0) {
				coefficientObjectList.at(elem.coefficientID).isPressureDependent = true;
			}
		}
	}
}
void modelXML::calcultateCoefficients() {
	for (auto& elem : reactionObjectList) {
		coefficientObjectList.at(elem.coefficientID).initialized = false;
	}
	for (auto& elem : reactionObjectList) {
		if (coefficientObjectList.at(elem.coefficientID).initialized) {
			continue;
		}
		if (coefficientObjectList.at(elem.coefficientID).isTemperatureDependent) {
			if (coefficientObjectList.at(elem.coefficientID).isPressureDependent) {
				coefficientObjectList.at(elem.coefficientID).value = coefficientObjectList.at(elem.coefficientID).A0 * exp(-1.0 * coefficientObjectList.at(elem.coefficientID).Ea * 1000.0 / (reactor.temperature + 273.15) / 8.314) * exp(-1.0 * coefficientObjectList.at(elem.coefficientID).deltaV / 10.0 * reactor.pressure / (reactor.temperature + 273.15) / 8.314);
			}
			else {
				coefficientObjectList.at(elem.coefficientID).value = coefficientObjectList.at(elem.coefficientID).A0 * exp(-1.0 * coefficientObjectList.at(elem.coefficientID).Ea * 1000.0 / (reactor.temperature + 273.15) / 8.314);
			}
		}
		if (coefficientObjectList.at(elem.coefficientID).reactionOrder == 2) {
			if (coefficientObjectList.at(elem.coefficientID).order2EE) {
				coefficientObjectList.at(elem.coefficientID).valueMC = 2.0 * coefficientObjectList.at(elem.coefficientID).value / (reactor.controlVolume * AvogadroConstantNA);
			}
			else {
				coefficientObjectList.at(elem.coefficientID).valueMC = coefficientObjectList.at(elem.coefficientID).value / (reactor.controlVolume * AvogadroConstantNA);
			}
			continue;
		}
		if (coefficientObjectList.at(elem.coefficientID).reactionOrder == 1) {
			coefficientObjectList.at(elem.coefficientID).valueMC = coefficientObjectList.at(elem.coefficientID).value;
		}
		if (coefficientObjectList.at(elem.coefficientID).reactionOrder == 3) {
			coefficientObjectList.at(elem.coefficientID).valueMC = coefficientObjectList.at(elem.coefficientID).value / (reactor.controlVolume * AvogadroConstantNA) / (reactor.controlVolume * AvogadroConstantNA);
		}
	}
}
void modelXML::initModel(double _controlVolume, double _temperature, double _pressure) {
	// determine number of molecules
	double csum = 0;
	for (auto& elem : speciesObjectList) {
		csum += elem.concentration;
	}
	uint64_t molecules = llround(_controlVolume * AvogadroConstantNA * csum);
	initModel(molecules, _temperature, _pressure);
}
void modelXML::initModel(uint64_t _molecules, double _temperature, double _pressure) {
	for (auto& elem : coefficientObjectList) {
		elem.initialized = false;
	}
	double csum = 0;
	for (auto& elem : speciesObjectList) {
		csum += elem.concentration;
	}
	if (csum > 0.0) {
		for (auto& elem : speciesObjectList) {
			elem.molecules = llround(elem.concentration / csum * _molecules);
		}
	}
	else {
		for (auto& elem : speciesObjectList) {
			elem.molecules = 0;
		}
	}
	reactor.controlVolume = static_cast <double>(_molecules) / AvogadroConstantNA / csum;
	reactor.t0controlVolume = reactor.controlVolume;
	reactor.temperature = _temperature;
	reactor.pressure = _pressure;
	initCoefficients();
	calcultateCoefficients();
}
const vector<speciesObject> modelXML::getSpecies() const {
	return speciesObjectList;
}
const vector<coefficientObject> modelXML::calculateAndGetCoefficients() {
	calcultateCoefficients();
	return coefficientObjectList;
}
const vector<reactionObject> modelXML::getReactions() const {
	return reactionObjectList;
}
const vector<pluginObject> modelXML::getPlugins() const {
	return pluginObjectList;
}
void modelXML::scaleControlVolume(double _factor) {
	reactor.controlVolume *= _factor;
	calcultateCoefficients();
}
void modelXML::setControlVolume(double _controlVolume) {
	reactor.controlVolume = _controlVolume;
	calcultateCoefficients();
}
void modelXML::setTemperature(double _temperature) {
	reactor.temperature = _temperature;
	calcultateCoefficients();
}
double modelXML::getRatioConcMolecules() const {
	return 1.0 / (reactor.controlVolume * AvogadroConstantNA);
}

double modelXML::getVolumet0() {
	return reactor.t0Volume;
}
double modelXML::getControlVolume() {
	return reactor.controlVolume;
}
double modelXML::getVolume() {
	return reactor.volume;
}
double modelXML::getTemperature() {
	return reactor.temperature;
}

} // namespace

