// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* polymerspecies.h
*
*  Created on: 16.01.2014
*/
#include <stdio.h>
#include <algorithm>
#include "modelObjects.h"
#include "Fragment.h"
#include "Polymer.h"
#include "polymerSpecies.h"
#include "MersenneTwister.h"
#include "RandomBuffer.h"

using namespace std;
namespace mcPolymer_ns {

polymerSpecies::polymerSpecies(int _id) {
	id = _id;
	properties.numMonomers.resize(numberMonomers, 0);
}
polymerSpecies:: ~polymerSpecies() {
}

int polymerSpecies::getID() const {
	return id;
}

Polymer* polymerSpecies::getPolymer() const {
	size_t index;
	double randr;
	do {
		randr = mtrandr();
		index = static_cast<size_t>(randr * p.size());
		if (index < p.size()) {
			break;
		}
	} while (true);
	return p[index];
}
void polymerSpecies::newP1(int _id) {
	Polymer* p1;
	p1 = new Polymer(_id);
	p1->activateECR();
	if (transferToPolymer) {
		bintree->addPolymer(p1);
	}
	else {
		p.push_back(p1);
	}
}
void polymerSpecies::incrP(int _id) {
	Polymer* p1;
	p1 = getPolymer();
	p1->addM(_id);
	++properties.numMonomers.at(_id);
}
Polymer* polymerSpecies::deleteP() {
	return deleteP(0);
}
Polymer* polymerSpecies::deleteP(uint64_t _numPolymers) {
	double randr;
	size_t index;
	Polymer* pn = nullptr;
	if (transferToPolymer) {
		uint64_t targetPolymer = static_cast<uint64_t>(_numPolymers * randBufferInstance->getRand());
		pn = bintree->getPolymerLastLevel(targetPolymer);
		return pn;
	}
	else {
		if (p.size() == 0) {
			return nullptr;
		}

		if (p.size() > 1) {
			do {
				randr = mtrandr();
				index = static_cast<size_t>(randr * p.size());
				if (index < p.size()) {
					break;
				}
			} while (true);

			pn = p[index];
			p[index] = p.back();
			p.pop_back();
		}
		else {
			// == 1 !
			pn = p.at(0);
			p.pop_back();
		}
		return pn;
	}
}
Polymer* polymerSpecies::deletePolymerForTransfer(int _id) {
	double randr = mtrandr();
	uint64_t targetMonomer;
	
	while (true) {
		targetMonomer = static_cast<uint64_t>(bintree->getSumMonomers() * randr);
		if (targetMonomer > 0) {
			break;
		}
	}
	Polymer* polymer = bintree->getPolymer(targetMonomer);
	return polymer;
}
uint64_t polymerSpecies::getNumTransferMonomers() const {
	return bintree->getSumMonomers();
}

void polymerSpecies::addPolymer(Polymer* p1) {
	if (transferToPolymer) {
		bintree->addPolymer(p1);
	}
	else {
		p.push_back(p1);
	}
}
void polymerSpecies::addPolymers(vector<Polymer*> _p) {
	p.insert(end(p), begin(_p), end(_p));
	if (transferToPolymer) {
		for (auto& elem : _p) {
			bintree->addPolymer(elem);
		}
	}
}
vector<Polymer*> polymerSpecies::getAllPolymers() {
	return p;
}

void polymerSpecies::addPolymerFromCombination(Polymer* p1, Polymer* p2) {
	
	PropagatingRadicalType radicalP1 = p1->getPropagatingRadicalType();
	PropagatingRadicalType radicalP2 = p2->getPropagatingRadicalType();

	// transfer polymer 2
	size_t nFragments = p2->getFragmentSize();
	for (size_t i = 0; i < nFragments; i++) {
		p1->addFragment(p2->deleteFragment());
	}
	size_t nConnects = p2->getConnects();
	for (size_t i = 0; i < nConnects; i++) {
		p1->addConnect(p2->deleteConnect());
	}
	ConnectType cType = ConnectType::Undefined;
	if (radicalP1 == radicalP2) {
		switch (radicalP1) {
			case PropagatingRadicalType::ECR:
				// ECR --> 1 + ECR --> 1 -->--> type 0
				cType = ConnectType::Termination;
				break;
			case PropagatingRadicalType::MCR:
				// MCR --> 2 + MCR --> 2 -->--> type 0
				cType = ConnectType::Termination;
				p1->addSCB();
				p2->addSCB();
				break;
			case PropagatingRadicalType::MCRL:
				// MCRL --> 3 + MCRL --> 3 -->--> type 3
				cType = ConnectType::Crosslink;
				break;
			case PropagatingRadicalType::Undefined:
				cType = ConnectType::Undefined;
				break;
		}
	}
	else {
		if (((radicalP1 == PropagatingRadicalType::ECR) && (radicalP2 == PropagatingRadicalType::MCR)) 
				|| ((radicalP2 == PropagatingRadicalType::ECR) && (radicalP1 == PropagatingRadicalType::MCR))) {
			// ECR --> 1 + MCR --> 2 -->--> type 0
			cType = ConnectType::Termination;
			p2->addSCB();
		}
		else {
			if (((radicalP1 == PropagatingRadicalType::ECR) && (radicalP2 == PropagatingRadicalType::MCR)) 
				|| ((radicalP2 == PropagatingRadicalType::ECR) && (radicalP1 == PropagatingRadicalType::MCR))) {
				// ECR --> 1 + MCR --> 3 -->--> type 1
				cType = ConnectType::LCB;
			}
			else {
				if (((radicalP1 == PropagatingRadicalType::MCR) && (radicalP2 == PropagatingRadicalType::MCRL))
					|| ((radicalP2 == PropagatingRadicalType::MCR) && (radicalP1 == PropagatingRadicalType::MCRL))) {
					// MCR --> 2 + MCRL --> 3 -->--> type 1
					cType = ConnectType::LCB;
					p1->addSCB();
				}
			}
		}
	}
	
	p1->addConnect(p1->getPropagatingFragment(), p2->getPropagatingFragment(), cType);
	p1->removeRadical();
	delete p2;

	if (transferToPolymer) {
		bintree->addPolymer(p1);
	}
	else {
		p.push_back(p1);
	}
}
bool polymerSpecies::importSpecies(const std::string &_filename) {
	ifstream file;
	int size;
	size_t size_old;
	size_t i;

	file.open(_filename, ios::binary);
	if (!file.is_open()) {
		cout << "Failed open file " << _filename << endl;
		return false;
	}

	file.read((char *)&size, sizeof(int));
	size_old = p.size();
	p.resize(size_old + size);

	for (i = size_old; i < p.size(); i++) {
		p.at(i) = new Polymer();
		p.at(i)->importPolymer(&file);
	}

	file.close();
	return true;
}

void polymerSpecies::getCLD(Histogram* _CLD, const vector<Histogram*> &_monomerHistogram) {
	if (transferToPolymer) {
		p = bintree->analyzeAllPolymers();
	}
	if (p.size() > 0) {
		for (auto& elemP : p) {
			int cLength = elemP->getLength();
			int nSCB = elemP->getSCB();
			int nLCB = elemP->getNumLCB();
			_CLD->addValue(cLength, 1, nSCB, nLCB);
			for (size_t i = 0; i < _monomerHistogram.size(); i++) {
				int nM1 = elemP->getM(static_cast<int>(i));
				_monomerHistogram.at(i)->addValue(cLength, nM1);
			}
		}
	}
}
void polymerSpecies::exportCLD(const string& _filename, const modelXML* _mcPolymerModel) {

	vector<string> monomerName;
	vector<double> molarMass;
	int nMonomers = 0;
	const vector<speciesObject> speciesObjects = _mcPolymerModel->getSpecies();
	for (auto& elem : speciesObjects) {
		if (elem.isMonomer) {
			// monomer
			++nMonomers;
			monomerName.push_back(elem.name);
			molarMass.push_back(elem.optionalParameterMolarMass);
		}
	}
	Histogram* lH = new Histogram();
	Histogram** M1 = new Histogram * [nMonomers];
	for (int i = 0; i < nMonomers; i++) {
		M1[i] = new Histogram();
	}
	if (transferToPolymer) {
		p = bintree->analyzeAllPolymers();
	}
	if (p.size() > 0) {
		for (auto& elemP : p) {
			int cLength = elemP->getLength();
			int nSCB = elemP->getSCB();
			int nLCB = elemP->getNumLCB();
			lH->addValue(cLength, 1, nSCB, nLCB);
			for (int i = 0; i < nMonomers; i++) {
				int nM1 = elemP->getM(i);
				M1[i]->addValue(cLength, nM1);
			}
		}
		string filenamecld = _filename + ".CLD.dat";
		ofstream CLDfile(filenamecld, ofstream::out);
		CLDfile << "numMonomers:\t" << nMonomers << endl;

		vector<double> partMass;
		uint64_t sumMonomers = 0;
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << monomerName.at(i) << ":\t" << molarMass.at(i) << endl;
			// calculate average molar mass
			partMass.push_back(M1[i]->getSy1() * 1.0 * molarMass.at(i));
			sumMonomers += M1[i]->getSy1();
		}
		double averageMolarMass = 0.0;
		for (int i = 0; i < nMonomers; i++) {
			averageMolarMass += partMass.at(i);
		}
		averageMolarMass = averageMolarMass / (sumMonomers * 1.0);

		CLDfile << "Mn:\t" << lH->getPn() * averageMolarMass << "\tMw:\t" << lH->getPw() * averageMolarMass << "\tMw/Mn:\t" << lH->getPw() / lH->getPn() << endl;
		CLDfile << "numChains:\t" << lH->getSy1() << "\tnumSCB:\t" << lH->getSy2() << "\tnumLCB:\t" << lH->getSy3() << endl;
		CLDfile << "Pn:\t" << lH->getPn() << "\tPw:\t" << lH->getPw() << "\tPw/Pn:\t" << lH->getPw() / lH->getPn() << endl;
		CLDfile << "chainLength\tnumChains\tnumSCB\tnumLCB";
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << "\t" << monomerName.at(i);
		}
		CLDfile << endl;
		for (size_t i = 0; i < lH->getSize(); i++) {
			CLDfile << lH->getX(i) << "\t" << lH->getY1(i) << "\t" << lH->getY2(i) << "\t" << lH->getY3(i);
			for (int j = 0; j < nMonomers; j++) {
				CLDfile << "\t" << M1[j]->getY1(i);
			}
			CLDfile << endl;
		}
	}
	else {
		string filenamecld = _filename + ".CLD.dat";
		ofstream CLDfile(filenamecld, ofstream::out);
		CLDfile << "numMonomers:\t" << nMonomers << endl;
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << monomerName.at(i) << ":\t" << molarMass.at(i) << endl;
		}
		CLDfile << "Mn:\t" << 0.0 << "\tMw:\t" << 0.0 << "\tMw/Mn:\t" << "-" << endl;
		CLDfile << "numChains:\t" << 0 << "\tnumSCB:\t" << 0 << "\tnumLCB:\t" << 0 << endl;
	}

	delete lH;
	for (int i = 0; i < nMonomers; i++) {
		delete M1[i];
	}
	delete[] M1;
	if (transferToPolymer) {
		p.resize(0);
	}
}
void polymerSpecies::writeCLD(const string& _filename, const map<int, string>& _speciesNameMap, const map<int, int>& _speciesTypeMap, const map<int, double>& _molarMassMap) const {

	vector<string> monomerName;
	vector<double> molarMass;
	int nMonomers = 0;
	for (auto& elemT : _speciesTypeMap) {
		if ((elemT.second == 2) || (elemT.second == 52)) {
			// monomer
			++nMonomers;
			for (auto& elemN : _speciesNameMap) {
				if (elemT.first == elemN.first) {
					monomerName.push_back(elemN.second);
					for (auto& elemM : _molarMassMap) {
						if (elemN.first == elemM.first) {
							molarMass.push_back(elemM.second);
						}
					}
				}
			}
		}
	}
	Histogram* lH = new Histogram();
	Histogram** M1 = new Histogram*[nMonomers];
	for (int i = 0; i < nMonomers; i++) {
		M1[i] = new Histogram();
	}
	if (p.size() > 0) {
		for (auto& elemP : p) {
			int cLength = elemP->getLength();
			int nSCB = elemP->getSCB();
			int nLCB = elemP->getNumLCB();
			lH->addValue(cLength, 1, nSCB, nLCB);
			for (int i = 0; i < nMonomers; i++) {
				int nM1 = elemP->getM(i);
				M1[i]->addValue(cLength, nM1);
			}
		}
		string filenamecld = _filename + ".CLD.dat";
		ofstream CLDfile(filenamecld, ofstream::out);
		CLDfile << "numMonomers:\t" << nMonomers << endl;

		vector<double> partMass;
		uint64_t sumMonomers = 0;
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << monomerName.at(i) << ":\t" << molarMass.at(i) << endl;
			// calculate average molar mass
			partMass.push_back(M1[i]->getSy1()*1.0*molarMass.at(i));
			sumMonomers += M1[i]->getSy1();
		}
		double averageMolarMass = 0.0;
		for (int i = 0; i < nMonomers; i++) {
			averageMolarMass += partMass.at(i);
		}
		averageMolarMass = averageMolarMass / (sumMonomers * 1.0);

		CLDfile << "Mn:\t" << lH->getPn() * averageMolarMass << "\tMw:\t" << lH->getPw() * averageMolarMass << "\tMw/Mn:\t" << lH->getPw() / lH->getPn() << endl;
		CLDfile << "numChains:\t" << lH->getSy1() << "\tnumSCB:\t" << lH->getSy2() << "\tnumLCB:\t" << lH->getSy3() << endl;
		CLDfile << "Pn:\t" << lH->getPn() << "\tPw:\t" << lH->getPw() << "\tPw/Pn:\t" << lH->getPw() / lH->getPn() << endl;
		CLDfile << "chainLength\tnumChains\tnumSCB\tnumLCB";
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << "\t" << monomerName.at(i);
		}
		CLDfile << endl;
		for (size_t i = 0; i < lH->getSize(); i++) {
			CLDfile << lH->getX(i) << "\t" << lH->getY1(i) << "\t" << lH->getY2(i) << "\t" << lH->getY3(i);
			for (int j = 0; j < nMonomers; j++) {
				CLDfile << "\t" << M1[j]->getY1(i);
			}
			CLDfile << endl;
		}
	}
	else {
		string filenamecld = _filename + ".CLD.dat";
		ofstream CLDfile(filenamecld, ofstream::out);
		CLDfile << "numMonomers:\t" << nMonomers << endl;
		for (int i = 0; i < nMonomers; i++) {
			CLDfile << monomerName.at(i) << ":\t" << molarMass.at(i) << endl;
		}
		CLDfile << "Mn:\t" << 0.0 << "\tMw:\t" << 0.0 << "\tMw/Mn:\t" << "-" << endl;
		CLDfile << "numChains:\t" << 0 << "\tnumSCB:\t" << 0 << "\tnumLCB:\t" << 0 << endl;
	}

	delete lH;
	for (int i = 0; i < nMonomers; i++) {
		delete M1[i];
	}
	delete[] M1;
}
void polymerSpecies::exportPolymers(ofstream* _os) const {
	if (transferToPolymer) {
		bintree->exportPolymers(_os);
	}
	else {
		if (p.size() == 0) {
			return;
		}
		for (auto& elem : p) {
			elem->exportPolymer(_os);
		}
	}
}

void polymerSpecies::exportSelectedPolymers(uint64_t sizeSelected) const {
	std::string filename = "selectedPolymers.bin";
	std::ofstream file;
	file.open(filename, std::ios::binary);

	if (!file.is_open()) {
		std::cout << "Failed open file " << filename << std::endl;
		return;
	}
	size_t size = p.size();
	file.write((const char*)&sizeSelected, sizeof(size_t));

	for (size_t i = 0; i < size; i++) {
		if (p.at(i)->getSelection()) {
			p.at(i)->exportPolymer(&file);
			p.at(i)->deselect();
		}
	}
	file.close();
}


size_t polymerSpecies::getSize() const {
	if (transferToPolymer) {
		return bintree->getNumPolymers();
	}
	else {
		return p.size();
	}
}

int polymerSpecies::getLength(int _index) const {
	return p.at(_index)->getLength();
}
uint64_t polymerSpecies::getNumMonomer(int _id) const {
	uint64_t nMonomers = 0;
	for (const auto& elem : p) {
		nMonomers += elem->getM(_id);
	}
	return nMonomers;
}
vector<int> polymerSpecies::getChainLengthList() const {
	vector<int> chainLengthList;
	for (const auto& elem : p) {
		int chainLength = elem->getLength();
		chainLengthList.push_back(chainLength);
	}
	return chainLengthList;
}
void polymerSpecies::setTransferMonomer(int _id) {
	transferToPolymer = true;
	transferMonomerID = _id;
	// max Pn = 2^18 = 262144
	bintree = new binTree(18);
}
polymerSpeciesProperties polymerSpecies::getPolymerSpeciesProperties() {
	if (transferToPolymer) {
		p = bintree->analyzeAllPolymers();
	}
	properties.numPolymers = p.size();
	properties.numMonomers.resize(numberMonomers, 0);
	for (const auto& elem : p) {
		for (size_t i = 0; i < numberMonomers; i++) {
			properties.numMonomers.at(i) += elem->getM(i);
		}
		properties.numSCB += elem->getSCB();
		properties.numLCB += elem->getNumLCB();
	}
	if (transferToPolymer) {
		p.resize(0);
	}
	return properties;
}
} // namespace
