// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* class Simulation
*
*  Created on: 24.08.2008
*/

#define _USE_MATH_DEFINES
#include <iostream>
#include <string>
#include <algorithm>
#include "logging.h"
#include "modelObjects.h"
#include "modelXML.h"
#include "MersenneTwister.h"
#include "polymerSpecies.h"
#include "Plugin.h"
#include "RandomBuffer.h"
#include "Reaction.h"
#include "simulation.h"

using namespace std;
namespace mcPolymer_ns {

size_t numberMonomers = 0;

// class Simulation
polymerSpecies* Simulation::getPolymerSpecies(int _id) {
	for (auto& elem : macromolSpecies) {
		if (_id == elem->getID()) {
			return elem;
		}
	}
	return nullptr;
}
Simulation::Simulation(modelXML* _mcPolymerModel, const string& _outputDir) {
	const vector<speciesObject> speciesList = _mcPolymerModel->getSpecies();
	size_t numMacroSpecies = initMoleculeCounter(speciesList);
	initSolidsContentBalance();
	initMacromolSpecies(speciesList, numMacroSpecies);
	initCoefficients(_mcPolymerModel);
	initReactions(_mcPolymerModel, speciesList);
	registerPlugins(_mcPolymerModel, _outputDir);
	t = 0.0;
}

Simulation::~Simulation() {
}

size_t Simulation::initMoleculeCounter(const vector<speciesObject>& _speciesList) {
	size_t numMacroSpecies = 0;
	size_t countMonomers = 0;
	moleculeCounter.resize(_speciesList.size());
	moleculeCounterLastStep.resize(_speciesList.size());
	for (const auto& elem : _speciesList) {
		if (elem.isPolymer) {
			numMacroSpecies += 1;
		}
		if (elem.isMonomer) {
			countMonomers += 1;
		}
		moleculeCounter[elem.id] = elem.molecules;
	}
	moleculeCounterLastStep = moleculeCounter;
	numberMonomers = countMonomers;
	return numMacroSpecies;
}

void Simulation::initMacromolSpecies(const vector<speciesObject>& _speciesList, size_t _numMacroSpecies) {
	macromolSpecies.resize(_numMacroSpecies);
	valarray<int> macromolSpeciesIDlist;
	macromolSpeciesIDlist.resize(_numMacroSpecies);
	int countPolymer = 0;
	for (const auto& elem : _speciesList) {
		if (elem.isPolymer) {
			// collect speciesID
			macromolSpeciesIDlist[countPolymer] = elem.id;
			++countPolymer;
		}
	}
	if (macromolSpecies.size() > 0) {
		int tmpid = 0;
		for (auto& elem : macromolSpecies) {
			// speciesID == macromolSpeciesID
			elem = new polymerSpecies(macromolSpeciesIDlist[tmpid]);
			++tmpid;
		}
	}
}

void Simulation::initSolidsContentBalance() {
	solidsContentBalance.solidsContentMoleculeCounter.resize(numberMonomers, 0);
	solidsContentBalance.monomerMoleculeCounter.resize(numberMonomers);
	for (size_t i = 0; i < numberMonomers; i++) {
		solidsContentBalance.monomerMoleculeCounter[i] = &moleculeCounter[i];
	}
}

void Simulation::initCoefficients(modelXML* _modelXML) {
	vector<coefficientObject> coefficientList = _modelXML->calculateAndGetCoefficients();
	reactionRateCoefficientsMC.resize(coefficientList.size());
	for (size_t i = 0; i < coefficientList.size(); i++) {
		reactionRateCoefficientsMC[i] = coefficientList.at(i).valueMC;
	}
}

void Simulation::initReactions(const modelXML* _modelXML, const vector<speciesObject>& _speciesList) {
	const vector<reactionObject> reactionList = _modelXML->getReactions();
	reactions.resize(reactionList.size());
	for (const auto& elem : reactionList) {
		switch (elem.rktTemplate) {
		case ReactionTemplate::Backbiting_RadPos3: {
			vector<uint64_t*> eductlist;
			for (size_t i = 0; i < elem.eductIdList.size(); i++) {
				eductlist.push_back(&moleculeCounter[elem.eductIdList.at(i)]);
			}
			vector<uint64_t*> productlist;
			for (size_t i = 0; i < elem.productIDList.size(); i++) {
				productlist.push_back(&moleculeCounter[elem.productIDList.at(i)]);
			}
			reactions[elem.id] = new Backbiting_RadPos3(&reactionRateCoefficientsMC[elem.coefficientID],
				eductlist,
				productlist,
				getPolymerSpecies(elem.eductIdList.at(0)),
				getPolymerSpecies(elem.productIDList.at(0)));
			break;
		}
		case ReactionTemplate::ElementaryReaction: {
			vector<uint64_t*> eductlist;
			for (size_t i = 0; i < elem.eductIdList.size(); i++) {
				eductlist.push_back(&moleculeCounter[elem.eductIdList.at(i)]);
			}
			vector<uint64_t*> productlist;
			for (size_t i = 0; i < elem.productIDList.size(); i++) {
				productlist.push_back(&moleculeCounter[elem.productIDList.at(i)]);
			}
			reactions[elem.id] = new ElementaryReaction(&reactionRateCoefficientsMC[elem.coefficientID],
															eductlist, productlist);
			break;
		}
		case ReactionTemplate::Fragmentation: {
			reactions[elem.id] = new Fragmentation(&reactionRateCoefficientsMC[elem.coefficientID],
				&moleculeCounter[elem.eductIdList.at(0)],
				&moleculeCounter[elem.productIDList.at(0)],
				&moleculeCounter[elem.productIDList.at(1)],
				getPolymerSpecies(elem.eductIdList.at(0)),
				getPolymerSpecies(elem.productIDList.at(0)),
				getPolymerSpecies(elem.productIDList.at(1)));
			break;
		}
		case ReactionTemplate::Initiation: {
			reactions[elem.id] = new Initiation(&reactionRateCoefficientsMC[elem.coefficientID],
													&moleculeCounter[elem.eductIdList.at(0)],
													&moleculeCounter[elem.eductIdList.at(1)],
													&moleculeCounter[elem.productIDList.at(0)],
													getPolymerSpecies(elem.productIDList.at(0)),
													elem.eductIdList.at(1),
													&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(1)]);
			break;
		}
		case ReactionTemplate::InitiatorDecomposition:
			reactions[elem.id] = new InitiatorDecomposition(&reactionRateCoefficientsMC[elem.coefficientID],
				&moleculeCounter[elem.eductIdList.at(0)],
				&moleculeCounter[elem.productIDList.at(0)],
				_speciesList.at(elem.eductIdList.at(0)).optionalParameterF);
			break;
		case ReactionTemplate::Propagation_SCB: {
			reactions[elem.id] = new Propagation_SCB(&reactionRateCoefficientsMC[elem.coefficientID],
				&moleculeCounter[elem.eductIdList.at(0)],
				&moleculeCounter[elem.eductIdList.at(1)],
				&moleculeCounter[elem.productIDList.at(0)],
				getPolymerSpecies(elem.eductIdList.at(0)),
				getPolymerSpecies(elem.productIDList.at(0)),
				elem.eductIdList.at(1),
				&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(1)]);
			break;
		}
		case ReactionTemplate::Propagation_PnPnp1: {
			reactions[elem.id] = new Propagation_PnPnp1(&reactionRateCoefficientsMC[elem.coefficientID],
														&moleculeCounter[elem.eductIdList.at(0)],
														&moleculeCounter[elem.eductIdList.at(1)],
														&moleculeCounter[elem.productIDList.at(0)],
														getPolymerSpecies(elem.eductIdList.at(0)),
														getPolymerSpecies(elem.productIDList.at(0)),
														elem.eductIdList.at(1),
														&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(1)]);
			break;
		}
		case ReactionTemplate::Propagation_PnQnp1: {
			reactions[elem.id] = new Propagation_PnQnp1(&reactionRateCoefficientsMC[elem.coefficientID],
														&moleculeCounter[elem.eductIdList.at(0)],
														&moleculeCounter[elem.eductIdList.at(1)],
														&moleculeCounter[elem.productIDList.at(0)],
														getPolymerSpecies(elem.eductIdList.at(0)),
														getPolymerSpecies(elem.productIDList.at(0)),
														elem.eductIdList.at(1),
														&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(1)]);
			break;
		}
		case ReactionTemplate::Termination_PnPmDnDm: {
			reactions[elem.id] = new Termination_PnPmDnDm(&reactionRateCoefficientsMC[elem.coefficientID],
														&moleculeCounter[elem.eductIdList.at(0)],
														&moleculeCounter[elem.eductIdList.at(1)],
														&moleculeCounter[elem.productIDList.at(0)],
														&moleculeCounter[elem.productIDList.at(1)],
														getPolymerSpecies(elem.eductIdList.at(0)),
														getPolymerSpecies(elem.eductIdList.at(1)),
														getPolymerSpecies(elem.productIDList.at(0)),
														getPolymerSpecies(elem.productIDList.at(1)));
			break;
		}
		case ReactionTemplate::Termination_PnPmDnpm: {
			reactions[elem.id] = new Termination_PnPmDnpm(&reactionRateCoefficientsMC[elem.coefficientID],
														&moleculeCounter[elem.eductIdList.at(0)],
														&moleculeCounter[elem.eductIdList.at(1)],
														&moleculeCounter[elem.productIDList.at(0)],
														getPolymerSpecies(elem.eductIdList.at(0)),
														getPolymerSpecies(elem.eductIdList.at(1)),
														getPolymerSpecies(elem.productIDList.at(0)));
			break;
		}
		case ReactionTemplate::ThermalInitiation3: {
			reactions[elem.id] = new ThermalInitiation3(&reactionRateCoefficientsMC[elem.coefficientID],
				&moleculeCounter[elem.eductIdList.at(0)],
				&moleculeCounter[elem.productIDList.at(0)],
				getPolymerSpecies(elem.productIDList.at(0)),
				elem.eductIdList.at(1));
			break;
		}
		case ReactionTemplate::Transfer: {
			vector<uint64_t*> eductlist;
			for (size_t i = 0; i < elem.eductIdList.size(); i++) {
				eductlist.push_back(&moleculeCounter[elem.eductIdList.at(i)]);
			}
			vector<uint64_t*> productlist;
			for (size_t i = 0; i < elem.productIDList.size(); i++) {
				productlist.push_back(&moleculeCounter[elem.productIDList.at(i)]);
			}
			reactions[elem.id] = new Transfer(&reactionRateCoefficientsMC[elem.coefficientID],
													eductlist,
													productlist,
													getPolymerSpecies(elem.eductIdList.at(0)),
													getPolymerSpecies(elem.productIDList.at(0)));
			break;
		}
		case ReactionTemplate::TransferToMonomer: {
			reactions[elem.id] = new TransferToMonomer(&reactionRateCoefficientsMC[elem.coefficientID],
													&moleculeCounter[elem.eductIdList.at(0)],
													&moleculeCounter[elem.eductIdList.at(1)],
													&moleculeCounter[elem.productIDList.at(0)],
													&moleculeCounter[elem.productIDList.at(1)],
													getPolymerSpecies(elem.eductIdList.at(0)),
													getPolymerSpecies(elem.productIDList.at(0)),
													getPolymerSpecies(elem.productIDList.at(1)),
													elem.eductIdList.at(1),
													&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(1)]);
			break;
		}
		case ReactionTemplate::TransferToPolymer: {
			reactions[elem.id] = new TransferToPolymer(&reactionRateCoefficientsMC[elem.coefficientID],
				&moleculeCounter[elem.eductIdList.at(0)],
				&moleculeCounter[elem.eductIdList.at(1)],
				&moleculeCounter[elem.productIDList.at(0)],
				&moleculeCounter[elem.productIDList.at(1)],
				getPolymerSpecies(elem.eductIdList.at(0)),
				getPolymerSpecies(elem.eductIdList.at(1)),
				getPolymerSpecies(elem.productIDList.at(0)),
				getPolymerSpecies(elem.productIDList.at(1)),
				elem.eductIdList.at(2),
				elem.productIDList.at(2),
				&solidsContentBalance.solidsContentMoleculeCounter[elem.eductIdList.at(2)],
				&solidsContentBalance.solidsContentMoleculeCounter[elem.productIDList.at(2)],
				PropagatingRadicalType::MCRL);
			for (auto& elemSpecies : macromolSpecies) {
				if (elemSpecies->getID() == elem.eductIdList.at(1)) {
					elemSpecies->setTransferMonomer(elem.eductIdList.at(2));
					break;
				}
			}
			break;
		}
		default:
			break;
		} // switch
	}
}

void Simulation::runTo(double _destTime, modelXML* _mcPolymerModel) {
	const vector<coefficientObject> coefficientList = _mcPolymerModel->calculateAndGetCoefficients();
	for (size_t i = 0; i < coefficientList.size(); i++) {
		reactionRateCoefficientsMC[i] = coefficientList.at(i).valueMC;
	}
	processPlugins(_mcPolymerModel, _destTime);
	moleculeCounterLastStep = moleculeCounter;
	valarray<bool> recalculateProbabilities(true, reactions.size());
	valarray<double> reactionProbability(0.0, reactions.size());

	while (t < _destTime) {
		double sumProbabilities = 0.0;
		for (size_t reactIdx = 0; reactIdx < reactions.size(); reactIdx++) {
				reactionProbability[reactIdx] = reactions[reactIdx]->calculatePartialProbability();
				sumProbabilities += reactionProbability[reactIdx];
		}

		if (sumProbabilities == 0.0) {
			t = _destTime;
			return;
		}

		double randr = randBufferInstance->getRand() * sumProbabilities;
		double cumulativeSum = 0.0;
		for (size_t reactIdx = 0; reactIdx < reactions.size(); reactIdx++) {
			cumulativeSum += reactionProbability[reactIdx];
			if (randr <= cumulativeSum) {
				reactions[reactIdx]->execute();
				break;
			}
		}
		randr = randBufferInstance->getRand() + 1.0e-40;
		t -= log(randr) / sumProbabilities;
	}
}

double Simulation::runUntilRadicalTransfer(int _idRadical, double _maxTime, modelXML* _mcPolymerModel) {
	const vector<coefficientObject> coefficientList = _mcPolymerModel->calculateAndGetCoefficients();
	for (size_t i = 0; i < coefficientList.size(); i++) {
		reactionRateCoefficientsMC[i] = coefficientList.at(i).valueMC;
	}
	valarray<double> reactionProbability(0.0, reactions.size());
	while (true) {
		double sumProbabilities = 0.0;
		for (size_t reactIdx = 0; reactIdx < reactions.size(); reactIdx++) {
			reactionProbability[reactIdx] = reactions[reactIdx]->calculatePartialProbability();
			sumProbabilities += reactionProbability[reactIdx];
		}
		if (sumProbabilities == 0.0) {
			t = _maxTime;
			return t;
		}
		double randr = randBufferInstance->getRand() * sumProbabilities;
		double cumulativeSum = 0.0;
		for (size_t reactIdx = 0; reactIdx < reactions.size(); reactIdx++) {
			cumulativeSum += reactionProbability[reactIdx];
			if (randr <= cumulativeSum) {
				reactions[reactIdx]->execute();
				break;
			}
		}
		randr = randBufferInstance->getRand() + 1.0e-40;
		t -= log(randr) / sumProbabilities;
		if (moleculeCounter[_idRadical] == 1) {
			return t;
		}
		if (t >= _maxTime) {
			return _maxTime;
		}
	}
}
uint64_t Simulation::getMolecules(int _id) {
	return moleculeCounter[_id];
}
uint64_t Simulation::getMoleculesMonomerInPolymer(int _id) {
	return solidsContentBalance.solidsContentMoleculeCounter[_id];
}
void Simulation::setMoleculesMonomerInPolymer(int _id, uint64_t _numMolecules) {
	solidsContentBalance.solidsContentMoleculeCounter[_id] = _numMolecules;
}
polymerSpeciesProperties Simulation::getPolymerSpeciesProperties(int _id) const {
	polymerSpeciesProperties properties;
	int indexMacro = 0;
	for (auto& elem : macromolSpecies) {
		if (elem->getID() == _id) {
			properties = macromolSpecies[indexMacro]->getPolymerSpeciesProperties();
		}
		++indexMacro;
	}
	return properties;
}

int Simulation::setTime(double _newTime) {
	if (_newTime > t) {
		t = _newTime;
	}
	return 1;
}
int Simulation::getTime(double* _internalTime) {
	*_internalTime = t;
	return 1;
}
void Simulation::setMolecules(int _id, uint64_t _count) {
	moleculeCounter[_id] = _count;
}

void Simulation::addMolecules(int _id, uint64_t _count) {
	moleculeCounter[_id] += _count;
}
void Simulation::addPolymerMolecule(int _speciesID, Polymer* _p) {
	moleculeCounter[_speciesID] += 1;
	for (size_t idx = 0; idx < macromolSpecies.size(); idx++) {
		if (macromolSpecies[idx]->getID() == _speciesID) {
			macromolSpecies[idx]->addPolymer(_p);
			break;
		}
	}
	for (size_t i = 0; i < solidsContentBalance.solidsContentMoleculeCounter.size(); i++) {
		solidsContentBalance.solidsContentMoleculeCounter[i] += _p->getM(int(i));
	}
}
Polymer* Simulation::getPolymerMolecule(int _speciesID) {
	moleculeCounter[_speciesID] -= 1;
	Polymer* p = nullptr;
	for (size_t idx = 0; idx < macromolSpecies.size(); idx++) {
		if (macromolSpecies[idx]->getID() == _speciesID) {
			p = macromolSpecies[idx]->deleteP();
			break;
		}
	}
	if (p != nullptr) {
		for (size_t i = 0; i < solidsContentBalance.solidsContentMoleculeCounter.size(); i++) {
			solidsContentBalance.solidsContentMoleculeCounter[i] -= p->getM(int(i));
		}
	}
	return p;
}
vector<Polymer*> Simulation::getAllPolymers(int _speciesID) {
	for (size_t idx = 0; idx < macromolSpecies.size(); idx++) {
		if (macromolSpecies[idx]->getID() == _speciesID) {
			return(macromolSpecies[idx]->getAllPolymers());
		}
	}
	vector<Polymer*> noPolymers;
	return (noPolymers);
}
void Simulation::setRateCoeficient(int _id, double _value) {
	reactionRateCoefficientsMC[_id] = _value;
}

void Simulation::updateMCRates() {
	// !!!
}

void Simulation::importSpecies(int _indexS, int _indexM, const string& _filename, int _format) {
	// !!!
	/*
	polymerSpecies* polyS;
	uint64_t numberMolecules;
	polyS = ci->Reactants[_indexS - 1]->getPolymerS();
	if (_format == 0) {
		// binary
		polyS->importSpecies(_filename);
		return;
	}
	if (_format == 1) {
		// chain length distribution
		numberMolecules = ci->Reactants[_indexS - 1]->getMolcount();
		polyS->importSpeciesCLD(_filename, _indexM, numberMolecules);
		return;
	}
	*/
}

bool Simulation::exportPolymerSpecies(int _id, const std::string& _filename, FileTypePolymerSpecies _fileType, const modelXML* _mcPolymerModel) const {
	int indexMacro = 0;
	for (auto& elem : macromolSpecies) {
		if (elem->getID() == _id) {
			switch (_fileType) {
			case FileTypePolymerSpecies::CLD: {
				macromolSpecies[indexMacro]->exportCLD(_filename, _mcPolymerModel);
				return true;
			}
			case FileTypePolymerSpecies::Binary: {
				string filenamebin = _filename + ".bin";
				ofstream binfile(filenamebin, ofstream::out | ofstream::binary);
				vector<string> monomerName;
				vector<double> molarMass;
				const vector<speciesObject> speciesObjects = _mcPolymerModel->getSpecies();
				for (auto& elem : speciesObjects) {
					if (elem.isMonomer) {
						monomerName.push_back(elem.name);
						molarMass.push_back(elem.optionalParameterMolarMass);
					}
				}
				size_t countMonomers = monomerName.size();
				binfile.write((const char*)&countMonomers, sizeof(size_t));

				for (auto& elem : monomerName) {
					string wstr = elem;
					size_t wstrl = wstr.length();
					binfile.write((const char*)&wstrl, sizeof(size_t));
					binfile.write((const char*)wstr.c_str(), wstrl);
				}
				for (auto& elem : molarMass) {
					double MM = elem;
					binfile.write((const char*)&MM, sizeof(double));
				}
				size_t numPolymers = moleculeCounter[_id];
				binfile.write((const char*)&numPolymers, sizeof(size_t));
				macromolSpecies[indexMacro]->exportPolymers(&binfile);
				return true;
			}
			default:
				break;
			}
		}
		++indexMacro;
	}
	return false;
}
bool Simulation::exportPolymers(const string& _filename, FileTypePolymerSpecies _fileType, const modelXML* _mcPolymerModel) const {
	const vector<speciesObject> speciesObjects = _mcPolymerModel->getSpecies();
	vector<Histogram*> monomerHistogram;
	vector<string> monomerName;
	vector<double> molarMass;
	for (auto& elem : speciesObjects) {
		if (elem.isMonomer) {
			monomerName.push_back(elem.name);
			molarMass.push_back(elem.optionalParameterMolarMass);
			Histogram* MH = new Histogram();
			monomerHistogram.push_back(MH);
		}
	}
	if (monomerName.size() == 0) {
		// monomers not present in current kinetic model
		return false;
	}
	switch (_fileType) {
	case FileTypePolymerSpecies::CLD: {
		// FIXME: subroutines
		Histogram* CLD = new Histogram();

		for (auto& elem : macromolSpecies) {
			elem->getCLD(CLD, monomerHistogram);
		}
		if (CLD->getSx() == 0) {
			// no polymer chains present
			return false;
		}
		string filenamecld = _filename + ".CLD.dat";
		ofstream CLDfile(filenamecld, ofstream::out);
		CLDfile << "numMonomers:\t" << monomerName.size() << endl;

		vector<double> partMass;
		uint64_t sumMonomers = 0;
		for (size_t i = 0; i < monomerHistogram.size(); i++) {
			CLDfile << monomerName.at(i) << ":\t" << molarMass.at(i) << endl;
			// calculate average molar mass
			partMass.push_back(monomerHistogram.at(i)->getSy1() * 1.0 * molarMass.at(i));
			sumMonomers += monomerHistogram.at(i)->getSy1();
		}
		double averageMolarMass = 0.0;
		for (size_t i = 0; i < monomerHistogram.size(); i++) {
			averageMolarMass += partMass.at(i);
		}
		averageMolarMass = averageMolarMass / (sumMonomers * 1.0);

		CLDfile << "Mn:\t" << CLD->getPn() * averageMolarMass << "\tMw:\t" << CLD->getPw() * averageMolarMass << "\tMw/Mn:\t" << CLD->getPw() / CLD->getPn() << endl;
		CLDfile << "numChains:\t" << CLD->getSy1() << "\tnumSCB:\t" << CLD->getSy2() << "\tnumLCB:\t" << CLD->getSy3() << endl;
		CLDfile << "Pn:\t" << CLD->getPn() << "\tPw:\t" << CLD->getPw() << "\tPw/Pn:\t" << CLD->getPw() / CLD->getPn() << endl;
		CLDfile << "chainLength\tnumChains\tnumSCB\tnumLCB";
		for (size_t i = 0; i < monomerName.size(); i++) {
			CLDfile << "\t" << monomerName.at(i);
		}
		CLDfile << endl;
		for (size_t i = 0; i < CLD->getSize(); i++) {
			CLDfile << CLD->getX(i) << "\t" << CLD->getY1(i) << "\t" << CLD->getY2(i) << "\t" << CLD->getY3(i);
			for (size_t j = 0; j < monomerHistogram.size(); j++) {
				CLDfile << "\t" << monomerHistogram.at(j)->getY1(i);
			}
			CLDfile << endl;
		}
		return true;
	}
	case FileTypePolymerSpecies::Binary: {
		string filenamebin = _filename + ".bin";
		ofstream binfile(filenamebin, ofstream::out | ofstream::binary);
		vector<string> monomerName;
		vector<double> molarMass;
		const vector<speciesObject> speciesObjects = _mcPolymerModel->getSpecies();
		for (auto& elem : speciesObjects) {
			if (elem.isMonomer) {
				monomerName.push_back(elem.name);
				molarMass.push_back(elem.optionalParameterMolarMass);
			}
		}
		size_t countMonomers = monomerName.size();
		binfile.write((const char*)&countMonomers, sizeof(size_t));

		for (auto& elem : monomerName) {
			string wstr = elem;
			size_t wstrl = wstr.length();
			binfile.write((const char*)&wstrl, sizeof(size_t));
			binfile.write((const char*)wstr.c_str(), wstrl);
		}
		for (auto& elem : molarMass) {
			double MM = elem;
			binfile.write((const char*)&MM, sizeof(double));
		}
		size_t numPolymers = 0;
		for (auto& elem : macromolSpecies) {
			numPolymers += moleculeCounter[elem->getID()];
		}
		binfile.write((const char*)&numPolymers, sizeof(size_t));
		for (auto& elem : macromolSpecies) {
			elem->exportPolymers(&binfile);
		}
		return true;
	}
	default:
		break;
	}
	return false;
}

vector<int> Simulation::getPolymerSpeciesChainLengthList(int _id) {
	vector<int> chainLengthList;
	for (auto& elem : macromolSpecies) {
		if (elem->getID() == _id) {
			chainLengthList = elem->getChainLengthList();
			break;
		}
	}
	return chainLengthList;
}
void Simulation::registerPlugins(modelXML* _modelXML, const string& _outputDir) {
	registeredPlugins = _modelXML->getPlugins();
	const vector<coefficientObject> coefficientList = _modelXML->calculateAndGetCoefficients();
	for (auto& elem : registeredPlugins) {
		string filename = _outputDir + "/logPlugin_" + elem.templateLabelFilename + "_ID_" + to_string(elem.id) + ".dat";
		ofstream resLogFile(filename);
		switch (elem.type) {
		case PluginType::TerminationHamielec: {
			resLogFile << "time\tconversionStyrene\tkt\tktMC" << endl;
			resLogFile << "0.0\t0.0\t" << coefficientList.at(elem.coefficientID).value << "\t" << coefficientList.at(elem.coefficientID).valueMC << endl;
			TerminationHamielec* pHamielec = new TerminationHamielec(elem.id, _outputDir);
			pHamielec->setCountStyrene0(getMolecules(elem.speciesIdList.at(0)));
			plugin.push_back(pHamielec);
			break;
		}
		case PluginType::TerminationChainLengthDependent: {
			resLogFile << "time\tnumP\tkt\tktMC" << endl;
			resLogFile << "0.0\t0\t" << coefficientList.at(elem.coefficientID).value << "\t" << endl;
			TerminationChainLengthDependent* pCLD = new TerminationChainLengthDependent(elem.id, _outputDir);
			pCLD->setParameter(elem.chainLengthDependenceParameter.ic, elem.chainLengthDependenceParameter.as, elem.chainLengthDependenceParameter.al, elem.chainLengthDependenceParameter.as2, elem.chainLengthDependenceParameter.as3, elem.chainLengthDependenceParameter.as4, elem.chainLengthDependenceParameter.kt11p1, elem.chainLengthDependenceParameter.kt11p2, elem.chainLengthDependenceParameter.kt11p3);
			plugin.push_back(pCLD);
			break;
		}
		default:
			break;
		}
	}
}
void Simulation::processPlugins(modelXML* _modelXML, double _simulationTimeSeconds) {
	for (const auto& elem : registeredPlugins) {
		string outputDir = "";
		for (size_t i = 0; i < plugin.size(); i++) {
			if (elem.id == plugin.at(i)->getID()) {
				outputDir = plugin.at(i)->getOutputDir() + "/";
			}
		}
		string filename = outputDir + "logPlugin_" + elem.templateLabelFilename + "_ID_" + to_string(elem.id) + ".dat";
		ofstream resLogFile(filename, ofstream::app);
		switch (elem.type) {
		case PluginType::TerminationHamielec:
			processPluginTerminationHamielec(elem, _modelXML, _simulationTimeSeconds, resLogFile);
			break;
		case PluginType::TerminationChainLengthDependent:
			processPluginTerminationChainLengthDependent(elem, _modelXML, _simulationTimeSeconds, resLogFile);
			break;
		case PluginType::Undefined:
			logError("Simulation::processPlugins() processing undefined plugin type");
			break;
		}
	}
}

void Simulation::processPluginTerminationHamielec(const pluginObject& _object, modelXML *_modelXML, double _simulationTimeSeconds, std::ofstream& _resLogFile) {
	for (size_t i = 0; i < plugin.size(); i++) {
		if (_object.id != plugin.at(i)->getID())
			continue;
		TerminationHamielec* p = dynamic_cast<TerminationHamielec*>(plugin.at(i));
		double ct = static_cast<double>(getMolecules(_object.speciesIdList.at(0)));
		double c0 = static_cast<double>(p->getCountStyrene0());
		double conversionStyrene = (c0 - ct) / c0;
		double f = p->getScaleFactor(_modelXML->getTemperature(), conversionStyrene);
		double currentValueMC = _modelXML->calculateAndGetCoefficients().at(_object.coefficientID).valueMC * f;
		setRateCoeficient(_object.coefficientID, currentValueMC);
		_resLogFile << _simulationTimeSeconds << "\t" << conversionStyrene << "\t" << _modelXML->calculateAndGetCoefficients().at(_object.coefficientID).value * f << "\t" << currentValueMC << endl;
		break;
	}
}

void Simulation::processPluginTerminationChainLengthDependent(const pluginObject& _object, modelXML *_modelXML, double _simulationTimeSeconds, std::ofstream& _resLogFile) {
	for (size_t i = 0; i < plugin.size(); i++) {
		if (_object.id != plugin.at(i)->getID())
			continue;
		TerminationChainLengthDependent* p = dynamic_cast<TerminationChainLengthDependent*>(plugin.at(i));
		const vector<reactionObject> reactionList = _modelXML->getReactions();
		vector<int> cli;
		vector<int> clj;
		for (auto& elemReaction : reactionList) {
			if (elemReaction.coefficientID == _object.coefficientID) {
				if (elemReaction.eductIdList.size() == 2) {
					if (elemReaction.eductIdList.at(0) == elemReaction.eductIdList.at(1)) {
						vector<int> polymerSpeciesChainLengthList = getPolymerSpeciesChainLengthList(elemReaction.eductIdList.at(0));
						if (polymerSpeciesChainLengthList.size() > 1) {
							for (size_t j = 0; j < polymerSpeciesChainLengthList.size(); j++) {
								for (size_t k = j + 1; k < polymerSpeciesChainLengthList.size(); k++) {
									if (j != k) {
										cli.push_back(polymerSpeciesChainLengthList.at(j));
										clj.push_back(polymerSpeciesChainLengthList.at(k));
									}
								}
							}
						}
					}
					else {				
						vector<int> polymerSpeciesChainLengthList1 = getPolymerSpeciesChainLengthList(elemReaction.eductIdList.at(0));
						vector<int> polymerSpeciesChainLengthList2 = getPolymerSpeciesChainLengthList(elemReaction.eductIdList.at(1));
						if ((polymerSpeciesChainLengthList1.size() > 0) && (polymerSpeciesChainLengthList2.size() > 0)) {
							for (size_t k = 0; k < polymerSpeciesChainLengthList1.size(); k++) {
								for (size_t j = 0; j < polymerSpeciesChainLengthList2.size(); j++) {
									cli.push_back(polymerSpeciesChainLengthList1.at(k));
									clj.push_back(polymerSpeciesChainLengthList2.at(j));
								}
							}
						}
					}
				}
			}
		}
		if (cli.size() > 0) {
			double ktMC = 0.0;
			double kt = 0.0;
			for (const auto& elemCoefficient : _modelXML->calculateAndGetCoefficients()) {
				if (elemCoefficient.id == _object.coefficientID) {
					kt = elemCoefficient.value;
					ktMC = elemCoefficient.valueMC;
				}
			}
			double factorCliClj = p->calcFactorCliClj(cli, clj);
			kt *= factorCliClj;
			ktMC *= factorCliClj;
			setRateCoeficient(_object.coefficientID, ktMC);
			_resLogFile << _simulationTimeSeconds << "\t" << cli.size() << "\t" << kt << "\t" << ktMC << endl;
		}
		else {
			double ktMC = 0.0;
			double kt = 0.0;
			for (const auto& elemCoefficient : _modelXML->calculateAndGetCoefficients()) {
				if (elemCoefficient.id == _object.coefficientID) {
					kt = elemCoefficient.value;
					ktMC = elemCoefficient.valueMC;
				}
			}
			setRateCoeficient(_object.coefficientID, ktMC);
			_resLogFile << _simulationTimeSeconds << "\t0\t" << kt << "\t" << ktMC << endl;
		}
	}
}

} // namespace
