// mcPolymer - Monte Carlo simulation program of radical polymerization
// Copyright (C) 2008 - 2021 Marco Drache, Georg Drache, Benjamin Hosemann, Katrin Brandl
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

/*
* class Simulation
*
*  Created on: 24.08.2008
*/

#pragma once

#include <vector>
#include <valarray>
#include "modelObjects.h"
#include "modelXML.h"
#include "Reaction.h"
#include "Plugin.h"
#include "polymerSpecies.h"


namespace mcPolymer_ns {

class Simulation {
public:
	Simulation(modelXML* _mcPolymerModel,const std::string& _outputDir);
	void runTo(double _destTime, modelXML* _mcPolymerModel);
	double runUntilRadicalTransfer(int _idRadical, double _maxTime, modelXML* _mcPolymerModel);
	virtual ~Simulation();

	int setTime(double _newTime);
	int getTime(double *_internalTime);
	void updateMCRates();

	bool exportPolymerSpecies(int _id, const std::string& _filename, FileTypePolymerSpecies _fileType, const modelXML* _mcPolymerModel) const;
	bool exportPolymers(const std::string& _filename, FileTypePolymerSpecies _fileType, const modelXML* _mcPolymerModel) const;
	void setMolecules(int index, uint64_t count);
	void setRateCoeficient(int _id, double _value);
	void addMolecules(int _id, uint64_t _plusmole);
	void addPolymerMolecule(int _speciesID, Polymer* _p);
	Polymer* getPolymerMolecule(int _speciesID);
	std::vector<Polymer*> getAllPolymers(int _speciesID);
	void importSpecies(int indexS, int indexM, const std::string& _filename, int _format);

	uint64_t getMolecules(int _id);
	uint64_t getMoleculesMonomerInPolymer(int _id);
	void setMoleculesMonomerInPolymer(int _id, uint64_t _numMolecules);

	polymerSpeciesProperties getPolymerSpeciesProperties(int _id) const;
	std::vector<int> getPolymerSpeciesChainLengthList(int _id);

private:
	struct cld {
		double al = 0.0;
		double as = 0.0;
		double ic = 0.0;
	};
	struct reactionObjectSimulation {
		double kMC;
		int coefficientID;
		double reactionProbability = 0;
		std::valarray<uint64_t*> educts;
		std::valarray<uint64_t*> products;
		std::valarray<polymerSpecies*> macromolEducts;
		std::valarray<polymerSpecies*> macromolProducts;
		ReactionTemplate rktTemplate = ReactionTemplate::Undefined;
		bool order2EE;
		size_t eductsSize;
		size_t productsSize;
		double optionalDoubleParameter;
		std::valarray<int> eductID;
		std::valarray<int> productID;
	};
	struct solidsContent {
		std::valarray<uint64_t*> monomerMoleculeCounter;
		std::valarray<uint64_t> solidsContentMoleculeCounter;
	};

	size_t initMoleculeCounter(const std::vector<speciesObject>& _speciesList);
	void initSolidsContentBalance();
	void initMacromolSpecies(const std::vector<speciesObject>& _speciesList, size_t _numMacroSpecies);
	void initCoefficients(modelXML* _modelXML);
	void initReactions(const modelXML* _modelXML, const std::vector<speciesObject>& _speciesList);

	void registerPlugins(modelXML* _modelXML, const std::string& _outputDir);
	void processPlugins(modelXML* _modelXML, double _time);

	void processPluginTerminationHamielec(const pluginObject& _object, modelXML *_modelXML, double _simulationTimeSeconds, std::ofstream& _resLogFile);
	void processPluginTerminationChainLengthDependent(const pluginObject& _object, modelXML* _modelXML, double _simulationTimeSeconds, std::ofstream& _resLogFile);
	polymerSpecies* getPolymerSpecies(int _id);
	std::valarray<uint64_t> moleculeCounter;
	std::valarray<uint64_t> moleculeCounterLastStep;
	std::valarray<double> reactionRateCoefficientsMC;
	std::valarray<polymerSpecies*> macromolSpecies;
	solidsContent solidsContentBalance;
	std::vector<pluginObject> registeredPlugins;
	std::vector<Plugin*> plugin;
	std::valarray<Reaction*> reactions;
	double t = 0.0;
};

} // namespace
